/*----------------------- Rogaland University Centre -------------------------
## \section{blurr}
## 
## 
## \subsection*{Purpose}
## Perform blurring (smoothing/low-pass filtering) on an image filtered by 
## the program mcf without blurring or qmf.
## 
## \subsection*{Description}
## The blurring of a particular filtered frame is determined by the
## centre frequency of the corresponding passband. 
## This program normally reads the filter structure in the 
## same manner as the program qmf and performs blurring by the
## appropriate filter.
## It is also possible to specify a fixed centre frequency, as an
## alternative to specifying he filter structure.
## Normal operation in image segmentation is to preceed the application
## of this program by the application of the tanh program
## 	
## See: ``Multi-channel filtering techniques for texture segmentation and
##      surface quality inspection'',
##      Farshid Farrokhnia, Ph.D. dissertation,
##      Michigan State University, Dept. of Elect. Eng., 1990
## 
## \subsection*{Usage}
## \begin{verbatim}
## blurr [-sn struct<integer> | 
##        -s struct<string> | 
##        -sb struct<string> | 
##        -cf centre frequency<real>]
##       [-sf gaussian window factor<real>]
## \end{verbatim}
## \begin{description}
## \item[-s]
## 	Filter structure string, defined as a quad tree.
## 	See the module {\em filtertree} (CompileFiltStructTree)
## 	for a description of the structure.
## \item[-sb]
## 	Filterstructure string, defined as a binary tree.
## 	See the module {\em filtertree} (CompileFiltStructTreeFromBin)
## 	for a description of the structure.
## \item[-sn]
## 	Filter structure number, between 1 and 11 (default 4).
## 	See the module {\em filtertree} (BuildFiltStructTree)
## 	for a description of the structure.
## \item[-cf]
## 	Centre frequency which to tune the Gaussian lowpass filter to.
## 	Refer to Farshid Farrokhnia's Ph. D. dissertation for a description
## 	of the relationship between centre frequency and gaussian lowpass
## 	filter.
## \item[-d]
## 	Blurr a critically sampled filter output, as created by 
## 	{\tt qmf -d $\cdots$}.
## \end{description}
## 
## 
## \subsection*{Date created}
## March 1993
 *------------------------ Rogaland University Centre -----------------------*/



/* Header files */
#include "qmf.h"
#include "mcf.h"

static Flag_Format flagfmt[]={
  {"d",{LASTFLAG},0,{
    {PTBOOLEAN,"FALSE"},
    LASTPARAMETER}},
  {"sn",{"s","sb","cf",LASTFLAG},1,{
    {PTBOOLEAN,"TRUE"},
    {PTINT,"4","struct"},
    LASTPARAMETER}},
  {"sb",{"sn","s","cf",LASTFLAG},1,{
    {PTBOOLEAN,"FALSE"},
    {PTSTRING,"(,)","struct"},
    LASTPARAMETER}},
  {"s",{"sn","sb","cf",LASTFLAG},1,{
    {PTBOOLEAN,"FALSE"},
    {PTSTRING,"(,)","struct"},
    LASTPARAMETER}},
  {"cf",{"s","sn","sb",LASTFLAG},1,{
    {PTDOUBLE,"0.5","centre frequency"},
    LASTPARAMETER}},    
  {"sf",{LASTFLAG},1,{
    {PTDOUBLE,DEFAULT_SIGMA_FACTOR,"gaussian window factor"}, 
    LASTPARAMETER}},
  LASTFLAG};

int types[]={PFCOMPLEX,LASTTYPE};

main(argc,argv)
     int argc;
     char *argv[];
{
  struct header hd;		/* Input file header */
  struct header hdp;		/* Converted input header */
  int method, method_c;		/* Conversion method */
  Boolean decimateFlag;		/* Parameter -d */
  Boolean structStr1DFlag;	/* Structure string specified (bin) */
  Boolean structStrFlag;	/* Structure string specified (quad) */
  Boolean structNoFlag;		/* Structure number specified */
  int struct_no;
  char *structStr1D;		/* Structure description string (bin) */
  char *structStr;		/* Structure description string (quad) */
  FilterTree filtertree;	/* Filter structure tree */
  int n_subbnds;		/* Number of subbands */
  double centrFreq;		/* Centre frequency */
  double sigmaFactor;
  int frame=0;

  hipserrprt = HEL_ERROR;	/* Suppress messaging from HIPS routines */

  /* Initialize variables */
  Progname=strsave(*argv);
  parseargs(argc,argv,flagfmt,&decimateFlag,&structNoFlag,&struct_no,
	    &structStr1DFlag,&structStr1D,&structStrFlag,&structStr,
	    &centrFreq,&sigmaFactor,FFNONE);

  /* Build the filter structure binary tree */
  if (structStrFlag) 
    filtertree = CompileFiltStructTree(structStr,"");
  else if (structStr1DFlag) 
    filtertree = CompileFiltStructTreeFromBin(structStr1D,"");
  else if (structNoFlag)
    filtertree = BuildFiltStructTree(struct_no,"");
 
  /* Read the input image header and write the output */
  read_hdr_a(&hd);
  method = set_conversion(&hd,&hdp,types);
  write_headeru(&hd,argc,argv);

  if (structStrFlag || structStr1DFlag || structNoFlag) {
    n_subbnds = getNSubbands (filtertree);
    if (n_subbnds != hd.num_frame && !decimateFlag) 
      perr(HE_FRMATCH);
    PrintFiltStructTree (filtertree);
    blurr_frame (&hd,&hdp,method,filtertree,sigmaFactor,0.,0.,0.5,0.5,
		 &frame,decimateFlag);
    if (decimateFlag)
      write_imagec(&hd,&hdp,method,TRUE,frame);
    fprintf(stderr,"\n");
  }
  else {
    /* Don't calculate centre frequency from filtertree, use the parameter
       supplied from the command line */
    for (frame=0; frame<hdp.num_frame; frame++) {
      read_imagec(&hd,&hdp,method,frame);	
      mcf_gauss(&hdp,&hdp,sigmaFactor/centrFreq);
      write_imagec(&hd,&hdp,method,TRUE,frame);
    }
  }
}




blurr_frame (hd,hdp,method,ft,sf,lx,ly,ux,uy,frame,decimate)
     struct header *hd, *hdp;
     int method;
     FilterTree ft;	/* Filter structure tree */
     double sf;		/* Factor for sigma in blurrer */
     float lx, ly;	/* Lower left corner in the freq. domain */
     float ux, uy;	/* Upper right corner in the freq. domain */
     			/* Remember that max frequency is 0.5 */
     int *frame;	/* Input/output frame */
     Boolean decimate;	/* Operate on decimated image? */
{
  int ro, kappa;
  double sigma;
  double centreFreq;
  int fr, fc, nr, nc;	/* Parameters to setroi */

  if (ft != NULL) 
    for (ro=0; ro<2; ro++) 
      for (kappa=0; kappa<2; kappa++) 
        blurr_frame (hd,hdp,method,
		     ft->subband[ro][kappa],
		     (decimate ? sf/2.0 : sf),
		     (ro==0 ? lx : (lx+ux)/2.0),
		     (kappa==0 ? ly : (ly+uy)/2.0),
		     (ro==1 ? ux : (lx+ux)/2.0),
		     (kappa==1 ? uy : (ly+uy)/2.0),
                     frame,decimate);
  else {
    fprintf(stderr,"'");	/* Life sign */
    centreFreq = 
      sqrt(sqr((lx+ux)/2.0) + sqr((ly+uy)/2.0));
    sigma = sf/centreFreq;
    if (decimate) {
      if ((*frame)==0) {
	read_imagec(hd,hdp,method,*frame);	
	(*frame)++;
      }
      fr = (int) (2.0*lx*hdp->orows);
      fc = (int) (2.0*ly*hdp->ocols);
      nr = (int) (2.0*(ux-lx)*hdp->orows);
      nc = (int) (2.0*(uy-ly)*hdp->ocols);
      setroi(hdp,fr,fc,nr,nc);
      mcf_gauss(hdp,hdp,sigma);
      setroi(hdp,0,0,hdp->orows,hdp->ocols);
    }
    else {
      read_imagec(hd,hdp,method,*frame);	
      mcf_gauss(hdp,hdp,sigma);
      write_imagec(hd,hdp,method,TRUE,*frame);
      (*frame)++;
    }
  }
}
  
