/*----------------------- Rogaland University Centre -------------------------
## \section{centrfreq}
## 
## 
## \subsection*{Purpose}
## Determine the centre frequencies of a general Gabor filter giving the
## maximum step signature for a given texture pair.
## 
## 
## \subsection*{Description}
## See: ``Determining Gabor-filter parameters for texture segmentation'',
##      Dennis F. Dunn, William E. Higgins and Joseph Wakeley,
##      SPIE Vol. 1826, Intelligent robots and computer vision XI (1992),
##      Pages 51-63.
## 
## \subsection*{Usage}
## \begin{verbatim}
## centrfreq
##     [-s % of samples to use for 
##           estimating centre
##           frequencies<real>]
##     [-sigma SigmaX<real> [SigmaY<real>]]
##     [-seed random number seed<integer>] 
##     [-2mcf]
##     imagefile1 [<] imagefile2
## \end{verbatim}
## \begin{description}
## \item[-s]
## 	Percentage of pixels to use in computation (default 5)
## \item[-sigma]
## 	Parameter in the Gabor filter
## \item[-seed]
## 	Pseudo-random number generator seed
## \item[-2mcf]
## 	Make the output be of the format required by mcf
## \end{description}
## 
## 
## 
## \subsection*{Date created}
## March 1993
 *------------------------ Rogaland University Centre -----------------------*/



#include <stdlib.h>
#include <limits.h>
#include <float.h>
#include <hipl_format.h>
#include <math.h>
#include "mcf.h"
#include "math.h"

#define INFO_STR "Computing centre frequencies"

static Flag_Format flagfmt[] = {
  {"s",{LASTFLAG},1,{
    {PTDOUBLE,"5.0","% of samples to use for estimating centre frequencies"}, 
    LASTPARAMETER}},
  {"sigma",{LASTFLAG},1,{
    {PTDOUBLE,"50.0","SigmaX"}, 
    {PTDOUBLE,"-1.0","SigmaY"}, 
    LASTPARAMETER}},
  {"seed",{LASTFLAG},1,{{PTINT,"-1","random number seed"}, LASTPARAMETER}},
  {"2mcf",{LASTFLAG},0,{{PTBOOLEAN,"FALSE"}, LASTPARAMETER}},
  LASTFLAG};

/* Convert imported data to float */
int types[] = {PFFLOAT,LASTTYPE};

double SigmaX, SigmaY;



main(argc,argv)
     int argc;
     char **argv;
{
  Filename filename[2];
  double sperc;
  struct header mean[2], var[2], PE;
  int i,U,V;
  float min;
  int seed;
  Boolean toMcf;

  hipserrprt = HEL_ERROR;
  Progname = strsave(*argv);
  parseargs(argc,argv,flagfmt,&sperc,&SigmaX,&SigmaY,&seed,&toMcf,
	    FFTWO,&(filename[0]),&(filename[1]));

  if(seed < 0)
    seed = time(NULL);
  srand48((long int) seed);
  if(SigmaY < 0)
    SigmaY = SigmaX;

  /* Compute the mean and variance */
  for(i=0; i<2; i++) 
    getMeanVar (&(mean[i]),&(var[i]),filename[i],sperc,i);

  /* Compute PE find centre frequencies */
  get_PE (&(mean[0]),&(mean[1]),&(var[0]),&(var[1]),&PE);
  min_posn(&PE,&U,&V,&min);

  printPstFin(INFO_STR,100,100);
  if (toMcf) {
    /* Print output as parameterstring to mcf */
    printf("-Gen %*.*lf %*.*lf -U0 1 %*.*lf  -V0 1 %*.*lf\n",
	   DBL_DIG+2,DBL_DIG,SigmaX,DBL_DIG+2,DBL_DIG,SigmaY,
	   DBL_DIG+2,DBL_DIG,TO_FR(U,PE.cols),
	   DBL_DIG+2,DBL_DIG,TO_FR(V,PE.rows));
  }
  fprintf(stderr,"U0=%d, V0=%d and min(PE)=%e\n", U,V,min);

  return (0);
}



getMeanVar(mean,var,filename,sperc,pass)
     struct header *mean;
     struct header *var;
     Filename filename;		/* Image file */
     double sperc;		/* Percent of pixels used in computations */
     int pass;			/* First or second image (0 or 1)? */
     /* Compute the mean and variance from equations from Dunn et.al. */
{
  struct header hd;	/* Input image */
  struct header hdp;	/* Input image, converted */
  int method;
  FILE *fp;
  Coord *s;
  int s_num;
  float mean_pix;

  fp = hfopenr(filename);
  fread_hdr_a(fp,&hd,filename);
  method = fset_conversion(&hd,&hdp,types,filename);
  fread_imagec(fp,&hd,&hdp,method,0,filename);
  /* Remove the DC component; set mean to 0 */
  h_mean(&hdp,&mean_pix,FALSE);
  h_linscale(&hdp,&hdp,1.0,-mean_pix);

  /* Make list of pixel coordinates in s */
  setList(sperc,hd.rows,hd.cols,&s,&s_num);

  /* Compute mean */
  init_header(mean,"","",1,"",hdp.orows,hdp.ocols,hdp.pixel_format,1,"");
  alloc_imagez(mean);
  getMean(&hdp,mean,s,s_num,pass);

  /* Compute variance */
  init_header(var,"","",1,"",hdp.orows,hdp.ocols,hdp.pixel_format,1,"");
  alloc_imagez(var);
  getVar(&hdp,mean,var,s,s_num,pass);

  free_image(&hd);
  free_image(&hdp);
  free(s);
  return (HIPS_OK);
}



setList(perc,nr,nc,list,n)
     double perc;
     int nr, nc;
     Coord **list;
     int *n;
     /* Make a list with coordinates of perc% of the pixels in an
	nr x nc image */
{
  int i, j;
  Boolean bitmap[nr][nc];
  Coord *lstptr;

  *n=0;
  perc /= 100.0;
  for(i=0; i<nr; i++) {
    for(j=0; j<nc; j++) {
      if(drand48() <= perc) {
	(*n)++;
	bitmap[i][j] = TRUE;
      }
      else
	bitmap[i][j] = FALSE;
    }
  }
  *list = (Coord *) calloc(*n,sizeof(Coord));
  lstptr = *list;
  for(i=0; i<nr; i++) 
    for(j=0; j<nc; j++)
      if(bitmap[i][j]) {
	(*lstptr)[0] = i;
	(*lstptr)[1] = j;
	lstptr++;
      }

  return (HIPS_OK);
}



getMean(hdi,hdo,s,s_n,pass)
     struct header *hdi, *hdo;
     Coord *s;		/* List of pixels used for computation */
     int s_n;		/* Number of pixels used for computation */
     int pass;		/* 0-th r 1-st input image (for user status report) */
     /* Compute the mean from equations from Dunn et.al. */
{
  int sc;
  struct header hdt;
  float div=(float) s_n;

  init_hdr_alloc(&hdt,"","",1,"",hdi->orows,hdi->ocols,hdi->pixel_format,1,"");
  /* Traverse each pixel in the set determined by s */
  for(sc=0; sc<s_n; sc++) {
    printPstFin(INFO_STR,sc+1+2*s_n*pass,4*s_n+2);
    win_fourtr(hdi,&hdt,s,sc,SigmaX,SigmaY);
    h_abs(&hdt,&hdt);
    h_add(hdo,&hdt,hdo);
  }
  h_divscale(hdo,hdo,&div);
  free_image(&hdt);

  return (HIPS_OK);
}




getVar(hdi,mean,hdo,s,s_n,pass)
     struct header *hdi, *mean, *hdo;
     Coord *s;		/* List of pixels used for computation */
     int s_n;		/* Number of pixels used for computation */
     int pass;		/* 0-th r 1-st input image (for user status report) */
     /* Compute the variance from equations from Dunn et.al. */
{
  int sc;
  struct header hdt;
  float div = (float) s_n;

  init_hdr_alloc(&hdt,"","",1,"",hdi->orows,hdi->ocols,hdi->pixel_format,1,"");
  /* Traverse each pixel in the set determined by s */
  for(sc=0; sc<s_n; sc++) {
    printPstFin(INFO_STR,sc+1+s_n*(2*pass+1),4*s_n+2);
    win_fourtr(hdi,&hdt,s,sc,SigmaX,SigmaY);
    h_abs(&hdt,&hdt);
    h_diff(&hdt,mean,&hdt);
    h_power(&hdt,&hdt,2.0);
    h_add(hdo,&hdt,hdo);
  }
  /* Divide by number of images */
  h_divscale(hdo,hdo,&div);
  free_image(&hdt);

  return (HIPS_OK);
}




get_PE(mA,mB,vA,vB,PE)
     struct header *mA, *mB, *vA, *vB, *PE;
     /* Compute the probability of error-classification */
     /* This function performs a complex computation, and it will be best
	understood when compared with the litterature by Dunn et al., 
	referred to at the top of this file.
	To be consistent, it has been necessary to make a small alteration
	to Dunn's algorithm, since Dunn's algorithm will not be
	consistent when the two alternative limits (d's) neither lie
	between the two means. */
{
  int i,j;		/* Pixel-counters */
  int nr=mA->rows;	/* Number of rows (in all images) */
  int nc=mA->cols;	/* Number of columns */
  int nex=mA->orows - mA->rows;	/* Skip-distance, ref. ROI in HIPS */
  float *mApix,*mBpix,*PEpix,*vApix,*vBpix;	/* Pixel pointers */
  double mean[2],var[2],d[2],d_limit,tmp[2];	/* tTemporary var's */

  /* Initialize variables */
  init_hdr_alloc(PE   ,"","",1,"",mA->orows,mA->ocols,mA->pixel_format,1,"");
  mApix=(float *) mA->firstpix;
  mBpix=(float *) mB->firstpix;
  vApix=(float *) vA->firstpix;
  vBpix=(float *) vB->firstpix;
  PEpix=(float *) PE->firstpix;

  /* Perform the computation for each pixel in the image */
  for (i=0; i<nr; i++) {
    for (j=0; j<nc; j++) {
      /* Let var[0] and mean[0] correspond to the distribution with the
	 lowest mean and vice versa */
      if (*mApix < *mBpix) {
	var[0]  = (double) *vApix; 
	var[1]  = (double) *vBpix;
	mean[0] = (double) *mApix; 
	mean[1] = (double) *mBpix;
      }
      else {
	var[1]  = (double) *vApix; 
	var[0]  = (double) *vBpix;
	mean[1] = (double) *mApix;
	mean[0] = (double) *mBpix;
      }

      /* Compute the decision limit(s) */
      tmp[0] = (mean[0]*var[1] - mean[1]*var[0])
	/ (var[1] - var[0]);
      tmp[1] = sqrt(var[0]*var[1] 
		    * (sqr(mean[1]-mean[0]) 
		       + 2*(var[1]-var[0])
		       *log(sqrt(var[1]/var[0]))))
	/ (var[1] - var[0]);
      /* Let d[0] correspond to the lowest decision limit and vice versa */
      d[0] = tmp[0] - ABS(tmp[1]);
      d[1] = tmp[0] + ABS(tmp[1]);

      /* Neither d1 nor d2 lies between mA and mB */
      if (d[0]<mean[0] && d[1]>mean[1]) {
	if (var[0] < var[1])
	  *PEpix = (float) G(d[0],mean[0],var[0]) + 1 - G(d[1],mean[0],var[0])
	    + G(d[1],mean[1],var[1]) - G(d[0],mean[1],var[1]);
	else
	  *PEpix = (float) G(d[1],mean[1],var[1]) + 1 - G(d[1],mean[1],var[1])
	    + G(d[1],mean[0],var[0]) - G(d[0],mean[0],var[0]);
      }

      /* Either d1 or d2 lies between mA and mB */
      else {
	if (mean[0] <= d[0] && d[0] <= mean[1])
	  d_limit = d[0];
	else
	  d_limit = d[1];
	*PEpix = (float) G(d_limit,mean[1],var[1])
	  + 1 - G(d_limit,mean[0],var[0]);
      }

      /* Increment pointers */
      mApix++; mBpix++; PEpix++; vApix++; vBpix++;
    }
    mApix += nex; mBpix += nex; vApix += nex; vBpix += nex; PEpix += nex;
  }
  return (HIPS_OK);
}






min_posn(hd,x,y,min)
     struct header *hd;
     int *x,*y;
     float *min;
     /* Get the pixel coordinates for the pixel with the minimal pixelvalue */
{
  float *pix;
  int i,j,nex;

  nex = hd->orows - hd->rows;
  pix = (float *) hd->firstpix;
  *min = *pix;
  *x=0;
  *y=0;
   
  for (i=0; i<hd->rows; i++) {
    for (j=0; j<hd->cols; j++) {
      if(*pix < *min) {
	*x=i;
	*y=j;
	*min = *pix;
      }
      pix++;
    }
    pix += nex;
  }
  
  return (HIPS_OK);
}



