/*---------------------------------------------------------------------------
## \section{dec2seq}
## 
## \subsection*{Purpose}
##	Convert the output from "qmf -d ..." (which is a mosaic of critically
##	decimated subbands) into a sequence, where each subband is one frame.
##	The frames will be the size of the least decimated subband, and
##	more subsampled frames will have pixels replicated to fill the frame.
## 
## \subsection*{Usage }
## \begin{verbatim}
## dec2seq [-fs] 
##   [-sn struct<integer> | 
##    -s struct<string> | 
##    -sb struct<string>]
##   [imagefile]
## \end{verbatim}
## \begin{description}
## \item[-s]
## 	Filter structure string, defined as a quad tree.
## 	See the module {\em filtertree} (CompileFiltStructTree)
## 	for a description of the structure.
## \item[-sb]
## 	Filterstructure string, defined as a binary tree.
## 	See the module {\em filtertree} (CompileFiltStructTreeFromBin)
## 	for a description of the structure.
## \item[-sn]
## 	Filter structure number, between 1 and 11 (default 4).
## 	See the module {\em filtertree} (BuildFiltStructTree)
## 	for a description of the structure.
## \item[-fs]
##	Full size output (i.e. don't subsample with the least common
##	subsampling factor).
## \end{description}
##
## \subsection*{Data formats handled directly}
## \begin{itemize}
##    \item	byte
##    \item	short
##    \item	int
##    \item	float
## \end{itemize}
## 
## \subsection*{Date created}
## 6/3-94
+---------------------------------------------------------------------------+
|	Trygve Randen                                                       |
|	Rogaland University Center, Dept. of Elect. and Computer Eng.       |
|	P.O. Box 2557 Ullandhaug, 4004 Stavanger, Norway                    |
|	Tel.: +47 51 87 43 36, Fax.: +47 51 87 42 36                        |
|	E-mail: tranden@hsr.no                                              |
+--------------------------------------------------------------------------*/



/* Header files */
#include "qmf.h"
#include "mcf.h"

static Flag_Format flagfmt[]={
  {"sn",{"s","sb",LASTFLAG},1,{
    {PTBOOLEAN,"TRUE"},
    {PTINT,"4","struct"},
    LASTPARAMETER}},
  {"sb",{"sn","s",LASTFLAG},1,{
    {PTBOOLEAN,"FALSE"},
    {PTSTRING,"(,)","struct"},
    LASTPARAMETER}},
  {"s",{"sn","sb",LASTFLAG},1,{
    {PTBOOLEAN,"FALSE"},
    {PTSTRING,"(,)","struct"},
    LASTPARAMETER}},
  {"fs",{LASTFLAG},0,{
    {PTBOOLEAN,"FALSE"},
    LASTPARAMETER}},
  LASTFLAG};

int types[]={PFBYTE,PFSHORT,PFINT,PFFLOAT,LASTTYPE};

main(argc,argv)
     int argc;
     char *argv[];
{
  struct header hd;		/* Input file header (any type) */
  struct header hdp;		/* Input file header (float) */
  struct header hdo;		/* Output file header (any type) */
  int method;			/* Conversion methods */
  Boolean structStr1DFlag;	/* Structure string specified (bin) */
  Boolean structStrFlag;	/* Structure string specified (quad) */
  Boolean structNoFlag;		/* Structure number specified */
  Boolean fsFlag;		/* Full size output flag, -fs option */
  int structNo;
  char *structStr1D;		/* Structure description string (bin) */
  char *structStr;		/* Structure description string (quad) */
  FilterTree filtertree;	/* Filter structure tree */
  Filename filename;
  FILE *fp;
  int leastCommonDecimation;
  int frame;

  /* Initialize variables */
  Progname=strsave(*argv);
  parseargs(argc,argv,flagfmt,&structNoFlag,&structNo,
	    &structStr1DFlag,&structStr1D,&structStrFlag,&structStr,
	    &fsFlag,
	    FFONE,&filename);

  /* Build the filter structure binary tree */
  if (structStrFlag) 
    filtertree = CompileFiltStructTree(structStr,"");
  else if (structStr1DFlag) 
    filtertree = CompileFiltStructTreeFromBin(structStr1D,"");
  else if (structNoFlag)
    filtertree = BuildFiltStructTree(structNo,"");
  leastCommonDecimation=pow(2.0,(float) getUpperSubLevel (filtertree)+1);

  /* Read the input image */
  fp = hfopenr(filename);
  fread_hdr_a(fp,&hd,filename);
  method = fset_conversion(&hd,&hdp,types,filename);
  fread_imagec(fp,&hd,&hdp,method,0,filename);	

  /* Write the output header */
  dup_headern(&hd,&hdo);
  hdo.num_frame=getNSubbands(filtertree);
  if (!fsFlag)
    setsize(&hdo,
	    hd.orows/leastCommonDecimation,
	    hd.ocols/leastCommonDecimation);
  setformat(&hdo,hdp.pixel_format);
  alloc_image(&hdo);
  write_headeru(&hdo,argc,argv);

  /* Extract and write the subbands */
  extr_sub (&hdo,&hdp,filtertree,0,0,hd.orows-1,hd.ocols-1,&frame);
}




extr_sub (hdo,hdp,ft,fr,fc,tr,tc,frame)
     struct header *hdo, *hdp;
     FilterTree ft;
     int fr,fc,tr,tc;
     int *frame;
{
  int ro, kappa;

  if (ft != NULL) 
    for (ro=0; ro<2; ro++) 
      for (kappa=0; kappa<2; kappa++) 
        extr_sub (hdo, hdp, 
		  ft->subband[ro][kappa],
		  (ro==0 ? fr : (fr+tr+1)/2),
		  (kappa==0 ? fc : (fc+tc+1)/2),
		  (ro==1 ? tr : (fr+tr-1)/2),
		  (kappa==1 ? tc : (fc+tc-1)/2),
		  frame);
  else {
    if (hdo->pixel_format == PFBYTE)
      extr_byte (hdo,hdp,fr,fc,tr,tc,frame);
    else if (hdo->pixel_format == PFSHORT)
      extr_short (hdo,hdp,fr,fc,tr,tc,frame);
    else if (hdo->pixel_format == PFINT)
      extr_int (hdo,hdp,fr,fc,tr,tc,frame);
    else if (hdo->pixel_format == PFFLOAT)
      extr_float (hdo,hdp,fr,fc,tr,tc,frame);
  }
}
  


extr_byte (hdo,hdp,fr,fc,tr,tc,frame)
     struct header *hdo, *hdp;
     int fr,fc,tr,tc;
     int *frame;
{
  int i,j,k,l;
  byte **x, **y;
  int enlargement;

  /* Memory allocation etc. */
  x=(byte **) matrix2(hdp);
  y=(byte **) matrix2(hdo);
  
  /* Transfer pixels */
  enlargement=hdo->orows/(tr-fr+1);
  for (i=fr; i<=tr; i++) {
    for (j=fc; j<=tc; j++) {
      for (k=0; k<enlargement; k++) {
	for (l=0; l<enlargement; l++) {
	  y[(i-fr)*enlargement+k][(j-fc)*enlargement+l] = x[i][j];
	  }
      }
    }
  }
  
  /* Write frame */
  write_image(hdo,(*frame)++);
  
  /* Release allocated memory */
  free_matrix2(y);
  free_matrix2(x);
}



extr_short (hdo,hdp,fr,fc,tr,tc,frame)
     struct header *hdo, *hdp;
     int fr,fc,tr,tc;
     int *frame;
{
  int i,j,k,l;
  short **x, **y;
  int enlargement;

  /* Memory allocation etc. */
  x=(short **) matrix2(hdp);
  y=(short **) matrix2(hdo);
  
  /* Transfer pixels */
  enlargement=hdo->orows/(tr-fr+1);
  for (i=fr; i<=tr; i++) {
    for (j=fc; j<=tc; j++) {
      for (k=0; k<enlargement; k++) {
	for (l=0; l<enlargement; l++) {
	  y[(i-fr)*enlargement+k][(j-fc)*enlargement+l] = x[i][j];
	  }
      }
    }
  }
  
  /* Write frame */
  write_image(hdo,(*frame)++);
  
  /* Release allocated memory */
  free_matrix2(y);
  free_matrix2(x);
}



extr_int (hdo,hdp,fr,fc,tr,tc,frame)
     struct header *hdo, *hdp;
     int fr,fc,tr,tc;
     int *frame;
{
  int i,j,k,l;
  int **x, **y;
  int enlargement;

  /* Memory allocation etc. */
  x=(int **) matrix2(hdp);
  y=(int **) matrix2(hdo);
  
  /* Transfer pixels */
  enlargement=hdo->orows/(tr-fr+1);
  for (i=fr; i<=tr; i++) {
    for (j=fc; j<=tc; j++) {
      for (k=0; k<enlargement; k++) {
	for (l=0; l<enlargement; l++) {
	  y[(i-fr)*enlargement+k][(j-fc)*enlargement+l] = x[i][j];
	  }
      }
    }
  }
  
  /* Write frame */
  write_image(hdo,(*frame)++);
  
  /* Release allocated memory */
  free_matrix2(y);
  free_matrix2(x);
}



extr_float (hdo,hdp,fr,fc,tr,tc,frame)
     struct header *hdo, *hdp;
     int fr,fc,tr,tc;
     int *frame;
{
  int i,j,k,l;
  float **x, **y;
  int enlargement;

  /* Memory allocation etc. */
  x=(float **) matrix2(hdp);
  y=(float **) matrix2(hdo);
  
  /* Transfer pixels */
  enlargement=hdo->orows/(tr-fr+1);
  for (i=fr; i<=tr; i++) {
    for (j=fc; j<=tc; j++) {
      for (k=0; k<enlargement; k++) {
	for (l=0; l<enlargement; l++) {
	  y[(i-fr)*enlargement+k][(j-fc)*enlargement+l] = x[i][j];
	  }
      }
    }
  }
  
  /* Write frame */
  write_image(hdo,(*frame)++);
  
  /* Release allocated memory */
  free_matrix2(y);
  free_matrix2(x);
}

