/*----------------------- Rogaland University Centre -------------------------
## \section{dec\_synthesis}
## 
## 
## \subsection*{Purpose}
## Synthesis filterbank, to be appled to a image filtered
## by the decimated analysis filterbank {\tt dec_analysis}
## or a critically decimated image filtered by the filterbank
## {\tt analysis}.
## 
## \subsection*{Description}
## \subsubsection*{dec\_synthesis\_bank}
## 	Apply a decimated synthesis filterbank to the input
## 	array/image and write the result to the standard output
##  
## \subsection*{Usage}
## \subsubsection*{dec\_synthesis\_bank}
## \begin{verbatim}
## dec_synthesis_bank(hd,y,ft,frame_no)
##      struct header *hd;
##      float **y;        
##      FilterTree ft;
##      int frame_no;
## \end{verbatim}
## \begin{itemize}
##   \item      Input image header, floating point image
##   \item      Array-reference to the same image with the pixels accessible
##         	as an array: y[m][n]. See the module matrix2.
##   \item      Filterstructure tree see module {\em filtertree}.
##   \item      Frame number of input/output frame
## \end{itemize}
## 
## 
## 
## \subsection*{Date created}
## March 1993
 *------------------------ Rogaland University Centre -----------------------*/



/* Header files */
#include "qmf.h"


dec_synthesis_bank(hd,y,ft,frame_no)
     struct header *hd;		/* Input image */
     float **y;			/* Array-reference to the image */
     FilterTree ft;
     int frame_no;		/* Input image frame number */
{
  /* ro and kappa are exclusively used as counters denoting
     the polyphase branch, wheras m,n denote vertical and
     horizontal index, respectively. */
  int ro,kappa,i,j,k,l,m,n;
  struct header hd_c;		/* Header in output format (byte) */
  int method_c;			/* Output conversion method */
  int over_flows, under_flows;
  
  /* Set the conversion method and frame for converting to byte output */
  method_c=find_method(PFBYTE,PFFLOAT);
  dup_header(hd,&hd_c);
  setformat(&hd_c,PFBYTE);

  /* Filter the image recursively */
  synthesis_2x2bnd (hd,y,ft,0,0,hd->rows/2,hd->cols/2,getNSubLevels(ft),
		FALSE,FALSE);

  /* The output picture is stored as a byte image */
  under_flows=0;
  over_flows=0;
  for (m=0;m<hd->rows;m++) {
    for (k=0;k<hd->cols;k++) {
      if ( y[m][k] >127.0 ) {
	y[m][k]=127.0;
	over_flows++;
      }
      if ( y[m][k] < -128.0 ) {
	y[m][k]= -128.0;
	under_flows++;
      }
      y[m][k]=floor(y[m][k]+128.0+0.5);
    }
  }
  write_imagec(&hd_c,hd,method_c,TRUE,frame_no);
  fprintf(stderr,"\noverflows: %d   underflows: %d\n",
	  over_flows, under_flows);
}




synthesis_2x2bnd(hd, y, ft, x_orig, y_orig, x_dim, y_dim, TrLevelDiff,
	     x_reorg, y_reorg)
     struct header *hd;
     float **y;
     FilterTree ft;
     int x_orig, y_orig;
     int x_dim, y_dim;
     int TrLevelDiff;
     Boolean x_reorg, y_reorg;
{
  /* ro and kappa are exclusively used as counters denoting
     the subband branch, wheras m,n denote vertical and
     horizontal index, respectively. */
  struct fir_filterspec filt_par;
  int ro,kappa;
  int ro_out,kappa_out;
  int i,j,k,l,m,n;
  int v_index, h_index;
  float **x[2][2];
  float tmp[2][2];
  float *line_in, *line_out;
  int shift;

  /* Filtering subbands recursively */
  for (ro=0; ro<2; ro++) {
    for (kappa=0; kappa<2; kappa++) {
      if (ft->subband[ro][kappa] != NULL) {
	synthesis_2x2bnd (hd,y,ft->subband[ro][kappa],
			  x_orig + ro*x_dim, y_orig + kappa*y_dim,
			  x_dim/2, y_dim/2, TrLevelDiff-1, 
			  ro!=0,kappa!=0);
      }
      else { 
	/* Scale subbands properly */
	setroi(hd, x_orig + ro*x_dim, y_orig + kappa*y_dim, x_dim, y_dim);
	h_linscale(hd,hd,1.0 / ((float) pow(2.0,(float) TrLevelDiff)),0.0);
	clearroi(hd);
      }
    }
  }

  /* Allocating space for 4 nondecimated images */
  for (ro=0;ro<2;ro++)
    for (kappa=0;kappa<2;kappa++)
      x[ro][kappa]=matrix(0,2*x_dim-1,0,2*y_dim-1);

  /* Organize subbands */
  for (ro=0;ro<2;ro++) {
    ro_out=x_reorg ? 1-ro : ro;
    for (kappa=0;kappa<2;kappa++) {
      kappa_out=y_reorg ? 1-kappa : kappa;
      for (m=0;m<x_dim;m++) {
	for (n=0;n<y_dim;n++) {
	  x[ro][kappa][2*m+1][2*n+0] = 0.0;
	  x[ro][kappa][2*m+0][2*n+1] = 0.0;
	  x[ro][kappa][2*m+1][2*n+1] = 0.0;
	  x[ro][kappa][2*m+0][2*n+0] = 
	    y[ro_out*x_dim+m+x_orig][kappa_out*y_dim+ n+y_orig];
	}
      }
    }
  }

  /* Inverse FFT if using an IIR (polyphase) filter */
  if (tolower(ft->filt_id[0]) == 'i') {
    for (m=0;m<2*x_dim;m++) {
      for (n=0;n<2*y_dim;n++) {
	tmp[0][0] = x[0][0][m][n];
	tmp[0][1] = x[0][1][m][n];
	tmp[1][0] = x[1][0][m][n];
	tmp[1][1] = x[1][1][m][n];

	x[0][0][m][n] = (tmp[0][0] + tmp[0][1] + tmp[1][0] + tmp[1][1])/4;
	x[0][1][m][n] = (tmp[0][0] - tmp[0][1] + tmp[1][0] - tmp[1][1])/4;
	x[1][0][m][n] = (tmp[0][0] + tmp[0][1] - tmp[1][0] - tmp[1][1])/4;
	x[1][1][m][n] = (tmp[0][0] - tmp[0][1] - tmp[1][0] + tmp[1][1])/4;
      }
    }
  }

  /* Inverse filtering */
  for (ro=0;ro<2;ro++) 
    for (kappa=0;kappa<2;kappa++) 
      filter_2x2bnd(ft->filt_id,x[ro][kappa],2*x_dim,2*y_dim,ro,kappa,
                    1,FALSE);

  /* Putting it all together */
  switch (tolower(ft->filt_id[0])) 
    {
    case 'i':	/* J. H. Husoy IIR filter, polyphase realization */
      for (ro=0;ro<2;ro++) 
	for (kappa=0;kappa<2;kappa++) 
	  for (m=0;m<x_dim;m++) 
	    for (n=0;n<y_dim;n++) 
	      y[2*m+ro+x_orig][2*n+kappa+y_orig]=
		x[ro][kappa][2*m][2*n];
      break;
    case 'f':	/* Johnston FIR filters, non-polyphase */
      for (m=0;m<2*x_dim;m++) 
	for (n=0;n<2*y_dim;n++) 
	  y[m+x_orig][n+y_orig]= 
	    + x[0][0][m][n] 
	    - x[0][1][m][n] 
	    - x[1][0][m][n] 
	    + x[1][1][m][n];
      break;
    case 'w':	/* Daubechies orthogonal wavelets */
      for (m=0;m<2*x_dim;m++) 
	for (n=0;n<2*y_dim;n++) 
	  y[m+x_orig][n+y_orig]= 
	    + x[0][0][m][n] 
	    + x[0][1][m][n] 
	    + x[1][0][m][n] 
	    + x[1][1][m][n];
      break;
    }

  /* Freeing space for 4 nondecimated images */
  for (ro=0;ro<2;ro++)
    for (kappa=0;kappa<2;kappa++)
      free_matrix(x[ro][kappa],0,2*x_dim-1,0,2*y_dim-1);
}
