/*----------------------- Rogaland University Centre -------------------------
## \section{decimate}
## 
## 
## \subsection*{Purpose}
## Decimate an image sequence in both vertical and horizontal direction.
## 
## \subsection*{Description}
## Read the input image and copy pixels on the decimation grid to 
## an image that is written to the standard out. 
## Only the region of interest is processed, the rest is left out.
## 
## This program may also perform ``simulated'' decimation. Simulated decimation
## means copying the pixelvalues that would have been kept in the decimation
## onto the pixelvalues that would have been removed.
## 
## 
## \subsection*{Usage}
## \begin{verbatim}
## decimate
##     [-N vertical decimation<integer> 
##      [horizontal decimation<integer>]]
##     [-sn struct<integer> | 
##      -s struct<string> | 
##      -sb struct<string>]
##     [imagefile]
## \end{verbatim}
## \begin{description}
## \item[-N]
## 	If only vertical decimation is specified, horizontal
## 	decimation is set equal to the vertical decimation.
## 	Default value is two.
## \item[-s]
## Filter structure string, defined as a quad tree.
## See the module {\em filtertree} (CompileFiltStructTree)
## for a description of the structure.
## \item[-sb]
## Filterstructure string, defined as a binary tree.
## See the module {\em filtertree} (CompileFiltStructTreeFromBin)
## for a description of the structure.
## \item[-sn]
## Filter structure number, between 1 and 11 (default 4).
## See the module {\em filtertree} (BuildFiltStructTree)
## for a description of the structure.
## \end{description}
## If any of the parameters {\tt -s}, {\tt -sb} or {\tt -sn} is used,
## then this program performs ``simulated'' decimation on the image
## corresponding to the critical sampling of a filter with the specified
## filterstructure, and the simulated decimatedimage is then decimated
## by the factor defined by {\tt -N}.
## 
## 
## 
## 
## \subsection*{Date created}
## March 1993
 *------------------------ Rogaland University Centre -----------------------*/




#include <hipl_format.h>
#include "qmf.h"

static Flag_Format flagfmt[]={
  {"N",{LASTFLAG},1,{
    {PTINT,"2","vertical decimation"},
    {PTINT,"-1","horizontal decimation"},
    LASTPARAMETER}},
  {"sn",{"s","sb",LASTFLAG},1,{
    {PTBOOLEAN,"FALSE"},
    {PTINT,"4","struct"},
    LASTPARAMETER}},
  {"sb",{"sn","s",LASTFLAG},1,{
    {PTBOOLEAN,"FALSE"},
    {PTSTRING,"(,)","struct"},
    LASTPARAMETER}},
  {"s",{"sn","sb",LASTFLAG},1,{
    {PTBOOLEAN,"FALSE"},
    {PTSTRING,"(,)","struct"},
    LASTPARAMETER}},
  LASTFLAG};

int types[]={PFFLOAT,LASTTYPE};



main(argc,argv)
     int argc;
     char **argv;
{
  Boolean structStr1DFlag;	/* Structure string specified (bin) */
  Boolean structStrFlag;	/* Structure string specified (quad) */
  Boolean structNoFlag;		/* Structure string specified (quad) */
  int struct_no;
  char *structStr1D;		/* Structure description string (bin) */
  char *structStr;		/* Structure description string (quad) */
  FilterTree filtertree;	/* Filter structure tree */
  struct header hd, hdi, hdo;
  int method,f,nex,*p,i,j,coord;
  float **x,**y;
  int vertN,horN;
  Filename filename;
  FILE *fp;

  /* Read the parameters */
  Progname = strsave(*argv);
  parseargs(argc,argv,flagfmt,&vertN,&horN,&structNoFlag,&struct_no,
            &structStr1DFlag,&structStr1D,&structStrFlag,&structStr,
	    FFONE,&filename);
  if (horN<0)
    horN = vertN;

  /* Build the filter structure binary tree */
  if (structStrFlag) 
    filtertree = CompileFiltStructTree(structStr,"");
  else if (structStr1DFlag) 
    filtertree = CompileFiltStructTreeFromBin(structStr1D,"");
  else if (structNoFlag)
    filtertree = BuildFiltStructTree(struct_no,"");

  if (structStrFlag || structStr1DFlag || structNoFlag) {
    PrintFiltStructTree (filtertree);
  }

  /* Read the input image header */
  fp = hfopenr(filename);
  fread_hdr_a(fp,&hd,filename);
  method = fset_conversion(&hd,&hdi,types,filename);

  /* If any filter structure is specified, this program will SIMULATE the
     corresponding critical subsampling. Simulation means that the output
     is not actually subsampled (because diffrent levels in the filtertree
     would mean different decimation factors), rather pixels that would 
     normally be thrown away is given the last pixel value (up to the left) */

  /* Create the output image */
  init_hdr_alloc(&hdo,"","",hdi.num_frame,"",
		 hdi.rows/vertN, hdi.cols/horN,
		 hdi.pixel_format,1,"");
  write_headeru(&hdo,argc,argv);

  if (structStrFlag || structStr1DFlag || structNoFlag) {
    f=0;
    recursive_decimate (filtertree,vertN,horN,2,&f,&hd,&hdi,&hdo,
			method,&fp,filename);
  }
  else {
    for (f=0;f<hdi.num_frame;f++) {
      decimate_frame (vertN,horN,1,f,&hd,&hdi,&hdo,method,&fp,filename);
    }
  }
  return(0);
}



recursive_decimate (ft, vertN, horN, simN, frame, hd, hdi, hdo, 
		    method, fp, filename)
     FilterTree ft;		/* Filter structure tree */
     int vertN, horN;		/* Decimation factor */
     int simN;			/* Simulated decimation factor */
     int *frame;		/* Frame number */
     struct header *hd;		/* Header of data in original format */
     struct header *hdi;	/* Header of input */
     struct header *hdo;	/* Header of output */
     int method;		/* Conversion method */
     FILE **fp;			/* Input file pointer */
     Filename filename;		/* Input file name */
{
  int ro, kappa;

  for (ro=0; ro<2; ro++) {
    for (kappa=0; kappa<2; kappa++) {
      if (ft->subband[ro][kappa] != NULL) {
        recursive_decimate (ft->subband[ro][kappa],vertN,horN,2*simN,
			    frame,hd,hdi,hdo, method, fp, filename);
      }
      else {
	decimate_frame (vertN,horN,simN,*frame,hd,hdi,hdo,method,fp,filename);
	(*frame)++;
      }
    }
  }
}
  



decimate_frame (vertN, horN, simN, frame, hd, hdi, hdo, method, fp, filename)
     int vertN, horN;		/* Decimation factor */
     int simN;			/* Simulated decimation factor */
     int frame;			/* Frame number */
     struct header *hd;		/* Header of data in original format */
     struct header *hdi;	/* Header of input */
     struct header *hdo;	/* Header of output */
     int method;		/* Conversion method */
     FILE **fp;			/* Input file pointer */
     Filename filename;		/* Input file name */
{
  float **x, **y;
  float mean;
  int i, j, k, l;

  /* Read the input image frame */
  fread_imagec(*fp,hd,hdi,method,frame,filename);
  y = (float **) matrix2(hdo);
  x = (float **) matrix2(hdi);

  /* Perform the simulated decimation */
  for(i=0; i<hdi->rows;i+=simN) 
    for(j=0;j<hdi->cols;j+=simN) 
      for(k=0;k<simN;k++) 
	for(l=0;l<simN;l++) 
	  x[i+k][j+l] = x[i][j];

  /* Perform the actual decimation */
  for(i=0; i<hdo->rows;i++) 
    for(j=0;j<hdo->cols;j++) 
      y[i][j] = x[i*vertN][j*horN];

  /* Complete the processing */
  free_matrix2(x);
  free_matrix2(y);
  write_image(hdo,frame);
}
