/*---------------------------------------------------------------------------
## \section{dscale}
## 
## \subsection*{Purpose}
##	Scale and convert an image from "qmf -d ..." to byte, giving
##	each subband maximum range.
## 
## \subsection*{Usage }
## \begin{verbatim}
## dscale 
##   [-sn struct<integer> | 
##    -s struct<string> | 
##    -sb struct<string>]
##   [imagefile]
## \end{verbatim}
## \begin{description}
##   \item[-s]
## 	Filter structure string, defined as a quad tree.
## 	See the module {\em filtertree} (CompileFiltStructTree)
## 	for a description of the structure.
##   \item[-sb]
## 	Filterstructure string, defined as a binary tree.
## 	See the module {\em filtertree} (CompileFiltStructTreeFromBin)
## 	for a description of the structure.
##   \item[-sn]
## 	Filter structure number, between 1 and 11 (default 4).
## 	See the module {\em filtertree} (BuildFiltStructTree)
## 	for a description of the structure.
##   \end{description}
##
## \subsection*{Date created}
## 2/3-94
+---------------------------------------------------------------------------+
|	Trygve Randen                                                       |
|	Rogaland University Center, Dept. of Elect. and Computer Eng.       |
|	P.O. Box 2557 Ullandhaug, 4004 Stavanger, Norway                    |
|	Tel.: +47 51 87 43 36, Fax.: +47 51 87 42 36                        |
|	E-mail: tranden@hsr.no                                              |
+--------------------------------------------------------------------------*/



/* Header files */
#include "qmf.h"
#include "mcf.h"

static Flag_Format flagfmt[]={
  {"sn",{"s","sb",LASTFLAG},1,{
    {PTBOOLEAN,"TRUE"},
    {PTINT,"4","struct"},
    LASTPARAMETER}},
  {"sb",{"sn","s",LASTFLAG},1,{
    {PTBOOLEAN,"FALSE"},
    {PTSTRING,"(,)","struct"},
    LASTPARAMETER}},
  {"s",{"sn","sb",LASTFLAG},1,{
    {PTBOOLEAN,"FALSE"},
    {PTSTRING,"(,)","struct"},
    LASTPARAMETER}},
  LASTFLAG};

int types[]={PFFLOAT,LASTTYPE};

main(argc,argv)
     int argc;
     char *argv[];
{
  struct header hd;		/* Input file header (any type) */
  struct header hdp;		/* Input file header (float) */
  struct header hdo;		/* Output file header (byte) */
  int method, method2;		/* Conversion methods */
  Boolean structStr1DFlag;	/* Structure string specified (bin) */
  Boolean structStrFlag;	/* Structure string specified (quad) */
  Boolean structNoFlag;		/* Structure number specified */
  int structNo;
  char *structStr1D;		/* Structure description string (bin) */
  char *structStr;		/* Structure description string (quad) */
  FilterTree filtertree;	/* Filter structure tree */
  Filename filename;
  FILE *fp;

  hipserrprt = HEL_ERROR;	/* Suppress messaging from HIPS routines */

  /* Initialize variables */
  Progname=strsave(*argv);
  parseargs(argc,argv,flagfmt,&structNoFlag,&structNo,
	    &structStr1DFlag,&structStr1D,&structStrFlag,&structStr,
	    FFONE,&filename);

  /* Build the filter structure binary tree */
  if (structStrFlag) 
    filtertree = CompileFiltStructTree(structStr,"");
  else if (structStr1DFlag) 
    filtertree = CompileFiltStructTreeFromBin(structStr1D,"");
  else if (structNoFlag)
    filtertree = BuildFiltStructTree(structNo,"");
 
  /* Read the input image */
  fp = hfopenr(filename);
  fread_hdr_a(fp,&hd,filename);
  method = set_conversion(&hd,&hdp,types);
  fread_imagec(fp,&hd,&hdp,method,0,filename);	

  /* Scale the subbands individually */
  scale_subbands(&hdp,filtertree,0,0,hdp.orows-1,hdp.ocols-1);

  /* Write the output image */
  dup_headern(&hdp,&hdo);
  setformat(&hdo,PFBYTE);
  alloc_image(&hdo);
  h_tob(&hdp,&hdo);
  write_headeru(&hdo,argc,argv);
  write_image(&hdo,TRUE,0);
}




scale_subbands (hd,ft,fr,fc,tr,tc)
     struct header *hd;
     FilterTree ft;
     int fr,fc,tr,tc;
{
  int ro, kappa;

  if (ft != NULL) 
    for (ro=0; ro<2; ro++) 
      for (kappa=0; kappa<2; kappa++) 
        scale_subbands (hd, ft->subband[ro][kappa],
			(ro==0 ? fr : (fr+tr+1)/2),
			(kappa==0 ? fc : (fc+tc+1)/2),
			(ro==1 ? tr : (fr+tr-1)/2),
			(kappa==1 ? tc : (fc+tc-1)/2));
  else {
    setroi(hd,fr,fc,tr-fr+1,tc-fc+1);
    max_range(hd,hd,PFBYTE);
    setroi(hd,0,0,hd->orows,hd->ocols);
  }
}
  
