/*----------------------- Rogaland University Centre -------------------------
## \section{filter\_2x2bnd}
## 
## 
## \subsection*{Purpose}
## General filtering by QMF (Quadrature Mirror Filters)
## 
## \subsection*{Description}
## \subsubsection*{filter\_2x2bnd}
## 	Filter the input image array by a filter given by a
## 	filter identifier
## 
## \subsection*{Usage}
## \subsubsection*{filter\_2x2bnd}
## \begin{verbatim}
## filter_2x2bnd(filt_id,img,x_dim,y_dim,
##      ro,kappa,N,analysis)
##      char *filt_id;   
##      float **img;     
##      int x_dim, y_dim;
##      int ro, kappa; 
##      int N;         
##      int analysis;
## \end{verbatim}
## \begin{itemize}
##   \item Filter-identifier, a character string like i.e. `f16b', 
##         `i\_2\_1\_smpl' etc.
##         Note that a filter identifier starting with `f' is assumed to
##         represent a FIR filter with the sum of the coeffisients 
## 		equal to 1.0, an identifier starting with `i'
##         is a polyphase-allpass IIR filter and an identifier starting
##         with `w' is a wavelet FIR filter.
##         See: John Haakon Husoey, PhD dissertation, 1990, Norwegian
##         Institute of Technology
##   \item Image array, accessible as a float array x[m][n]
##   \item x (m - vertical) dimension, number of pixels
##   \item y (n - horizontal) dimension, number of pixels
##   \item Vertical subband, ro, 0 or 1; 0 is lowpass
##   \item Horizontal subband, kappa, 0 or 1
##   \item N: Filter by $F(z^N)$.
##   \item A boolean flag indicating wheter this is an analysis bank
##         (TRUE) or a synthesis bank (FALSE)
## \end{itemize}
## 
## \subsection*{Date created}
## March 1993
 *------------------------ Rogaland University Centre -----------------------*/



/* Header files */
#include "qmf.h"



filter_2x2bnd(filt_id,img,x_dim,y_dim,ro,kappa,N,analysis)
     char *filt_id;		/* Filter identificator */
     float **img;		/* Image to be filtered */
     int x_dim, y_dim;		/* Region of interest size */
     int ro, kappa;		/* Subband number in x and y dir. resp. */
     int N;			/* Filter by F(z^N) */
     int analysis;		/* TRUE if analysis bank */
     /* Perform 2-band filtering in each dimension */
{
  float tmp[x_dim][y_dim];
  float *line_in, *line_out;
  int m, n, i, j;
  int v_index,h_index;
  int x_shift, y_shift;			/* Time-shift after filtering */
  struct fir_filterspec prototyp_fir;
  struct fir_filterspec fir_filter;
  struct filterspec prototyp_iir;
  struct filterspec iir_filter;
  Boolean FIR;
  float tmp00,tmp01,tmp10,tmp11;
  float filt_coeff_sum;
  float swap;

  /* Print a dot as a verification that the program is active */
  fprintf(stderr,".");

  /* Identify wheter the filter is FIR or IIR and read the filtertaps, and
     build the filter tree based on the current filter tree level */
  switch (tolower (filt_id[0])) 
    {
    case 'f' :	/* Johnston FIR filter */
      read_fir (&prototyp_fir,filt_id);
      strcpy(fir_filter.filterfile, prototyp_fir.filterfile);
      fir_filter.no_of_taps = prototyp_fir.no_of_taps * N;
      uporder(prototyp_fir.p0,fir_filter.p0,prototyp_fir.no_of_taps,N);
      uporder(prototyp_fir.p1,fir_filter.p1,prototyp_fir.no_of_taps,N);
      x_shift = N*(prototyp_fir.no_of_taps/2) - (analysis ? 0 : 1); 
      y_shift = x_shift;
      fir_2x2 (&fir_filter,img,x_dim,y_dim,ro,kappa);
      break;
    case 'i' :	/* J. H. Husoy IIR filter */
      read_iir (&prototyp_iir,filt_id);
      if (prototyp_iir.no_of_branches != 2) {
	fprintf(stderr,"Only binary tree filters are supported\n");
	exit (1);
      }

      /* Construct actual filterbank */
      strcpy(iir_filter.filterfile, prototyp_iir.filterfile);
      iir_filter.no_of_branches =  prototyp_iir.no_of_branches;
      for (i=0; i< prototyp_iir.no_of_branches; i++) {
	iir_filter.filter_order[i] = prototyp_iir.filter_order[i] * 2*N;
	uporder(prototyp_iir.a[i],iir_filter.a[i],
		prototyp_iir.filter_order[i],2*N);
      }
      if (analysis) {
	iir_2x2 (&iir_filter,img,x_dim,y_dim,ro,kappa,N);
	x_shift = 0; 
	y_shift = 0;
      }
      else { /* Synthesis filter, the FFT must be performed in advance */
	allpass_2x2 (&iir_filter,img,x_dim,y_dim,ro,kappa,FALSE);
	x_shift = 0; 
	y_shift = 0;
      }
      break;
    case 'w' :	/* Wavelet filter */
      /* Only Daubechies orthogonal filters are supported, presently */

      /* Read the filter file and process the filter coefficients */
      read_fir (&prototyp_fir,filt_id);
      filt_coeff_sum = 0.0;
      for (i=0; i<prototyp_fir.no_of_taps; i++)
	filt_coeff_sum += prototyp_fir.p0[i];
      for (i=0; i<prototyp_fir.no_of_taps; i++) {
	prototyp_fir.p0[i] *= 1.4142 / filt_coeff_sum ;
	prototyp_fir.p1[i] *= 1.4142 / filt_coeff_sum ;
      }
      if (!analysis) {
	for (i=0; i<prototyp_fir.no_of_taps/2; i++) {
	  swap = prototyp_fir.p0[i];
	  prototyp_fir.p0[i] = prototyp_fir.p0[prototyp_fir.no_of_taps-i-1];
	  prototyp_fir.p0[prototyp_fir.no_of_taps-i-1] = swap;
	  swap = prototyp_fir.p1[i];
	  prototyp_fir.p1[i] = prototyp_fir.p1[prototyp_fir.no_of_taps-i-1];
	  prototyp_fir.p1[prototyp_fir.no_of_taps-i-1] = swap;
	}
      }
      strcpy(fir_filter.filterfile, prototyp_fir.filterfile);
      fir_filter.no_of_taps = prototyp_fir.no_of_taps * N;
      uporder(prototyp_fir.p0,fir_filter.p0,prototyp_fir.no_of_taps,N);
      uporder(prototyp_fir.p1,fir_filter.p1,prototyp_fir.no_of_taps,N);
      x_shift = N*(prototyp_fir.no_of_taps/2) - (analysis ? 0 : 1); 
      y_shift = x_shift;

      /* Filter the image */
      fir_2x2 (&fir_filter,img,x_dim,y_dim,ro,kappa);

      /* Scale the filtered image apropriately */
      for (m=0; m<x_dim; m++) {
	for (n=0; n<y_dim; n++) {
	  img[m][n] = 0.5 * img[m][n];
	}
      }
      break;
    default :
      fprintf(stderr,"Filter id %s does not conform to standard\n",filt_id);
      exit (1);
    }

  /* Data rearragement (time shifting) */
  for (m=0; m<x_dim; m++) {
    v_index = m + x_shift;
    if (v_index >= x_dim) 
      v_index %= x_dim;
    for (n=0; n<y_dim; n++) {
      h_index = n + y_shift;
      if (h_index >= y_dim)
	h_index %= y_dim;

      tmp[m][n] = img[v_index][h_index];
    }
  }  
  for (m=0; m<x_dim; m++) {
    for (n=0; n<y_dim; n++) {
      img[m][n] = tmp[m][n];
    }
  }
}




uporder (x,y,len,N)
     float *x, *y;
     int len,N;
{
  int i;

  for (i=0; i<len*N; i++) 
    y[i]=0;
  for (i=0; i<len; i++) 
    y[i*N]=x[i];
}




fir_2x2 (filt_par,img,x_dim,y_dim,ro,kappa)
     struct fir_filterspec *filt_par;
     float **img;
     int x_dim, y_dim;
     int ro, kappa;
{
  int m, n;
  float *line_in =vector(-MAX_NTAPS/2, MAX(x_dim,y_dim)-1+MAX_NTAPS/2);
  float *line_out=vector(0           , MAX(x_dim,y_dim)-1);

  /* Filtering in 1st. direction. (i.e. y-direcion or horizontally) */
  for (m=0;m<x_dim;m++) {
    extr_subband_line(line_in,img,0,kappa,1,m, y_dim,0);
    firfilt_p(filt_par,kappa,line_in,line_out,y_dim);
    insert_subband_line(line_out,img,0,kappa,1,m, y_dim,0);
  }

  /* Filtering in 2nd. direction. (i.e. x-direcion or vertically) */
  for (n=0;n<y_dim;n++) {
    extr_subband_line(line_in,img,ro,kappa,0,n, x_dim,0);
    firfilt_p(filt_par,ro,line_in,line_out,x_dim);
    insert_subband_line(line_out,img,ro,kappa,0,n, x_dim,0);
  }

  /* Freeing space for filter lines */
  free_vector(line_in, -MAX_NTAPS/2, MAX(x_dim,y_dim)-1+MAX_NTAPS/2);
  free_vector(line_out,0           , MAX(x_dim,y_dim)-1);
}




iir_2x2 (filt_par,img,x_dim,y_dim,ro,kappa,N)
     struct filterspec *filt_par;
     float **img;
     int x_dim, y_dim;
     int ro, kappa;
     int N;
{
  int x_bnd, y_bnd;
  int m, n;
  int m_ndx, n_ndx;
  float **allpass = matrix2_f(x_dim,y_dim);
  float **out     = matrix2_f(x_dim,y_dim);
  int i, j;

  /* Zero the output */
  for (m=0; m<x_dim; m++)
    for (n=0; n<y_dim; n++)
      out[m][n]=0.0;

  for (x_bnd=0; x_bnd<2; x_bnd++) {
    for (y_bnd=0; y_bnd<2; y_bnd++) {

      /* Copy the input */
      for (m=0; m<x_dim; m++) 
	for (n=0; n<y_dim; n++) 
	  allpass[m][n] = img[m][n];
      
      /* Compute the polyphase components */
      allpass_2x2 (filt_par,allpass,x_dim,y_dim,x_bnd,y_bnd,TRUE);

      /* Add the polyphase component */
      for (m=0; m<x_dim; m++) {
	m_ndx = (m+N*x_bnd > x_dim ? m+N*x_bnd-x_dim : m+N*x_bnd);
	for (n=0; n<y_dim; n++) {
	  n_ndx = (n+N*y_bnd > y_dim ? n+N*y_bnd-y_dim : n+N*y_bnd);
	  if (   (!x_bnd &&  y_bnd && kappa)
	      || ( x_bnd && !y_bnd && ro)
	      || ( x_bnd &&  y_bnd && (ro ^ kappa)))
	    out[m][n] -= allpass[m_ndx][n_ndx];
	  else
	    out[m][n] += allpass[m_ndx][n_ndx];
	}
      }
    }
  }

  /* Copy the output */
  for (m=0; m<x_dim; m++)
    for (n=0; n<y_dim; n++)
      img[m][n] = out[m][n]/4;

  free_matrix2_f (x_dim,y_dim,allpass);
  free_matrix2_f (x_dim,y_dim,out);
}


allpass_2x2 (filt_par,img,x_dim,y_dim,ro,kappa,analysis)
     struct filterspec *filt_par;
     float **img;
     int x_dim, y_dim;
     int ro, kappa;
     Boolean analysis;
{
  int m, n;
  float *line_in =matrix1_f(MAX(x_dim,y_dim));
  float *line_out=matrix1_f(MAX(x_dim,y_dim));

  /* Filtering in 1st. direction. (i.e. y-direcion or horizontally) */
  for (m=0;m<x_dim;m++) {
    extr_subband_line(line_in,img,ro,kappa,1,m, y_dim,!analysis);
    allpass_c(filt_par,kappa,line_in,line_out,y_dim);
    insert_subband_line(line_out,img,ro,kappa,1,m, y_dim,!analysis);
  }

  /* Filtering in 2nd. direction. (i.e. x-direcion or vertically) */
  for (n=0;n<y_dim;n++) {
    extr_subband_line(line_in,img,ro,kappa,0,n, x_dim,!analysis);
    allpass_c(filt_par,ro,line_in,line_out,x_dim);
    insert_subband_line(line_out,img,ro,kappa,0,n, x_dim,!analysis);
  }

  /* Freeing space for filter lines */
  free(line_in);
  free(line_out);
}
