/*----------------------- Rogaland University Centre -------------------------
## \section{filtertree}
## 
## 
## \subsection*{Purpose}
## This library provides a general way of specifying
## the subband-split structure of a herarchical tree-structured 
## QMF filterbank. 
## 
## 
## \subsection*{Description}
## Each node in the type FilterTree represents one set of QMF
## filters, splitting the input in a LL, LH, HL and LL band.
## \subsubsection*{BuildFiltStructTree}
## 	Build a filter structure quad tree based on a filterstructure number
## \subsubsection*{CompileFiltStructTree}
## 	Convert a string of parentheses and commas to a filterstructure
## 	tree
## \subsubsection*{CompileFiltStructTreeFromBin}
## 	Convert a string of parentheses and commas to a filterstructure
## 	tree. The input string is representing a binary tree, and the 
## 	conversion binary to quad is done by setting 
## 	``((,),(,))''=``((,,,),(,,,),(,,,),(,,,))'', 
## 	``((,),)''=``((,,,),,,)'' and ``(,(,))''=``(,,,(,,,))''.
## \subsubsection*{getNSubbands}
## 	Get the number of subands below the current tree node
## \subsubsection*{getNSubLevels}
## 	Get the number of filterlevels below the current tree node
## 	as the input to CompileFiltStructTree based on a filtertree.
## 
## 
## \subsection*{Usage}
##
## \subsubsection*{BuildFiltStructTree}
## \begin{verbatim}
## FilterTree
## BuildFiltStructTree (struct_no,filter)
##      int struct_no;
##      char *filter;
## \end{verbatim}
## \begin{itemize}
##   \item      Structure number, integer between 1 and 11
##   \item	Filter-id string, filter to use on all nodes of the filtertree
## \end{itemize}
##
## \subsubsection*{CompileFiltStructTree}
## \begin{verbatim}
## FilterTree
## CompileFiltStructTree (structStr,filter)
##      char *structStr;
##      char *filter;
## \end{verbatim}
## \begin{itemize}
##   \item Structure description string.
##         Format:
## 	        [filter-id]([LL-band split],[LH-band split],
##                 [HL-band split],[HH-band split])
##         Examples:
##         \begin{verbatim}
##         ((,,,),,,)
##         f8a(,(,,,),f32d(,,,),)
##         \end{verbatim}
##   \item Filter-id string, filter to use on all nodes of the filtertree
##         where the structure description string doesn't specify a
##         different filter.
## \end{itemize}
##
## \subsubsection*{CompileFiltStructTreeFromBin}
## \begin{verbatim}
## FilterTree
## CompileFiltStructTreeFromBin (structStr,
##      filter)
##      char *structStr;
##      char *filter;
## \end{verbatim}
## \begin{itemize}
##   \item Structure description string.
##         Format:
## 	   [filter-id]([L-band split],[H-band split])
##         Examples:
##         \begin{verbatim}
##         ((,),)
##         ((,),((,),(,)))
##         ((,),)
##         \end{verbatim}
##   \item Filter-id string, filter to use on all nodes of the filtertree
##         where the structure description string doesn't specify a
##         different filter.
## \end{itemize}
##
## \subsubsection*{getNSubbands}
## \begin{verbatim}
## int
## getNSubbands (ft)
##      FilterTree ft;
## \end{verbatim}
## \begin{itemize}
##   \item Filter tree node
## \end{itemize}
##
## \subsubsection*{getNSubLevels}
## \begin{verbatim}
## getNSubLevels (ft)
##      FilterTree ft;
## \end{verbatim}
## \begin{itemize}
##   \item Filter tree node
## \end{itemize}
## 
## \subsection*{Date created}
## April 1993
 *------------------------ Rogaland University Centre -----------------------*/



/* Header files */
#include "qmf.h"


FilterTree
FiltStructTreeNode (filter)
     char *filter;
     /* Create a filter structure tree node and for each node */
{
  FilterTree ft;
  
  ft = (FilterTree) malloc(sizeof(struct FiltTreeStruct2D));
  ft->subband[0][0]=NULL;
  ft->subband[0][1]=NULL;
  ft->subband[1][0]=NULL;
  ft->subband[1][1]=NULL;
  strcpy(ft->filt_id,filter);

  return (ft);
}



FilterTree1D
FiltStructTreeNode1D (filter)
     char *filter;
     /* Create a filter structure tree node and for each node */
{
  FilterTree1D ft;
  
  ft = (FilterTree1D) malloc(sizeof(struct FiltTreeStruct1D));
  ft->subband[0]=NULL;
  ft->subband[1]=NULL;
  strcpy(ft->filt_id,filter);

  return (ft);
}



getNSubbands (ft)
     FilterTree ft;
{
  if (ft == NULL)
    return (0);
  else
    return (MAX(1,getNSubbands(ft->subband[0][0]))
	    +MAX(1,getNSubbands(ft->subband[0][1]))
	    +MAX(1,getNSubbands(ft->subband[1][0]))
	    +MAX(1,getNSubbands(ft->subband[1][1])));
}



getNSubLevels (ft)
     FilterTree ft;
{
  return (getLowerSubLevel (ft));
}



getLowerSubLevel (ft)
     FilterTree ft;
{
  int left=0;
  int right=0;
  
  if (ft==NULL)
    return (-1);
  else
    return (1+MAX(getNSubLevels(ft->subband[0][0]),
		  MAX(getNSubLevels(ft->subband[0][1]),
		      MAX(getNSubLevels(ft->subband[1][0]),
			  getNSubLevels(ft->subband[1][1])))));
}



getUpperSubLevel (ft)
     FilterTree ft;
{
  int left=0;
  int right=0;
  
  if (ft==NULL)
    return (-1);
  else
    return (1+MIN(getNSubLevels(ft->subband[0][0]),
		  MIN(getNSubLevels(ft->subband[0][1]),
		      MIN(getNSubLevels(ft->subband[1][0]),
			  getNSubLevels(ft->subband[1][1])))));
}



FilterTree
BinToQuadTree (fti)
     FilterTree1D fti;
{
  FilterTree fto;

  fto = FiltStructTreeNode (fti->filt_id);

  if (fti->subband[0] != NULL)
    fto->subband[0][0] = BinToQuadTree (fti->subband[0]);
  if (fti->subband[1] != NULL)
    fto->subband[1][1] = BinToQuadTree (fti->subband[1]);
  if (fti->subband[0] != NULL && fti->subband[1] != NULL) {
    fto->subband[0][1] = BinToQuadTree (fti->subband[1]);
    fto->subband[1][0] = BinToQuadTree (fti->subband[1]);
  }

  return (fto);
}


ReadFilterID (input,index,output,filter)
     char *input;
     int *index;
     char *output;
     char *filter;
{
  int start, end;

  start = *index;
  while (input[*index] != '(') {
    if (input[*index] == ')')
      syntax_error_fst(input,"missing `,'",*index);
    (*index)++;
    if (*index > strlen(input))
      syntax_error_fst (input,"missing `('",*index-1);
  }
  end = (*index)++;
  if (start==end)
    strcpy (output,filter);
  else {
    strncpy (output,&(input[start]),end-start);
    output[end-start] = '\0';
  }
}





syntax_error_fst (structStr,msg,index)
     char *structStr, *msg;
     int index;
{
  int i;

  fprintf(stderr,"Bad filter structure description string, %s\n", msg);
  fprintf(stderr,"\t%s\n",structStr);
  fprintf(stderr,"\t");
  for (i=0; i<index; i++)
    fprintf(stderr," ");
  fprintf(stderr,"^\n");
  exit (1);
}



FilterTree1D
compile_fst_1D (structStr,filter,i)
     char *structStr;
     char *filter;
     int *i;
     /* Recursive part of CompileFiltStructTree */
{
  char filter_id[MAX_FILT_ID];
  FilterTree1D node;
  int ro;

  /* Read filter ID */
  ReadFilterID (structStr,i,filter_id,filter);

  /* Create filter tree node */
  node = FiltStructTreeNode1D(filter_id);
  
  /* Recurse down through further levels */
  for (ro=0; ro<2; ro++) {
    if ((structStr [*i] != ',') && (structStr [*i] != ')')) {
      /* Further filterlevels */
      node->subband[ro] = compile_fst_1D (structStr,filter,i);
    }
    if (ro != 1) {
      if (structStr [(*i)++] != ',') {	/* Read delimiting comma */
	syntax_error_fst (structStr,"missing `,'",*i-1);
      }
    }
    else {
      if (structStr [(*i)++] != ')') {	/* Close up parentheses */
	syntax_error_fst (structStr,"missing `)'",*i-1);
      }
    }
  }

  return (node);
}



FilterTree
CompileFiltStructTreeFromBin (structStr,filter)
     char *structStr;
     char *filter;
     /* Compile the structStr binary structure desription string into a
	filter structure tree. If the structure string doesn't
	specify a filter at a specific level, then the filter ID
	filter is inserted into the tree 
	*/
{
  int i=0;
  FilterTree1D root;

  root = compile_fst_1D(structStr,filter,&i);
  if (i<strlen(structStr))
    syntax_error_fst (structStr,"string too long",i);
  return (BinToQuadTree(root));
}
     




FilterTree
compile_fst (structStr,filter,i)
     char *structStr;
     char *filter;
     int *i;
     /* Recursive part of CompileFiltStructTree */
{
  char filter_id[MAX_FILT_ID];
  FilterTree node;
  int ro, kappa;

  /* Read filter ID */
  ReadFilterID (structStr,i,filter_id,filter);

  /* Create filter tree node */
  node = FiltStructTreeNode(filter_id);
  
  /* Recurse down through further levels */
  for (ro=0; ro<2; ro++) {
    for (kappa=0; kappa<2; kappa++) {
      if ((structStr [*i] != ',') && (structStr [*i] != ')')) {
	/* Further filterlevels */
	node->subband[ro][kappa] = compile_fst (structStr,filter,i);
      }
      if (ro != 1 || kappa != 1) {
	if (structStr [(*i)++] != ',') {	/* Read delimiting comma */
	  syntax_error_fst (structStr,"missing `,'",*i-1);
	}
      }
      else {
	if (structStr [(*i)++] != ')') {	/* Close up parentheses */
	  syntax_error_fst (structStr,"missing `)'",*i-1);
	}
      }
    }
  }

  return (node);
}



FilterTree
CompileFiltStructTree (structStr,filter)
     char *structStr;
     char *filter;
     /* Compile the structStr quad-tree structure desription string into a
	filter structure tree. If the structure string doesn't
	specify a filter at a specific level, then the filter ID
	filter is inserted into the tree 
	*/
{
  int i=0;
  FilterTree root;

  root = compile_fst(structStr,filter,&i);
  if (i<strlen(structStr))
    syntax_error_fst (structStr,"string too long",i);
  return (root);
}
     

FilterTree
BuildFiltStructTree (struct_no,filter)
     int struct_no;
     char *filter;
     /* Build the filter structure tree corresponding to struct_no,
	with filter as the filter ID in each node */
{
  FilterTree1D root;

  switch (struct_no) 
    {
    case 1: 
      return CompileFiltStructTree("(,,,)",filter);
    case 2: 
      return CompileFiltStructTree("((,,,),,,)",filter);
    case 3: 
      return CompileFiltStructTree("((,,,),(,,,),(,,,),(,,,))",filter);
    case 4: 
      return CompileFiltStructTree("(((,,,),(,,,),(,,,),(,,,)),(,,,),(,,,),(,,,))",filter);
    case 5: 
      return CompileFiltStructTree("(((,,,),(,,,),(,,,),(,,,)),((,,,),(,,,),(,,,),(,,,)),((,,,),(,,,),(,,,),(,,,)),((,,,),(,,,),(,,,),(,,,)))",filter);
    case 6: 
      return CompileFiltStructTree("((((,,,),,,),,,),,,)",filter);
    case 7: 
      return CompileFiltStructTree("((((,,,),,,),(,,,),(,,,),(,,,)),(,,,),(,,,),(,,,))",filter);
    case 8: 
      return CompileFiltStructTree("((((,,,),(,,,),(,,,),(,,,)),(,,,),(,,,),(,,,)),(,,,),(,,,),(,,,))",filter);
    case 9: 
      return CompileFiltStructTree("(((((,,,),(,,,),(,,,),(,,,)),(,,,),(,,,),(,,,)),(,,,),(,,,),(,,,)),(,,,),(,,,),(,,,))",filter);
    case 10: 
      return CompileFiltStructTree("(((((((,,,),,,),,,),,,),,,),,,),,,)",filter);
    case 11: 
      return CompileFiltStructTree("((((,,,),((,,,),(,,,),(,,,),(,,,)),((,,,),(,,,),(,,,),(,,,)),((,,,),(,,,),(,,,),(,,,))),(,,,),(,,,),(,,,)),(,,,),(,,,),(,,,))",filter);
    case 12:
      return CompileFiltStructTreeFromBin ("((((,),(,)),((,),(,))),((,),(,)))",filter);
    case 13:
      return CompileFiltStructTreeFromBin ("(((((,),),(,)),((,),(,))),((,),(,)))",filter);
    default:
      perr(HE_RNG,"structure number");
    }
}


PrintFiltStructTree (ft)
     FilterTree ft;
{
  PrintFiltStructTree_rec (ft);
  fprintf(stderr,"\n");
}



PrintFiltStructTree_rec (ft)
     FilterTree ft;
{
  int ro, kappa;

  fprintf(stderr,"%s(",ft->filt_id);
  for (ro=0; ro<2; ro++) {
    for (kappa=0; kappa<2; kappa++) {
      if (ft->subband[ro][kappa] != NULL)
	PrintFiltStructTree_rec (ft->subband[ro][kappa]);
      if (ro != 1 || kappa != 1)
	fprintf(stderr,",");
      else
	fprintf(stderr,")");
    }
  }
}
