/*----------------------- Rogaland University Centre -------------------------
## \section{hips2lvq}
## 
## 
## \subsection*{Purpose}
## Convert HIPS image sequences to a LVQ (Leaning Vector Quantization)
## datafile.
## 
## \subsection*{Description}
## The HIPS images are stored frame by frame while the LVQ datafiles are stored
## vector by vector (the elements in the vectors are pixelvalues from
## corresponding pixel positions in the frames).
## This implies that the program must build the LVQ datafile column by column
## as the frames are read in.
## For large images, this means that it needs a lot of disk space temporarily.
##
## Alternately, the program can take decimated (as created by
## {\tt qmf -d $\cdots$}) files as input (by using either of the filter
## structure parameters -- {\tt -s -sn -sb}). 
## Then the first input frame is assumed to be in 
## decimated format, while the rest are assumed to be full rate frames.
## This operation is much faster than the full rate operation.
## 
## \subsection*{Usage}
## \begin{verbatim}
## hips2lvq [-d decimals<integer>] 
##    [-train training LVQ file<filename>]
##    [-out LVQ datafile<filename>]
##    [-p vector retain percentage<real>]
##    [-seed random number seed<integer>] 
##    [-m mask file<string>]
##    [-sn struct<integer> | 
##     -s struct<string> | 
##    -sb struct<string>]
##    [imagefile1 imagefile2...imagefileN]
## \end{verbatim}
## \begin{description}
##   \item[-d]
## 	Number of decimals in output, default 0.
##   \item[-train]
## 	Let the last column be the tag data.
## 	The training dataset will be stored in the specified file,
## 	(p percent of the samples).
##   \item[-p]
## 	Percent of vectors to be kept. 
## 	Used when the program creates training
## 	data to LVQ, when LVQ is supposed to be trained on a subset 
## 	of the pixels in the image.
##   \item[-seed]
## 	Random seed.
## 	A negative value implies use of system clock.
##   \item[-m]
## 	Mask image. Only pixels where the corresponding pixels in the mask
## 	image are nonzero will be exported to LVQ format.
##   \item[-s]
##      Filter structure string, defined as a quad tree.
##      See the module {\em filtertree} (CompileFiltStructTree)
##      for a description of the structure.
##   \item[-sb]
##      Filterstructure string, defined as a binary tree.
##      See the module {\em filtertree} (CompileFiltStructTreeFromBin)
##      for a description of the structure.
##   \item[-sn]
##      Filter structure number, between 1 and 11 (default 4).
##      See the module {\em filtertree} (BuildFiltStructTree)
##      for a description of the structure.
##   \item[Input files]
## 	Minimum one input file, there may be as many as wanted.
## \end{description}
## 
## 
## 
## 
## 
## \subsection*{Date created}
## January 1993
 *------------------------ Rogaland University Centre -----------------------*/



#include <stdlib.h>
#include <limits.h>
#include <float.h>
#include <hipl_format.h>
#include <math.h>
#include "mcf.h"
#include "qmf.h"
#include "error.h"
#include "math.h"

/* Conversion */
int types[] = {PFFLOAT,LASTTYPE};
int maskTypes[] = {PFINT,LASTTYPE};

/* Define the commandline-format */
static Flag_Format flagfmt[] = {
  {"d",{LASTFLAG},1,{
    {PTINT,"0","decimals"}, 
    LASTPARAMETER}},
  {"train",{LASTFLAG},1,{
    {PTBOOLEAN,"FALSE"},
    {PTFILENAME," ","training LVQ file"},
    LASTPARAMETER}},
  {"out",{LASTFLAG},1,{
    {PTFILENAME,"<stdout>","LVQ datafile"},
    LASTPARAMETER}},
  {"p",{LASTFLAG},1,{
    {PTDOUBLE,"100","vector retain percentage"}, 
    LASTPARAMETER}},
  {"seed",{LASTFLAG},1,{{PTINT,"0","random number seed"}, LASTPARAMETER}},
  {"m",{LASTFLAG},1,{
    {PTBOOLEAN,"FALSE"},
    {PTSTRING,"","mask file"},
    LASTPARAMETER}},
  {"sn",{"s","sb",LASTFLAG},1,{
    {PTBOOLEAN,"FALSE"},
    {PTINT,"4","struct"},
    LASTPARAMETER}},
  {"sb",{"sn","s",LASTFLAG},1,{
    {PTBOOLEAN,"FALSE"},
    {PTSTRING,"(,)","struct"},
    LASTPARAMETER}},
  {"s",{"sn","sb",LASTFLAG},1,{
    {PTBOOLEAN,"FALSE"},
    {PTSTRING,"(,)","struct"},
    LASTPARAMETER}},
  LASTFLAG};







main(argc,argv)
     int argc;
     char **argv;
{
  Filename *filelist, trainFile, outFile;
  FILE *fpi, *fpo;		/* fpi: inputfile, fpo: outputfile */
  int numFiles,file,frame,method,seed;
  Boolean train;		/* Same as parameter -train */
  double p;			/* Same as parameter -p */
  struct header hdMask;		/* Mask file header */
  struct header hdpMask;	/* Mask file header */
  int maskMethod;		/* Mask file conversion method */
  char *maskFile;		/* Mask file name */
  FILE *fpMask;			/* Mask file pointer */
  Boolean structNoFlag;		/* Structure number specified */
  Boolean structStr1DFlag;	/* Structure string specified (bin) */
  Boolean structStrFlag;	/* Structure string specified (quad) */
  int structNo;			/* Structure description number  */
  char *structStr1D;		/* Structure description string (bin) */
  char *structStr;		/* Structure description string (quad) */
  FilterTree filtertree;	/* Filter structure tree */
  int vectDim;
  int inFrameNo;
  Boolean maskFlag;		/* Mask file flag */
  int **mask;			/* Array-reference to the mask */
  struct header hd, hdp;
  FILE *fpo_tmp[2];		/* Files used to create columns of output */
  float **pix;			/* Pointer to input picture-element */
  int d;				/* Same as parameter -d */
  char *tmp_file[2] 
    = {tempnam(NULL,".tmp."),tempnam(NULL,".tmp.")};
  int decimation=1;

  /* Initialize variables and parse commandline */
  hipserrprt = HEL_ERROR;
  Progname = strsave(*argv);
  parseargs(argc,argv, flagfmt, &d, &train, &trainFile, &outFile,
	    &p, &seed, &maskFlag, &maskFile,
	    &structNoFlag, &structNo, 
	    &structStr1DFlag, &structStr1D, 
	    &structStrFlag, &structStr,	    
	    FFLIST,&numFiles,&filelist);
  if(seed < 0)
    seed = time(NULL);
  srand(seed);
  vectDim = 0;
  inFrameNo = 0;

  /* Build the filter structure binary tree */
  if (structStrFlag) 
    filtertree = CompileFiltStructTree(structStr,"");
  else if (structStr1DFlag) 
    filtertree = CompileFiltStructTreeFromBin(structStr1D,"");
  else if (structNoFlag)
    filtertree = BuildFiltStructTree(structNo,"");

  /* Read the mask file */
  if (maskFlag) {
    fpMask = hfopenr(maskFile);
    fread_hdr_a(fpMask,&hdMask,maskFile);
    maskMethod = fset_conversion(&hdMask,&hdpMask,maskTypes,maskFile);
    fread_imagec(fpMask,&hdMask,&hdpMask,maskMethod,0,maskFile);
    mask = (int **) matrix2(&hdpMask);
  }

  /* Scan through every frame on every inputfile and convert it to one column
     in the outputfile. Only nonzero pixels in the mask-file will be
     put out, if a mask file is specified */
  for(file=0; file<numFiles; file++) {
    fpi = hfopenr(filelist[file]);
    fread_hdr_a(fpi,&hd,filelist[file]);
    method = fset_conversion(&hd,&hdp,types,filelist[file]);
    fprintf(stderr,"File %d of %d, %d frames, frame: ",
	    file+1,numFiles,hd.num_frame);
    for(frame=0; frame<hd.num_frame; frame++) {
      fprintf(stderr,"%d ",frame+1);
      fread_imagec(fpi,&hd,&hdp,method,frame,filelist[file]);
      pix = (float **) matrix2(hdp);
      inFrameNo++;

      /* Locate the filepointer at the top */
      if (inFrameNo > 1)
	rewind(fpo_tmp[(inFrameNo+1) % 2]);
      if((fpo_tmp[inFrameNo % 2] = fopen(tmp_file[inFrameNo%2],"w+")) == NULL)
	error_handler(ERR_OPEN_FILE,"temporary file");

      if (file==0 && frame==0 
	  && (structStrFlag || structStr1DFlag || structNoFlag)) {
	vectDim+=getNSubbands(filtertree);
	decimation=crit_rate_hips2lvq (filtertree, inFrameNo, hdp, 
				       fpo_tmp, pix, d);
      }
      else {
	vectDim++;
	full_rate_hips2lvq (decimation, inFrameNo, vectDim, maskFlag, mask, 
			    hdp, fpo_tmp, pix, d);
      }
      fclose(fpo_tmp[(inFrameNo + 1) % 2]);
    }
    fprintf(stderr,"\n");
  }
  
  /* Copy data from temporary file to outputfile */
  fprintf(stderr,"Writing..\n");
  if(train)
    hips2lvq_save (trainFile,&(fpo_tmp[inFrameNo % 2]),
		   train,p,vectDim);
  hips2lvq_save (outFile,&(fpo_tmp[inFrameNo % 2]),
		 train,100.0,vectDim);
  unlink(tmp_file[0]);
  unlink(tmp_file[1]);

  return(0);
}




/* Full rate HIPS to LVQ conversion */
full_rate_hips2lvq (int decimation,
		    int inFrameNo,
		    int vectDim,
		    Boolean maskFlag,  	/* Mask file flag */
		    int **mask,	       	/* Array-reference to the mask */
		    struct header hd,
		    FILE **fpo_tmp,  	/* Files used to create output */
		    float **pix,	/* Pointer to input picture-element */
		    int d)		/* Same as parameter -d */
{
  register int x,y;		/* x,y: coordinates in current inputframe */
  char *str;			/* Current line in outputfile */
  int maxStrLen;

  /* Allocate space for temporary string, big enough to hold
     the largest possible numeric value formatted as a string at
     each position in the vector */
  maxStrLen = vectDim*(DBL_DIG + strlen("+9.E+99")+5);
  str = (char *) malloc(maxStrLen);

  /* Build new datacolumns */
  for(x=0; x<hd.orows; x+=decimation) {
    for(y=0; y<hd.ocols; y+=decimation) {
      if (!maskFlag || (maskFlag && mask[x][y])) {
	if (inFrameNo > 1) {
	  /* Not first column => read prev. columns and insert space */
	  if(fgets(str,maxStrLen,fpo_tmp[(inFrameNo+1) % 2])==NULL)
	    error_handler(ERR_READ_FILE,"temporary file");
	  strcpy((char *)strstr(str,"\n")," ");  
	  /* Remove newline, insert space between vector elements */
	}
	else
	  strcpy(str,"");
	if(fprintf(fpo_tmp[inFrameNo%2],"%s%1.*f\n",str,d,(float) pix[x][y])<0)
	  error_handler(ERR_WRITE_FILE,"temporary file");
      }
    }
  }
  free(str);
}




print_vector(int x,
	     int y,
	     int x_orig,
	     int y_orig,
	     int x_dim,
	     int y_dim,
	     FilterTree ft,
	     float decimation,
	     int inFrameNo,
	     FILE **fpo_tmp,
	     float **pix,
	     int d)
{
  int ro, kappa;
  int dec=(int) decimation;

  if (ft != NULL) {
    for(ro=0; ro<2; ro++)
      for(kappa=0; kappa<2; kappa++) {
	print_vector(x,
		     y,
		     x_orig + (ro    == 0 ? 0 : x_dim/2),
		     y_orig + (kappa == 0 ? 0 : y_dim/2),
		     x_dim/2,
		     y_dim/2,
		     ft->subband[ro][kappa],
		     decimation*2.0,
		     inFrameNo, fpo_tmp, pix, d);
      }
  }
  else {
    if(fprintf(fpo_tmp[inFrameNo%2],"%1.*f ", d, 
	       (float) pix[x_orig+x/dec][y_orig+y/dec])<0)
      error_handler(ERR_WRITE_FILE,"temporary file");
  }
}




/* Critical rate HIPS to LVQ conversion */
int
crit_rate_hips2lvq (FilterTree filtertree,
		    int inFrameNo,
		    struct header hd,
		    FILE **fpo_tmp,  	
		    float **pix,	
		    int d)
{
  int x,y;		/* x,y: coordinates in current inputframe */
  int leastCommonDecimation=pow(2.0,(float) getUpperSubLevel (filtertree)+1);

  for(x=0; x<hd.orows/leastCommonDecimation; x++) {
    for(y=0; y<hd.ocols/leastCommonDecimation; y++) {
      print_vector(x,y, 0,0, hd.orows,hd.ocols, 
		   filtertree, 1.0/leastCommonDecimation,
		   inFrameNo, fpo_tmp, pix, d);
      if(fprintf(fpo_tmp[inFrameNo%2],"\n")<0)
	error_handler(ERR_WRITE_FILE,"temporary file");
    }
  }
  return leastCommonDecimation;
}


  

/* Save p% of the lines of the temporary file into fname. Maximum number of
   vectorelements is given by vectDim*/
hips2lvq_save (fname,tmpfile,train,p,vectDim)
     Filename fname;
     FILE **tmpfile;
     Boolean train;
     double p;
     int vectDim;
{
  FILE *fpo;
  char *str;
  int maxStrLen;

  maxStrLen = vectDim*(DBL_DIG + strlen("+9.E+99")+5);
  str = (char *) malloc(maxStrLen);
  if (strcmp(fname,"<stdout>")==0)
    fpo=stdout;
  else 
    if ((fpo = fopen(fname,"w"))==NULL)
      error_handler(ERR_OPEN_FILE,fname);
  str = (char *) malloc(maxStrLen);
  rewind(*tmpfile);
  /* Write header to the outputfile  */
  if(train) {
    if(fprintf(fpo,"%d\n",vectDim-1)<0)
      error_handler(ERR_WRITE_FILE,fname);
  }
  else {
    if(fprintf(fpo,"%d\n",vectDim)<0)
      error_handler(ERR_WRITE_FILE,fname);
  }
  p *= RAND_MAX / 100.0;
  while(fgets(str,maxStrLen,*tmpfile)) {
    /* Only print out p% of the vectors */
    if(rand() <= p) 
      if(fprintf(fpo,"%s",str)<0) 
	error_handler(ERR_WRITE_FILE,fname);
  }
  if(!feof(*tmpfile))
    error_handler(ERR_READ_FILE,"temporary file");
  free(str);
  fclose(fpo);
  
  return(0);
}



