/*----------------------- Rogaland University Centre -------------------------
## \section{lvq2hips}
## 
## 
## \subsection*{Purpose}
## Convert a LVQ (Learning Vector Quantization) datafile to HIPS-2 format.
## The main use of this program is converting back classified 
## data to an image.
## 
## \subsection*{Description}
## The input LVQ-file is assumed to be the output from classify,
## without any blank lines or any comment-lines. If not, the operation
## may not be predictable.
## The result is written to the standard output.
## 
## \subsection*{Usage}
## \begin{verbatim}
## lvq2hips [-m mask file<string>] [-vis]
##   [-p [prototype<filename>] | 
##    -s rows<integer> [cols<integer>]]
##   [imagefile]
## \end{verbatim}
## \begin{description}
##   \item[-m]
## 	Mask image. The LVQ datafile is assumned to contain only the 
## 	pixels where the corresponding pixels in the mask
## 	image are nonzero.
##   \item[-vis]
## 	The input is assumed to be of SOM (Self Organizing Map) format.
##   \item[-p]
## 	HIPS-2 image (header) file, input, this file is used to determine the
## 	size of the output image
##   \item[-s]
## 	Size of the output image
##   \item[imagefile]
## 	LVQ datafile. Input.
## \end{description}
## 
## \subsection*{Date created}
## January 1993
 *------------------------ Rogaland University Centre -----------------------*/


#include <hipl_format.h>
#include <math.h>
#include "mcf.h"
#include "error.h"

#define MAX_STR_LEN 128

/* Conversion */
int maskTypes[] = {PFINT,LASTTYPE};

/* Define the commandline-format */
static Flag_Format flagfmt[] = {
  {"m",{LASTFLAG},1,{
    {PTBOOLEAN,"FALSE"},
    {PTSTRING,"","mask file"},
    LASTPARAMETER}},
  {"vis",{LASTFLAG},0,{
    {PTBOOLEAN,"FALSE"},
    LASTPARAMETER}},
  {"p",{"s",LASTFLAG},0,{
    {PTBOOLEAN,"FALSE"},
    {PTFILENAME," ","prototype"},
    LASTPARAMETER}},
  {"s",{"p",LASTFLAG},1,{
    {PTINT,"128","rows"},
    {PTINT,"-1","cols"},
    LASTPARAMETER}},
  LASTFLAG};

main(argc,argv)
     int argc;
     char **argv;
{
  struct header hdi, hdo;
  Filename file_i, file_lvq;
  FILE *fpi, *fp_lvq;	/* fpi: formatfile, fp_lvq: LVQ-file */
  int vecDim;  		/* LVQ-vector dimension */
  register int nex;	/* Skipdistance at end of ROI in row to next row */
  register float *pix;	/* Pointer to input picture-element */
  int i,j,k;
  Boolean pFlag;
  Boolean maskFlag;     /* Mask file flag */
  char *maskFile;       /* Mask file name */
  FILE *fpMask;         /* Mask file pointer */
  struct header hdMask; /* Mask file header */
  struct header hdpMask;        /* Mask file header */
  int maskMethod;       /* Mask file conversion method */
  int **mask;           /* Array-reference to the mask */
  Boolean visFlag;	/* Process the output from the SOM-PAK "visual" prog */
  char str[MAX_STR_LEN];
  int xdim,ydim,x,y;	/* visual: topology and parameters */
  int rows,cols;

  /* Initialize variables and parse commandline */
  hipserrprt = HEL_ERROR;
  Progname = strsave(*argv);
  parseargs(argc,argv,flagfmt,&maskFlag,&maskFile,&visFlag,
	    &pFlag,&file_i,&rows,&cols, 
	    FFONE,&file_lvq);

  if (pFlag) {
    fpi = hfopenr(file_i);
    fread_header(fpi,&hdi,file_i);
    rows=hdi.orows;
    cols=hdi.ocols;
  }
  else if(cols<0)
    cols = rows;

  init_hdr_alloc(&hdo,"","",1,"",rows,cols,PFFLOAT,1,"");
  write_headeru(&hdo,argc,argv);
  pix = (float *) hdo.firstpix;
  
  if ((fp_lvq = fopen(file_lvq,"r"))==NULL)
    error_handler(ERR_OPEN_FILE,file_lvq);
  
  if (visFlag) { /* !!! */
    /* Read the topology */
    if (fscanf(fp_lvq," %d %S %d %d %S ",&vecDim,str,&xdim,&ydim,str) == EOF)
      error_handler(ERR_READ_FILE,file_lvq);
  }
  else {
    /* Read the vector dimension */
    if (fscanf(fp_lvq," %d ",&vecDim) == EOF)
      error_handler(ERR_READ_FILE,file_lvq);
  }
  
  /* Read the mask file */
  if (maskFlag) {
    fpMask = hfopenr(maskFile);
    fread_hdr_a(fpMask,&hdMask,maskFile);
    maskMethod = fset_conversion(&hdMask,&hdpMask,maskTypes,maskFile);
    fread_imagec(fpMask,&hdMask,&hdpMask,maskMethod,0,maskFile);
    mask = (int **) matrix2(&hdpMask);
  }

  /* Read vectors */
  for (i=0; i<hdo.rows; i++) {
    for (j=0; j<hdo.cols; j++) {
      if (!maskFlag || (maskFlag && mask[i][j])) {
	if (visFlag) {
	  if (fscanf(fp_lvq," %d %d %S ",&x,&y,str) == EOF)
	    error_handler(ERR_READ_FILE,file_lvq);
	  *pix=(float) x*ydim + y + 1;
	}
	else {
	  /* Read all elements of the vector, keep only last element */
	  for (k=0; k<=vecDim; k++) {
	    if (fscanf(fp_lvq," %f ",pix) == EOF)
	      error_handler(ERR_READ_FILE,file_lvq);
	  }
	}
      }
      else
	*pix = 0.0;
      pix++;
    }
  }


  /* Write image */
  write_image(hdo,0);

  return (0);
}






