/*----------------------- Rogaland University Centre -------------------------
## \section{mcf (Multi Channel Filter)}
## 
## 
## \subsection*{Purpose}
## Multichannel Gabor filter with nonlinear transform and smmothing.
## Preprocessor for textured images for texture segmentation.
## 
## \subsection*{Description}
## Read in the input image, transform it by a DFT and apply a set of 
## one channel filters to this image.
## The set of filters includes:
## 
## \begin{itemize}
##   \item Filtering by a Gabor filter with frequency and orientation selective 
## properties
##   \item Some kind of linear transform on the image pixels.
## 	   Currently implemented is the absolute value nonlinearity and the
## 	   tanh($\cdots$) nonlinearity as described by Farrokhnia.
##   \item Averaging (blurring) of the filtered image by a Gaussian lowpass 
##	   filter
## \end{itemize}
## 
## The output from this program is a image sequence where the input
## image has one output image per filter in the filter set.
## The output is scaled to the maximum dynamic range of the output format.
## Be aware that this scaling implies that training a classifier on the output of
## the filter-response of one specific image and classification on the
## output of another is impossible, because the two outputs will probably
## be scaled differently.
## 
## See: ``Multi-channel filtering techniques for texture segmentation and
##      surface quality inspection'',
##      Farshid Farrokhnia, Ph.D. dissertation,
##      Michigan State University, Dept. of Elect. Eng., 1990
## 
## And: ``Determining Gabor-filter parameters for texture segmentation'',
##      Dennis F. Dunn, William E. Higgins and Joseph Wakeley,
##      SPIE Vol. 1826, Intelligent robots and computer vision XI (1992),
##      Pages 51-63
## 
## 
## 
## \subsection*{Usage}
## 
## \begin{verbatim}
## mcf [-Even | -Gen [SigmaX<real> 
##                   [SigmaY<real>]] | 
##      -s struct<string> | 
##      -sn struct<integer> | 
##      -sb struct<string>]
##     [-suvf factor in sigma-u in 
##            filter<real> 
##      [..sigma-v..<real>]]
##     [-theta numargs<integer> arg1...argn]
##     [-U0 numargs<integer> arg1...argn]
##     [-V0 numargs<integer> arg1...argn]
##     [-tanh | -nnl] 
##     [-alpha alpha in nonlin. 
##             transf.<real>]
##     [-sf gaussian window fact.<real>] 
##     [-navg] [imagefile]
## \end{verbatim}
## 
## \begin{description}
## \item[-Even]
## 	Use the even symm. Gabor filter, as described by Farrokhnia
## \item[-Gen]
## 	Use the general Gabor filter, as described by Dunn.
## 	Parameters: SigmaX, SigmaY
## \item[-s]
## Filter structure string, defined as a quad tree.
## See the module {\em filtertree} (CompileFiltStructTree)
## for a description of the structure.
## \item[-sb]
## Filterstructure string, defined as a binary tree.
## See the module {\em filtertree} (CompileFiltStructTreeFromBin)
## for a description of the structure.
## \item[-sn]
## Filter structure number, between 1 and 11 (default 4).
## See the module {\em filtertree} (BuildFiltStructTree)
## for a description of the structure.
## \item[-tanh]
## 	Use the tanh($\cdots$) nonlinearity, as described by Farrokhnia
## \item[-alpha]
## 	alpha parameter in tanh($\cdots$) nonlinearity, as described by 
## 	Farrokhnia
## \item[-U0]
## 	Centre frequency parameter(s) to the Gabor filter and the tanh($\cdots$)
## 	nonlinearity
## \item[-V0]
## 	Second centre frequency parameter(s) for a general Gabor filter. 
## 	If specified, the number of values must equal the number of 
## 	values for U0. See Dunn.
## \item[-theta]
## 	Angle parameter(s) to the Gabor filter.
## \item[-suvf]
## 	Sigma U factor: parameter to the Gabor filter (Farrokhnia)
## 	Sigma V factor: parameter to the Gabor filter (Farrokhnia)
## \item[-sf]
## 	Factor to sigma in tanh($\cdots$)
## \item[-navg]
## 	Do not average the feature-images
## \item[-nnl]
## 	Do not perform any nonlinear transform, or rather only apply the
## 	magnitude operation.
## \end{description}
## 
## A parametervalue of -1 means automatic calculation
## This program will not accept inputsequences with more than one frame.
## Specifying no parameters implies applying a even-symmetric Gabor filter
## with Farrokhnias default parameters.
## 
## \subsection*{Date created}
## January 1993
 *------------------------ Rogaland University Centre -----------------------*/

#include <hipl_format.h>
#include <math.h>
#include <float.h>
#include "struct_defs.h"
#include "mcf.h"

/* Define the commandline-format */
static Flag_Format flagfmt[] = {
  {"Even",{"Gen","s","sn","sb",LASTFLAG},0,{
    {PTBOOLEAN,"TRUE"},
    LASTPARAMETER}},
  {"suvf",{LASTFLAG},1,{
    {PTDOUBLE,DEFAULT_SIGMA_U_FACTOR,"factor in sigma-u in filter"},
    {PTDOUBLE,DEFAULT_SIGMA_V_FACTOR,"..sigma-v.."},
    LASTPARAMETER}},
  {"theta",{LASTFLAG},1,{
    {PTLIST,"","numargs<integer> arg1 ... argn"}, 
    LASTPARAMETER}}, 
  {"U0",{LASTFLAG},1,{
    {PTLIST,"","numargs<integer> arg1 ... argn"}, 
    LASTPARAMETER}}, 
  {"V0",{LASTFLAG},1,{
    {PTLIST,"","numargs<integer> arg1 ... argn"}, 
    LASTPARAMETER}}, 
  {"Gen",{"Even","s","sn","sb",LASTFLAG},0,{
    {PTBOOLEAN,"FALSE"},
    {PTDOUBLE,"50","SigmaX"}, 
    {PTDOUBLE,"-1","SigmaY"}, 
    LASTPARAMETER}},
  {"tanh",{"nnl",LASTFLAG},0,{{PTBOOLEAN,"TRUE"},LASTPARAMETER}},
  {"alpha",{LASTFLAG},1,
     {{PTDOUBLE,"0.25","alpha in nonlin. transf."}, LASTPARAMETER}},
  {"sf",{LASTFLAG},1,{
    {PTDOUBLE,DEFAULT_SIGMA_FACTOR,"gaussian window fact"}, 
    LASTPARAMETER}},
  {"navg",{LASTFLAG},0,{{PTBOOLEAN,"FALSE"},LASTPARAMETER}},
  {"nnl",{"tanh",LASTFLAG},0,{{PTBOOLEAN,"FALSE"},LASTPARAMETER}},
  {"sn",{"Gen","Even","s","sb",LASTFLAG},1,{
    {PTBOOLEAN,"FALSE"},
    {PTINT,"4","struct"},
    LASTPARAMETER}},
  {"sb",{"Gen","Even","s","sn",LASTFLAG},1,{
    {PTBOOLEAN,"FALSE"},
    {PTSTRING,"(,)","struct"},
    LASTPARAMETER}},
  {"s",{"Gen","Even","sb","sn",LASTFLAG},1,{
    {PTBOOLEAN,"FALSE"},
    {PTSTRING,"(,)","struct"},
    LASTPARAMETER}},
  LASTFLAG};


main(argc,argv)
     int argc;
     char **argv;
{
  int U0_count,theta_count;	/* Count of parameters from command line */
  int fr,nfr;			/* Current and number of frames */
  double **U0_num;		/* List of numeric valued U0 */
  double **V0_num;		/* List of numeric valued V0 */
  double **theta_num;		/* List of numeric valued theta */
  double SigmaUfactor;		/* Parameter to the even symmetric Gabor filt*/
  double SigmaVfactor;		/* Parameter to the even symmetric Gabor filt*/
  double sigmaFactor;		/* Factor in parameter to the blurrer */
  double SigmaLP;		/* Parameter to the blurrer */
  Listarg U0, V0, theta;	/* Commandline parameters */
  Filename filename;		/* Input file name */
  struct header hd;		/* Input header (in original format) */
  struct header hdi;		/* Fourier-transformed input header */
  struct header hdo;		/* Output header */
  int method;			/* Conversion method */
  Boolean EvenGaborFlag;	/* Use the even symmetric Gabor filter */
  Boolean GenGaborFlag;		/* Use the general (complex) Gabor filter */
  Boolean tanhFlag;		/* Use the tanh(..) nonlinearity */
  Boolean nnl;			/* Don't use any nonlinearity */
  Boolean navg;			/* Don't blurr the image after filtering */
  double SigmaX, SigmaY;	/* Parameters to the general Gabor filter */
  double alpha;			/* Parameter to the nonlinearity */
  int nonlin;			/* Identifier telling what nonlin. to use */
  int blurr;			/* Identifier telling what blurring to use */
  int  struct_no;		/* Subband split structure number */
  Boolean structNoFlag;		/* Structure number specified */
  Boolean structStrFlag;	/* Structure string specified (quad) */
  Boolean structStr1DFlag;	/* Structure string specified (bin) */
  Boolean analysisFlag;		/* Analysis filtering */
  Boolean synthesisFlag;	/* Synthesis filtering */
  char *structStr;		/* Structure description string (quad) */
  char *structStr1D;		/* Structure description string (bin) */
  FilterTree filtertree;	/* Filter structure tree */

  hipserrprt = HEL_ERROR;	/* Suppress messaging from HIPS routines */

  /* Read the commandline */
  Progname = strsave(*argv);
  parseargs(argc,argv,flagfmt,&EvenGaborFlag,&SigmaUfactor,&SigmaVfactor,
	    &theta, &U0,&V0,&GenGaborFlag,&SigmaX,&SigmaY,
	    &tanhFlag,&alpha,&sigmaFactor,&navg,&nnl,
	    &structNoFlag,&struct_no,
            &structStr1DFlag,&structStr1D,&structStrFlag,&structStr,
	    FFONE,&filename);
  SigmaY = (SigmaY <= 0 ? SigmaX : SigmaY);
  nonlin = (tanhFlag ? TANH_NONLIN : NO_NONLIN);
  blurr = (navg ? NO_BLURR : BLURR);
  

  /* Read in the image and transform it into the Fourier-domain */
  read_ft(&hd,&hdi,&method,filename);

  /* Initialize parameters for the filtering process
     Use the scheme suggested by [Jain/Farrokhnia, 1991] if the parameters 
     are not specified on the commandline.
     */
  parseUV(&hd,&U0,&U0_num,TRUE);
  parseUV(&hd,&V0,&V0_num,FALSE);
  parseTheta(&theta,&theta_num);
  if (V0.argcount!=0 && V0.argcount != U0.argcount)
    perr(HE_SYNTAX,
	 "Number of V0 must either be zero or equal the number of U0");

  /* Write the header */
  if (EvenGaborFlag || GenGaborFlag)
    nfr = U0.argcount * theta.argcount;
  else if (structNoFlag || structStr1DFlag || structStrFlag) {
    if (structStrFlag) 
      filtertree = CompileFiltStructTree(structStr,"Gabor");
    else if (structStr1DFlag) 
      filtertree = CompileFiltStructTreeFromBin(structStr1D,"Gabor");
    else
      filtertree = BuildFiltStructTree(struct_no,"Gabor");
    PrintFiltStructTree (filtertree);
    nfr = getNSubbands (filtertree);
  }
  /* Allocate output image */
  dup_headern(&hdi,&hdo);
  hdo.num_frame = nfr;
  alloc_image(&hdo);
  hd.num_frame = nfr;
  write_headeru(&hd,argc,argv);

  /* Create one response-image per filterselection */
  fr = 0;
  
  if(EvenGaborFlag) {
    /* Gabor-filter as described by F. Farrokhnia */
    for (U0_count=0; U0_count<U0.argcount; U0_count++) {
      /* Normalize U0 */
      **(U0_num+U0_count) *= ((hdo.cols-1.0)/(sqr(hdo.cols)));
      for (theta_count=0; theta_count<theta.argcount; theta_count++) {
	if (**(U0_num+U0_count) !=0)
	  SigmaLP = sigmaFactor/(**(U0_num+U0_count));
	else
	  SigmaLP = sigmaFactor;
	one_channel_filter(&hdi,&hdo,
			   **(U0_num+U0_count),0.0, **(theta_num+theta_count),
			   SigmaUfactor * **(U0_num+U0_count),
			   SigmaVfactor * **(U0_num+U0_count),
			   SigmaLP, alpha, EVEN_GABOR, nonlin, blurr,FALSE);
	fprintf(stderr,"write\n");
	mcf_write_image(&hd,&hdo,fr++,TRUE);
      }
    }
  }
  else if (GenGaborFlag) {
    /* Gabor-filter as described by Dunn, 1992 */
    for (U0_count=0; U0_count<U0.argcount; U0_count++) {
      /* U0 and V0 are assumed to be normalized when input with -Dunn */
      for (theta_count=0; theta_count<theta.argcount; theta_count++) {
	if (sqr(**(U0_num+U0_count)) + sqr(**(V0_num+U0_count)) !=0)
	  SigmaLP = sigmaFactor/sqrt(sqr(**(U0_num+U0_count)) +
				     sqr(**(V0_num+U0_count)));
	else 
	  SigmaLP = sigmaFactor;
	one_channel_filter(&hdi,&hdo,
			   **(U0_num+U0_count),**(V0_num+U0_count),
			   **(theta_num+theta_count),
			   SigmaX, SigmaY, 
			   SigmaLP, alpha, GENERAL_GABOR, nonlin, blurr,FALSE);
	fprintf(stderr,"write.\n");
	mcf_write_image(&hd,&hdo,fr++,TRUE);
      }
    }
  }
  else if (structNoFlag || structStr1DFlag || structStrFlag) {
    /* Filter-tree Gabor filter */
    struct_gabor(filtertree,&hdi,&hd,method,sigmaFactor,alpha,nonlin,blurr);
  }
  return(0);
}

