/*----------------------- Rogaland University Centre -------------------------
## \section{mcf\_util}
## 
## 
## \subsection*{Purpose}
## Various utility routines for multichannel filtering
## 
## \subsection*{Description}
## \subsubsection*{parseUV}
## 	Convert the commandline-list for U0 or V0 to a numeric list.
## \subsubsection*{parseTheta}
## 	Convert the commandline-list for theta to a numeric list
## \subsubsection*{read\_ft}
## 	Read in the input image and convert it to its fourier
## 	transformed counterpart.
## \subsubsection*{mcf\_write\_image}
## Transform the frame into the maximum range of the output
## 	format, and write the frame 
## \subsubsection*{max\_range}
## Linear scale an image into the maximal range for the specified
## 	format.
## \subsubsection*{scale}
## 	Linear scale an image to the range defined by the calling parameters.
## 
## 
## \subsection*{Usage}
##
## \subsubsection*{parseUV}
## \begin{verbatim}
## parseUV (hd,uv, uv_num, genDef)
##      struct header *hd;
##      Listarg *uv;      
##      double ***uv_num; 
##      Boolean genDef;   
## \end{verbatim}
## \begin{itemize}
##   \item	Input image header (to determine number of rows and columns)
##   \item	List of U0 or V0 strings --- input.
##   \item	List of numeric U0 or V0 --- output.
##   \item	Flag indicating wheter to generate the defalt values 
##         	(for U0 only)
## \end{itemize}
##
## \subsubsection*{parseTheta}
## \begin{verbatim}
## parseTheta (theta, theta_num)
##      Listarg *theta;     
##      double ***theta_num;
## \end{verbatim}
## \begin{itemize}
##   \item	List of theta --- input
##   \item	List of numeric theta --- output 
## \end{itemize}
##
## \subsubsection*{read\_ft}
## \begin{verbatim}
## read_ft(hd, hdft, method, filename)
##      struct header *hd;  
##      struct header *hdft;
##      int *method;
##      Filename filename;
## \end{verbatim}
## \begin{itemize}
##   \item	Input format header (will be kept in order to convert back)
##   \item	Input header, converted to appropriate format and Fourier 
##        	transformed
##   \item	Conversion method
##   \item	Input filename
## \end{itemize}
##
## \subsubsection*{mcf\_write\_image}
## \begin{verbatim}
## mcf_write_image(hd,hdo,fr,MaxRange)
##      struct header *hd;
##      struct header *hdo;
##      int fr;
##      Boolean MaxRange;
## \end{verbatim}
## \begin{itemize}
##   \item	Input format header (the same format as the output will be 
##         stored as)
##   \item	Output header
##   \item	Frame number
##   \item	Flag, when TRUE, convert the output written image to
##         	the maximum dynamic range of the output format.
##         	This flag will normally be FALSE for floating point images, and
##         	TRUE for fixed point images, such as byte images.
## \end{itemize}
##
## \subsubsection*{max\_range}
## \begin{verbatim}
## max_range(hdi,hdo,format)
##      struct header *hdi,*hdo;
##      int format;
## \end{verbatim}
## \begin{itemize}
##   \item	Input image header
##   \item	Output image header
##   \item	Pixel format
## \end{itemize}
##
## \subsubsection*{scale}
## \begin{verbatim}
## scale (hdi,hdo,n_min,n_max)
##      struct header *hdi,*hdo;
##      float n_min,n_max;
## \end{verbatim}
## \begin{itemize}
##   \item	Input image header
##   \item	Output image header
##   \item      Minimal value of the output
##   \item      Maximal value of the output
## \end{itemize}
## 
## \subsection*{Date created}
## January 1993
 *------------------------ Rogaland University Centre -----------------------*/


#include <math.h>
#include <limits.h>
#include <float.h>
#include <hipl_format.h>
#include "mcf.h"



parseUV (hd,uv, uv_num, genDef)
     struct header *hd;	/* Input-image header */
     Listarg *uv;	/* Commandline-list of U0 or V0 */
     double ***uv_num;	/* List of numeric-value U0 or V0 */
     Boolean genDef;	/* Generate default values (for U0 only) */
{
  int i;
  int skip_low;		/* Number of lowest radial frequencies to skip */

  /* Specify radial frequencies for the filterbank */
  if(uv->argcount==0 && genDef)  {
    /* No list of U is specified, generate default set */
    /* Updated 10/2-94, old line: uv->argcount = log2((hd->cols)/4)-1; */
    uv->argcount = 5;
  }

  *uv_num =(double **) calloc(uv->argcount,sizeof(double *));

  for(i=0;i<uv->argcount;i++) {
    *(*uv_num+i)=(double *)malloc(sizeof(double));
    if(uv->args==NULL) {
      /* No arguments are specified, use arguments as specified by
	 Jain/Farrokhnia [1991] */
      skip_low = (log2((hd->cols)/4)+1) - 5;
      /* Updated 10/2-94, old line: **(*uv_num+i) = pow(2,i)*sqrt(2); */
      **(*uv_num+i) = pow(2,i+skip_low)*sqrt(2);
    }
    else {
      /* Parse arguments from the commandline */
      **(*uv_num+i) = atof(*(uv->args+i));
    }
  }
  return(HIPS_OK);
}
  

parseTheta (theta, theta_num)
     Listarg *theta;		/* Commandline-list of theta */
     double ***theta_num;	/* List of numeric-value theta */
{
  int i;

  /* Specify angles for the filterbank in degrees */
  if(theta->argcount==0)  
    theta->argcount = 4;
  *theta_num =(double **) calloc(theta->argcount,sizeof(double *));
  for(i=0;i<theta->argcount;i++) {
    *(*theta_num+i)=(double *)malloc(sizeof(double));
    if(theta->args==NULL)
      /* No arguments are specified, use arguments as specified by
	 Jain/Farrokhnia [1991] */
      **(*theta_num+i) = i*45;
    else
      /* Parse arguments from the commandline */
      **(*theta_num+i) = atof(*(theta->args+i));
  }
  return(HIPS_OK);
}



read_ft(hd, hdft, method, filename)
     struct header *hd;		/* Original sequenceheader */
     struct header *hdft; 	/* Header of the fourier-transformed seq. */
     int *method;
     Filename filename;
{
  struct header hdp, hd_im;
  FILE *fp;
  float mean,scale,min,max;
  /* Define the conversion done by the reading of the input-image */
  int types[] = {PFFLOAT,LASTTYPE};

  /* Read the image */
  fp = hfopenr(filename);
  fread_hdr_a(fp,hd,filename);
  *method=set_conversion(hd,&hdp,types);
  fread_imagec(fp,hd,&hdp,*method,0,filename);
  fclose(fp);

  /* Set the mean to 0 in order to avoid filterresponse to the mean gray */
  h_mean(&hdp,&mean,FALSE);
  h_linscale(&hdp,&hdp,1.,-mean);

  /* Transform the image into the complex domain 
     (the Fourier-transform takes a complex image as input) */
  dup_headern(&hdp,&hd_im);
  alloc_imagez(&hd_im);
  dup_headern(&hdp,hdft);
  setformat(hdft,PFCOMPLEX);
  alloc_image(hdft);
  h_combine(&hdp,&hd_im,hdft,FALSE);
  free_image(&hdp);
  free_image(&hd_im);

  /* Fourier-transform the image */
  h_fourtr(hdft);

  /* Scale the image to [-1,1] */
  h_minmax(hdft,&min,&max,FALSE);
  scale = MAX(abs(max),abs(min));
  h_divscale(hdft,hdft,&scale);

  /* Set conversion method as the method from input to output format */
  *method = find_method (hd->pixel_format,hdft->pixel_format);

  return(HIPS_OK);
}


mcf_write_image(hd,hdo,fr,MaxRange)
     struct header *hd;		/* Original sequence header */
     struct header *hdo;       	/* Header of filtered image */
     int fr;			/* Frame-number */
     Boolean MaxRange;		/* If TRUE: transform the image to */
				/* the maximum range */
{
  struct header hdf;
  int method;

  dup_headern(hdo,&hdf);
  setformat(&hdf,PFFLOAT);
  alloc_image(&hdf);
  h_tof(hdo,&hdf);
  if(MaxRange) {
    /* Tranform the image into maximum range */
    max_range(&hdf,&hdf,hd->pixel_format);
  }
  method = find_method(hd->pixel_format,hdf.pixel_format);
  return (write_imagec(hd,&hdf,method,TRUE,fr));
}


max_range(hdi,hdo,format)
     struct header *hdi,*hdo;	/* Input and output sequence */
     int format;		/* Desired pixel-format */
{
  switch (format) {
    case PFMSBF  	:
    case PFLSBF    	:
      return(scale(hdi,hdo,(float ) 0         ,(float) 1));
      break;
    case PFBYTE   	: 
      return(scale(hdi,hdo,(float ) 0         ,(float) UCHAR_MAX));
      break;
    case PFSHORT   	: 
      return(scale(hdi,hdo,(float ) SHRT_MIN  ,(float ) SHRT_MAX));
      break;
    case PFINT   	: 
      return(scale(hdi,hdo,(float ) INT_MIN   ,(float ) INT_MAX));
      break;
    case PFFLOAT   	: 
      return(scale(hdi,hdo,(float ) -FLT_MAX  ,(float ) FLT_MAX));
      break;
    case PFCOMPLEX   	: 
      return(scale(hdi,hdo,(float ) -FLT_MIN  ,(float ) FLT_MAX));
      break;
    case PFDOUBLE   	: 
      return(scale(hdi,hdo,(float ) -FLT_MAX  ,(float ) FLT_MAX));
      break;
    case PFDBLCOM   	: 
      return(scale(hdi,hdo,(float ) -FLT_MAX  ,(float ) FLT_MAX));
      break;
    case PFSBYTE   	: 
      return(scale(hdi,hdo,(float ) SCHAR_MIN ,(float ) SCHAR_MAX));
      break;
    case PFUSHORT   	: 
      return(scale(hdi,hdo,(float ) 0         ,(float ) USHRT_MAX));
      break;
    case PFUINT   	: 
      return(scale(hdi,hdo,(float ) 0         ,(float ) UINT_MAX));
      break;
    default       	: return(perr(HE_FMT,"- unable to scale output"));
    }
}



scale (hdi,hdo,n_min,n_max)
     struct header *hdi,*hdo;
     float n_min,n_max;
{
  float o_min, o_max;	/* Current minimum and maximum */
  float b, c;		/* Parameters to h_linscale */

  h_minmax(hdi,&o_min,&o_max,FALSE);
  if (o_max > o_min) {
    b = (n_max - n_min) / (o_max - o_min) ;
    c = n_min - b * o_min;
  }
  else {
    b = 1.0;
    c = 0.0;
  }

  return(h_linscale(hdi,hdo,b,c));
}			      


