/*----------------------- Rogaland University Centre -------------------------
## \section{mktex}
## 
## \subsection*{Purpose}
## Generate synthetic textures consisting of repeated line-segment
## texture elements.
## 
## \subsection*{Description}
## Program to generate artificial textures as defined in a .tx file.
## Relies on the OutputImage routine in {\tt mktex_imout.c}.
## 
## The program generates textures consisting of texture elements with
## specified form repeated over a grid, with a range of random offsets
## and rotations.
## 
## Based on a program by: Y. King, April 1991,
## Laboratory for Computational Intelligence,
## University of British Columbia
## 
## \subsection*{Usage}
## \begin{verbatim}
## mktex [<] infile.tx [>] outfile.hips
## \end{verbatim}
## 
## \subsection*{Format of the inputfile}
## 
## {\tiny
## \begin{verbatim}
## texture: rectangle (<upper left corner>) - (<lower right corner>)
## 	orientation = [<min rot. angle>,<max rot. angle>]:<resolution>
## 	offset = [<max x offset>,<max y offset>]:[<x resolution>,
## 		  <y res.>]
## 	pixels = <greylevel-value of pixels>
## 
## 	<linesegments in texture element>
## \end{verbatim}
## }
## 
## The following example file defines an image with two textures, 
## triangles and arrows.
## 
## {\scriptsize
## \begin{verbatim}
## 	image: 128 x 256
## 	grid: 6 x 12
## 	
## 	texture: rectangle (0,0) - (127,127)
## 		orientation = [0,360]:90
## 		offset = [0,0]:[1,1]
## 		pixels = 255
## 	
## 		(-5,-5) - (5,-5)
## 		(5,-5) - (5,5)
## 		(5,5) - (-5,-5)
## 	
## 	texture: rectangle (0,127) - (127,255)
## 		orientation = [0,360]:90
## 		offset = [3,5]:[1,1]
## 		pixels = 255
## 	
## 		(-5,-5) - (5,-5)
## 		(5,-5) - (5,5)
## 		(5,-5) - (-5,5)
## \end{verbatim}
## }
## 
## It generates a texture with 128 pixels in the vertical (x) direction 
## and 256 pixels in the horisontal (y) direction (defined by the 
## `image:' line).
## 
## The texture elements will be rotated by 0, 90, 180, 270 or 360 degrees.
## The last texture will have offset from the grid by 0, 1, 2 or 3 pixels
## in vertical direction and 0, 1, 2, ... 5 pixels from the grid in 
## horizontal direction.
## 
## 
## 
## \subsection*{Date created}
## January 1993
 *------------------------ Rogaland University Centre -----------------------*/



#include <stdio.h>
#include <math.h>
#include "mktex.h"

/* Global variables */
int	*image;			/* integer image array (variable size) */
int	xsize, ysize;
int	xgrid, ygrid;
int	xstep, ystep;
TEXTURE	*textures = NULL;	/* list of textures */

/* Function prototypes */
int	main();
void	read_element();
void	write_texture();
void	plot_element();
void	rotate_point();
void	draw_line();
void	plot_point();



int main( argc, argv )     
     int	argc;
     char	*argv[];
{
  int x,y;
  char sh;
  TEXTURE *tex;

  srand(time(NULL));

  /* Parse simple command line */
  if (argc == 2) {
    freopen(argv[1], "r", stdin);
  }
  else if (argc == 3) {
    freopen(argv[1], "r", stdin);
    freopen(argv[2], "w", stdout);
  }
  else if (argc != 1) {
    fprintf(stderr,"Usage:\n\tmktex [infile.tx [outfile.iff]]\n");
    exit(1);
  }

  /* Read size of image and texture grid */
  scanf(" image: %d x %d ", &xsize, &ysize);
  scanf(" grid: %d x %d ", &xgrid, &ygrid);
  xstep = ROUND((double)xsize/xgrid);
  ystep = ROUND((double)ysize/ygrid);
  
  /* Read in texture descriptions */
  while(scanf(" texture: %c %*[^(] ", &sh) == 1) {
    tex = (TEXTURE *) malloc(sizeof(TEXTURE));
    tex->shape = sh;
    scanf(" ( %lf , %lf ) - ( %lf , %lf ) ",&tex->tl.x,&tex->tl.y,
	  &tex->br.x,&tex->br.y);
    read_element(tex);    
    tex->next = textures;
    textures = tex;
  }
  
  if (ferror(stdin)) {
    fprintf(stderr,"Problem with input.\n");
    exit(2);
  }
  
  /* Allocate big chunk of memory for image */
  image = (int *) malloc(xsize * ysize * sizeof(int));
  
  /* Zero the array (this may be unneccessary) */
  for (x = 0; x < xsize; x++) 
    for (y = 0; y < ysize; y++)
      plot_point((double)x, (double)y, 0);

  /* Traverse the image, at intervals given by the texture grid, plotting 
     elements from the first texture in the list that is meant to go at each
     location. */
/* !!!! */
  for (x = xstep/2; x < xsize; x += xstep) {
    for (y = ystep/2; y < ysize; y += ystep) {
      for (tex = textures; tex != NULL; tex = tex->next) {
	/* Note: circular textures not implemented yet ... */
	if ((tex->shape == 'r') &&
	    (x >= tex->tl.x) && (x <= tex->br.x) &&
	    (y >= tex->tl.y) && (y <= tex->br.y)) {
	  plot_element(tex, x, y);
	  
	  /* Remove "break;" for overlay textures */
	  break;
	}
      }
    }
  }
  
  /* OutputImage should take an integer array and output an image file to
     stdout in whatever desired format. */
  
  OutputImage(image, xsize, ysize,argc,argv);
}



/*--------------------------------------------------
 *	read_element
 *		Read element definition from stdin.
 */

void	read_element( tex )
     
     TEXTURE	*tex;
{
  LINELIST	*line;
  double	x1,y1,x2,y2;
  
  scanf(" orientation = [ %d , %d ] : %d ", &(tex->min_orientation),
	&(tex->max_orientation), &(tex->res_orientation));
  scanf(" offset = [ %d , %d ] : [ %d , %d ] ", &(tex->dx), &(tex->dy),
	&(tex->res_dx), &(tex->res_dy));
  scanf(" pixels = %d ", &(tex->pixels));
  
  tex->lines = NULL;
  
  while (scanf(" ( %lf , %lf ) - ( %lf , %lf ) ", &x1,&y1,&x2,&y2) == 4) {
    line = (LINELIST *) malloc(sizeof(LINELIST));
    line->start.x = x1;
    line->start.y = y1;
    line->end.x = x2;
    line->end.y = y2;
    
    line->next = tex->lines;
    tex->lines = line;
  }
}



/*--------------------------------------------------
 *	Plot a texture element at the specified
 *	(x,y) position in the global image array.
 */

void	plot_element(tex, x, y)  
     TEXTURE	*tex;
     int	x,y;
{
  LINELIST	*line;
  POINT	p1, p2;
  double	dx,dy;
  double	theta;
  
  if(tex->res_dx==0 || tex->res_dy==0 || tex->res_orientation==0) {
    fprintf(stderr,"Error in input, resolution may not be zero\n");
    exit(1);
  }

  /* Create random displacement between 0 and dx in x-direction, at
     specified resolution */
  if(tex->dx != 0)
    dx = ((rand() >> 6) % ((tex->dx) / tex->res_dx + 1)) * tex->res_dx;
  else
    dx = tex->dx;

  /* Create random displacement between 0 and dy in y-direction, at
     specified resolution */
  if(tex->dy != 0)
    dy = ((rand() >> 6) % ((tex->dy) / tex->res_dy + 1)) * tex->res_dy;
  else
    dy = tex->dy;

  /* Create random rotation between minimum and maximum rotation, at
     specified resolution */
  if(tex->max_orientation - tex->min_orientation != 0) {
    theta = (double) 
      ((rand() >> 6) % ((tex->max_orientation - tex->min_orientation ) 
			/ tex->res_orientation + 1)) * tex->res_orientation;
    theta += tex->min_orientation;
    theta *= PI / 180.0;
  }
  else
    theta = 0.;
  
  /* Draw one texture-element */
  for (line = tex->lines; line != NULL; line = line->next) {
    rotate_point(&(line->start), theta, &p1);
    rotate_point(&(line->end), theta, &p2);
    p1.x += x + dx; p1.y += y + dy;
    p2.x += x + dx; p2.y += y + dy;
      
    draw_line(&p1, &p2, tex->pixels);
  }
}




/*--------------------------------------------------
 *	Rotate a point p1 by theta degrees to produce
 *	point p2. This is used to implement the
 *	rotation of texture elements.
 */

void	rotate_point(p1, theta, p2)  
     POINT	*p1;
     double	theta;
     POINT	*p2;
{
  p2->x = p1->x * cos(theta) + p1->y * sin(theta);
  p2->y = p1->y * cos(theta) - p1->x * sin(theta);
}




/*--------------------------------------------------
 *	Draw a line in the global image array.
 */
void	draw_line(p1,p2, val)
     POINT	*p1, *p2;
     int	val;
{
  double	dx, dy;
  double	m;
  double	x,y;
  
  dx = p2->x - p1->x;
  dy = p2->y - p1->y;
  if (fabs(dx) > fabs(dy)) {
    m = (double) dy / (double) dx;
    
    x = MIN(p1->x, p2->x);
    y = x == p1->x ? p1->y : p2->y;
    while (x <= MAX(ROUND(p1->x), ROUND(p2->x))) {
      plot_point(x, y, val);
      x += 1;
      y += m;
    }
  }
  else if (fabs(dy) > 0.0001) {
    m = (double) dx / (double) dy;
    y = MIN(p1->y, p2->y);
    x = y == p1->y ? p1->x : p2->x;
    while (y <= MAX(ROUND(p1->y), ROUND(p2->y))) {
      plot_point(x, y, val);
      x += m;
      y += 1;
    }
  }
}



/*--------------------------------------------------
 *	"plot" a point in the global image array.
 */

void	plot_point(x, y, val)
     
     double	x,y;
     int	val;
{
  x = ROUND(x);
  y = ROUND(y);
  if ((x >= 0) && (x < xsize) && (y >= 0) && (y < ysize)) 
    image[(int) x * ysize + (int) y] = val % PIXSIZE;
}


