/*----------------------- Rogaland University Centre -------------------------
## \section{one\_channel\_filter}
## 
## 
## \subsection*{Purpose}
## One channel filtering with some preprocessing options
## 
## \subsection*{Description}
##
## \subsubsection*{one\_channel\_filter}
## \begin{itemize}
##   \item        Filter with frequency and orientation selective properties.
##   \item        Some kind of linear transform on the image pixels.
##   \item        Smmothing (blurring) of the filtered image by a Gaussian
##         lowpass filter.
## \end{itemize}
## 
## \subsection*{Usage}
##
## \subsubsection*{one\_channel\_filter}
## \begin{verbatim}
## one_channel_filter (hdi,hdo,U0, V0, 
##             theta, Sigma1, Sigma2, 
##             SigmaLP, alpha, filter, 
##             nonlin, blurr, twoangle)
##      struct header *hdi, *hdo;
##      double U0, V0, theta, Sigma1, 
##             Sigma2, SigmaLP, alpha;      
##      int filter, nonlin, blurr;         
##      Boolean twoangle;
## \end{verbatim}
## \begin{itemize}
##   \item	Input image 
##   \item	Output image
##   \item	Vert. centre frequency
##   \item	Horz. centre frequency
##   \item	Rotation of filter
##   \item	Vert. sigma for filter
##   \item	Horz. sigma for filter
##   \item	Sigma for low-pass `feature' filter
##   \item	Alpha in tanh(..) nonlinearity
##   \item	Filter type identifier
##         	(NO_FILTER, GENERAL_GABOR or EVEN_GABOR)
##   \item	Nonlinearity identifier
##        	(NO_NONLIN or TANH_NONLIN)
##   \item	Blurring identifier (blurring=lowpass filtering)
##         	(NO_BLURR or BLURR)
##   \item      Use both positive and negative angle?
##         	If true, then the frequeny responses of both the Gabor
##         	filter with +theta angle and theta+90 degrees angles will be
##         	added to form one frequency response of the filter.
##         	Remember that filtering is performed by multiplication
## 		in the Fourier domain.
## \end{itemize}
## 
## \subsection*{Date created}
## January 1993
 *------------------------ Rogaland University Centre -----------------------*/


#include <hipl_format.h>
#include <math.h>
#include <float.h>
#include "mcf.h"


one_channel_filter (hdi,hdo,U0, V0, theta, Sigma1, Sigma2, SigmaLP, alpha,
		    filter, nonlin, blurr, twoangle)
     struct header *hdi;/* Input image */
     struct header *hdo;/* Output image */
     double U0; 	/* Vert. centre frequency */
     double V0;		/* Horz. centre frequency */
     double theta;	/* Rotation of filtre */
     double Sigma1;	/* Vert. sigma for filtre */
     double Sigma2;	/* Horz. sigma for filtre */
     double SigmaLP;	/* Sigma for low-pass feature filtre */
     double alpha;	/* Alpha in tanh(..) nonlinearity */
     int filter;	/* Filter type identifier */
     int nonlin;	/* Nonlinearity identifier */
     int blurr;		/* Blurring identifier */
     Boolean twoangle;	/* Use both theta and theta+90 as angle (sum resp.) */
{
  struct header hdf, hdf2;	/* Filter frequency response */

  /* Filtering */
  fprintf(stderr,"Filter, ");
  if (filter == NO_FILTER)
    h_copy (hdi,hdo);
  else {
    fprintf(stderr,"Gabor, ");
    dup_headern(hdi,&hdf);
    alloc_image(&hdf);
    if (filter == EVEN_GABOR)
      even_gabor(&hdf,U0,Sigma1,Sigma2,theta);
    else if (filter == GENERAL_GABOR) 
      gen_gabor(&hdf,U0,V0,Sigma1,Sigma2,theta);
    if (twoangle) {
      /* Use both theta and theta+90 degrees angles */
      dup_headern(&hdf,&hdf2);
      alloc_image(&hdf2);
      if (filter == EVEN_GABOR)
	even_gabor(&hdf2,U0,Sigma1,Sigma2,theta+90.0);
      else if (filter == GENERAL_GABOR) 
	gen_gabor(&hdf2,U0,V0,Sigma1,Sigma2,theta+90.0);
      h_add(&hdf,&hdf2,&hdf);
      free_image(&hdf2);
    }

    h_mul(hdi,&hdf,hdo);
    free_image(&hdf);
  }
  
  /* Inverse-fourier-transform the image and scale it to [-1,1] */
  mcf_invfourtr(hdo);
  
  /* Nonlinear transform */
  if (nonlin == TANH_NONLIN) {
    fprintf(stderr,"tanh(..), ");
    mcf_tanh_nonlin(hdo,hdo,alpha);
  }
  
  /* Average the feature-image */
  if (blurr == BLURR) {
    fprintf(stderr,"average, ");
    mcf_gauss(hdo,hdo,SigmaLP);
  }
}
