/*----------------------- Rogaland University Centre -------------------------
## \section{qmf\_util}
## 
## 
## \subsection*{Purpose}
## Utility routines for QMF filters
## 
## \subsection*{Description}
## Functions with only minor or no modifications from the functions written
## by John Haakon Husoey of the same name. Note that although the function may
## be the same, it may refer to structures that has been altered (in 
## {\tt struct_defs.h}).
## \subsubsection*{extr\_subband\_line}
## \subsubsection*{insert\_subband\_line}
## \subsubsection*{allpass\_c}
## \subsubsection*{read\_filtfile}
## \subsubsection*{vector}
## \subsubsection*{ivector}
## \subsubsection*{dvector}
## \subsubsection*{matrix}
## \subsubsection*{matrix1\_f}
## \subsubsection*{matrix2\_f}
## \subsubsection*{matrix4\_f}
## \subsubsection*{matrix2\_uch}
## \subsubsection*{dmatrix}
## \subsubsection*{imatrix}
## \subsubsection*{free\_vector}
## \subsubsection*{free\_ivector}
## \subsubsection*{free\_dvector}
## \subsubsection*{free\_matrix}
## \subsubsection*{free\_dmatrix}
## \subsubsection*{free\_imatrix}
## \subsubsection*{free\_submatrix}
## \subsubsection*{submatrix}
## \subsubsection*{convert\_matrix}
## \subsubsection*{free\_convert\_matrix}
## 
## 
## New functions of fuctions that has been completely rewritten:
## \subsubsection*{read\_fir}
## 	Read the FIR filter-taps from a file.
## 	The first number in the filterfile indicates number of taps.
## 	If the number of taps is negative, then the filter is 
## 	assumed to be non-symmetric.
## 	For symmetric filters, the filtertap file begins with the
## 	center tap, and only has half the tap-file.
##	This function will search for the filter specification in the
##	directory given in the environmentvariable FILT\_DIR.
## \subsubsection*{read\_iir}
## 	Read the IIR filter-taps from a file
##	This function will search for the filter specification in the
##	directory given in the environmentvariable FILT\_DIR.
## \subsubsection*{firfilt\_p}
## 	FIR filter with periodic extension.
## \subsubsection*{firfilt\_s}
## 	FIR filter with symmetric extension.
## 	NB! this filter will not work properly if the input from the
## 	level above is not subsampled.
## \subsubsection*{free\_matrix2\_f}
## 	Release memory occupied by {\tt matrix2\_f}.
## 
## 
## \subsection*{Usage}
##
## \subsubsection*{read\_fir}
## \begin{verbatim}
## read_fir(filt_par,filter_id)
##      struct fir_filterspec *filt_par;
##      char *filter_id;
## \end{verbatim}
## \begin{itemize}
##   \item		Filterparameter pointer.
##   \item		Filteridentifier.
## \end{itemize}
##
## \subsubsection*{read\_iir}
## \begin{verbatim}
## read_iir(filt_par,filter_id)
##      struct filterspec *filt_par;
##      char *filter_id;
## \end{verbatim}
## \begin{itemize}
## \item		Filterparameter pointer.
## \item		Filteridentifier.
## \end{itemize}
##
## \subsubsection*{firfilt\_s}
## \begin{verbatim}
## firfilt_s(filt_par,cur_branch,x,y,
##      len,ana_syn)
##      struct fir_filterspec *filt_par;
##      int cur_branch;
##      float *x, *y;
##      int len;
##      Boolean ana_syn;
## \end{verbatim}
## \begin{itemize}
##   \item	Filterparameters.
##   \item	Current branch: 0 or 1; lowpass or highpass.
##   \item	Input data vector.
##   \item	Output data vector.
##   \item	Vector length.
##   \item	Flag indicating wheter this is an synthesis bank (TRUE),
## 		or analysis (FALSE).
## \end{itemize}
##
## \subsubsection*{firfilt\_p}
## \begin{verbatim}
## firfilt_p(fil_par,cur_branch,x,y,len)
##      struct fir_filterspec *fil_par;
##      int     cur_branch;
##      float   *x, *y;
##      int     len;
## \end{verbatim}
## \begin{itemize}
##   \item		Filterparameters.
##   \item		Current branch: 0 or 1; lowpass or highpass.
##   \item		Input data vector.
##   \item		Output data vector.
##   \item		Vector length.
## \end{itemize}
##
## \subsubsection*{free\_matrix2\_f}
## \begin{verbatim}
## free_matrix2_f(n1,n2,ptr)
##      int n1, n2;
##      float **ptr;
## \end{verbatim}
## \begin{itemize}
##   \item		Vertical dimension.
##   \item		Horizontal dimension.
##   \item		Pointer to the matrix.
## \end{itemize}
## 
## 
## 
## \subsection*{Date created}
## March 1993
 *------------------------ Rogaland University Centre -----------------------*/



/* Header files */
#include "qmf.h"



extr_subband_line(x,y,ro,kappa,dim,other_coord,len,forw_rev)
     float **y;
     float *x;
     int ro,kappa,dim,other_coord,len,forw_rev;
{
  int i;
  
  if (dim == 0) /* x coord of vert dimension */ {
    if (forw_rev == 0) /* forward */ {
      for (i=0;i<len;i++) {
	x[i]=y[i][other_coord];
      }
    }
    else {
      for (i=len-1;i>=0;i--) {
	x[len-1-i]=y[i][other_coord];
      }
    }
  }
  else {
    if (forw_rev == 0) /* forward */ {
      for (i=0;i<len;i++) {
	x[i]=y[other_coord][i];
      }
    }
    else {
      for (i=len-1;i>=0;i--) {
	x[len-1-i]=y[other_coord][i];
      }
    }
  }
}




insert_subband_line(x,y,ro,kappa,dim,other_coord,len,forw_rev)
     float **y;
     float *x;
     int ro,kappa,dim,other_coord,len,forw_rev;
{
  int i;
  
  if (dim == 0) /* x coord of vert dimension */ {
    if (forw_rev == 0) {	/* forward */
      for (i=0;i<len;i++) {
	y[i][other_coord]=x[i];
      }
    }
    else {
      for (i=len-1;i>=0;i--) {
	y[len-1-i][other_coord]=x[i];
      }
    }
  }
  else {
    if (forw_rev == 0) /* forward */ {
      for (i=0;i<len;i++) {
	y[other_coord][i]=x[i];
      }
    }
    else {
      for (i=len-1;i>=0;i--) {
	y[other_coord][len-1-i]=x[i];
      }
    }
  }
}




firfilt_s(filt_par,cur_branch,x,y,len,ana_syn)
     struct fir_filterspec *filt_par;
     int cur_branch;
     float *x, *y;
     int len;
     Boolean ana_syn;
     /* FIR filter with symmetric extension */
{
  float sum;
  int n,i,k,tshift;
  
  if (!ana_syn) {
    /* Analysis: Symmetric extension */
    for(i=0;i<(filt_par->no_of_taps+1)/2;i++) {
      x[-i-1]=x[i];
      x[len+i]=x[len-1-i];
    }
    tshift=(filt_par->no_of_taps+1)/2;
  }
  else {        /* Synthesis: Antisymmetric extension in highpass channel */ 
    if (cur_branch==0) {
      for(i=0;i<filt_par->no_of_taps/4;i++) {
        x[-2*i-2]= x[2*i];
        x[-2*i-1]= 0.0;
        x[len+2*i]= x[len-2-2*i];
        x[len+2*i+1]= 0.0;
      }
    }
    else {
      for(i=0;i<filt_par->no_of_taps/4;i++) {
        x[-2*i-2]= -x[2*i];
        x[-2*i-1]= 0.0;
        x[len+2*i]= -x[len-2-2*i];
        x[len+2*i+1]= 0.0;
      }
    }
    tshift=filt_par->no_of_taps/2-1;
  }

  for (n=0;n<len;n++) {
    y[n]=0.0;
    for (k= -tshift;k< -tshift+filt_par->no_of_taps;k++) {
      if (cur_branch == 0) /* Lowpass */ {
        y[n] = y[n] + filt_par->p0[k+tshift]*x[n-k];
      }
      else {
        y[n] = y[n] + filt_par->p1[k+tshift]*x[n-k];
      }
    }
  }
}



firfilt_p(fil_par,cur_branch,x,y,len)
     struct fir_filterspec *fil_par;
     int     cur_branch;
     float   *x, *y;
     int     len;
     /* FIR filter incorporating periodic extension */
{
  float sum;
  int n, k, index;
  
  for (n=0;n<len;n++) {
    y[n]=0.0;
    for (k=0;k<fil_par->no_of_taps;k++) {
      index=n-k;
      if (index >= 0) {
	if (cur_branch == 0) {
	  y[n] = y[n] + fil_par->p0[k]*x[index];
	}
	else {
	  y[n] = y[n] + fil_par->p1[k]*x[index];
	}
      }
      else {
	if (cur_branch == 0) {
	  y[n] = y[n] + fil_par->p0[k]*x[len+index];
	}
	else {
	  y[n] = y[n] + fil_par->p1[k]*x[len+index];
	}
      }
    }
  }
}



char *
filterfile(filter_id)
     char *filter_id;
{
  static char filename[MAX_FILT_FILE_NAME];
  strcpy(filename,getenv(FILT_DIR));

  /* Remove trailing blanks */
  while(filename[strlen(filename)-1]==' ')
    filename[strlen(filename)-1]='\0';

  /* If necessary: add slash */
  if(filename[strlen(filename)]==' ')
    strcat(filename,"/");

  strcat(filename,filter_id);
  strcat(filename,FILT_EXT);
  return (filename);
}


read_iir(filt_par,filter_id)
     struct filterspec *filt_par;
     char *filter_id;
{
  strcpy(filt_par->filterfile,filterfile(filter_id));
  read_filtfile (filt_par);
}





read_fir(filt_par,filter_id)
     struct fir_filterspec *filt_par;
     char *filter_id;
{
  int i;
  FILE *f1;

  /* Build the filter filename */
  strcpy(filt_par->filterfile,filterfile(filter_id));

  /* Initialize variables */
  for (i=0;i<MAX_NTAPS;i++) {
    filt_par->p0[i]=0.0;
    filt_par->p1[i]=0.0;
  }
  
  /* Open filterfile */
  if ( (f1 = fopen(filt_par->filterfile,"r")) == NULL )
    perr(HE_OPEN,filt_par->filterfile);
  
  /* Read filterfile: number of taps */
  fscanf(f1," %d", &(filt_par->no_of_taps));

  /* restict to only even number of taps at present! */
  if ( (filt_par->no_of_taps % 2) != 0 ) {
    fprintf(stderr,"only even no. of taps permitted!\n");
    exit(1);
  }

  /* If number of taps is negative, then the filter being read in is 
     a non-symmetric filter, and ALL filtercoeffisients (not just the
     first half) has to be read from file */
  if (filt_par->no_of_taps < 0) {
    filt_par->no_of_taps = ABS(filt_par->no_of_taps);
    for (i=0;i<filt_par->no_of_taps;i++) {
      fscanf(f1,"%f",&(filt_par->p0[i]));
    }
  }
  else {
    for (i=0;i<filt_par->no_of_taps/2;i++) {
      fscanf(f1,"%f",&(filt_par->p0[filt_par->no_of_taps/2 + i]));
      filt_par->p0[filt_par->no_of_taps/2 - i - 1]=
	filt_par->p0[filt_par->no_of_taps/2 + i];
    }
  }
  fclose(f1);
  for(i=0;i<filt_par->no_of_taps;i++) {
    if (i%2==0)
      filt_par->p1[i]=filt_par->p0[filt_par->no_of_taps - i - 1];
    else
      filt_par->p1[i]= -filt_par->p0[filt_par->no_of_taps - i - 1];
  }
}




read_filtfile(filt_params)
     struct filterspec *filt_params;
{
  FILE *f1;
  int i,j, tmp;
  char x;
  float t0, t1, t2, t3;
  if ( (f1 = fopen(filt_params->filterfile,"r")) == NULL ) {
    fprintf(stderr,"can't open filter file!\n");
    exit(1);
  }
  fscanf(f1," %d %d ", &(filt_params->no_of_branches),
	 &(filt_params->no_of_branches));
  for (i=1;i<=filt_params->no_of_branches;i++) {
    fscanf(f1,"%d %d %d",
	   &(filt_params->filter_order[filt_params->no_of_branches-i]),
	   &tmp,&tmp);
    for (j=0;j<filt_params->filter_order[filt_params->no_of_branches-i];j++) {
      fscanf(f1,"%f",&(filt_params->a[filt_params->no_of_branches-i][j]));
    }
  }
  
  /* The filter coefficients have now been read in the form of
     cascaded first order sections, it is now desired to convert
     to a composite filter of appropriate order.
     */
  for (i=0;i<filt_params->no_of_branches;i++) {
    switch(filt_params->filter_order[i])
      {
      case 0: 
	break;
      case 1: 
	break;
      case 2: 
	t0=filt_params->a[i][0]*filt_params->a[i][1];
	t1=filt_params->a[i][0]
	  +filt_params->a[i][1];
	filt_params->a[i][0]=t0;
	filt_params->a[i][1]=t1;
	break;
      case 3: 
	t0=filt_params->a[i][0]*filt_params->a[i][1]*filt_params->a[i][2];
	t1=filt_params->a[i][0] *filt_params->a[i][1] 
	  + filt_params->a[i][0]*filt_params->a[i][2]
	  + filt_params->a[i][1] *filt_params->a[i][2];
	t2=filt_params->a[i][0] 
	  + filt_params->a[i][1] 
	  + filt_params->a[i][2];
	filt_params->a[i][0]=t0;
	filt_params->a[i][1]=t1;
	filt_params->a[i][2]=t2;
	break;
      case 4: 
	t0=filt_params->a[i][0]*filt_params->a[i][1]*filt_params->a[i][2]
	  *filt_params->a[i][3];
	t1=filt_params->a[i][0]*filt_params->a[i][1]*filt_params->a[i][2] 
	  + filt_params->a[i][0] *filt_params->a[i][2]*filt_params->a[i][3]
	  + filt_params->a[i][0] *filt_params->a[i][1]*filt_params->a[i][3]
	  + filt_params->a[i][1] *filt_params->a[i][2]*filt_params->a[i][3];
	t2=filt_params->a[i][0]*filt_params->a[i][1]
	  + filt_params->a[i][0]*filt_params->a[i][2]
	  + filt_params->a[i][0]*filt_params->a[i][3]
	  + filt_params->a[i][1]*filt_params->a[i][2]
	  + filt_params->a[i][1]*filt_params->a[i][3]
	  + filt_params->a[i][2]*filt_params->a[i][3];
	t3=filt_params->a[i][0] 
	  + filt_params->a[i][1] 
	  + filt_params->a[i][2] 
	  + filt_params->a[i][3];
	filt_params->a[i][0]=t0;
	filt_params->a[i][1]=t1;
	filt_params->a[i][2]=t2;
	filt_params->a[i][3]=t3;
	break;
      default:  fprintf(stderr,"no support for higher order flts \n");
	exit(1);
      }
  }
  
  fclose(f1);
  return 0;
}




allpass_c(fil_par,cur_branch,x,y,len)
     struct filterspec *fil_par;
     int        cur_branch;
     float      *x, *y;
     int        len;
{
  float sum;
  int n, l, index1, index2;
  
  for (n=0;n<len;n++) {
    sum=0.0;
    if (n < fil_par->filter_order[cur_branch]) {
      for (l=0;l<fil_par->filter_order[cur_branch];l++) {
	index1=n-l;
	index2=n-fil_par->filter_order[cur_branch]+l;
	if (index1 >= 0) {
	  if (index2 >= 0) {
	    sum=fil_par->a[cur_branch][l] *(x[index1] -y[index2]) + sum;
	  }
	  else {
	    sum=fil_par->a[cur_branch][l] *x[index1] + sum;
	  }
	}
	else {
	  if (index2 >= 0) {
	    sum=fil_par->a[cur_branch][l] *(-y[index2]) + sum;
	  }
	}
      }
    }
    else {
      /* n >= fil_par->filter_order[cur_branch] ==> no need to check indexes */
      for (l=0;l<fil_par->filter_order[cur_branch];l++) {
	sum= fil_par->a[cur_branch][l] 
	  *(x[n-l]-y[n-fil_par->filter_order[cur_branch]+l])+sum;
      }
    }
    if ((n - fil_par->filter_order[cur_branch]) >= 0) {
      y[n]=sum+x[n-fil_par->filter_order[cur_branch]];
    }
    else {
      y[n]=sum;
    }
  }
  /* linear convolution is completed, the circular modification is
     done by filtering the sequence a second time using values
     from the first filtering as initial conditions.
     */
  for (n=0;n<len;n++) {
    sum=0.0;
    if (n < fil_par->filter_order[cur_branch]) {
      for (l=0;l<fil_par->filter_order[cur_branch];l++) {
	index1=n-l;
	index2=n-fil_par->filter_order[cur_branch]+l;
	if (index1 < 0) index1=index1+len;
	if (index2 < 0) index2=index2+len;
	sum=fil_par->a[cur_branch][l] *(x[index1] -y[index2]) + sum;
      }
    }
    else {
      for (l=0;l<fil_par->filter_order[cur_branch];l++) {
	index1=n-l;
	index2=n-fil_par->filter_order[cur_branch]+l;
	sum=fil_par->a[cur_branch][l] *(x[index1] -y[index2]) + sum;
      }
    }
    if ((n - fil_par->filter_order[cur_branch]) >= 0) {
      y[n]=sum+x[n-fil_par->filter_order[cur_branch]];
    }
    else {
      y[n]=sum+x[n-fil_par->filter_order[cur_branch]+len];
    }
  }
}



float *matrix1_f(n1)
     int n1;
{
  float *m;
  
  m = (float *) malloc(n1*sizeof(float));
  if (!m) {
    fprintf(stderr,"allocation error in matrix1_d\n");
    exit(1);
  }
  return m;
}




float **matrix2_f(n1,n2)
     int n1, n2;
{
  int i ;
  float **m;
  
  m = (float **) malloc((unsigned) n1*sizeof(float*));
  if (!m) {
    fprintf(stderr,"allocation error in matrix2_d\n");
    exit(1);
  }
  for (i=0;i<n1;i++) {
    m[i]=(float *) malloc((unsigned) n2*sizeof(float));
    
    if (!m[i]) {
      fprintf(stderr,"allocation error in matrix2_d\n");
      exit(1);
    }
  }
  return m;
}


free_matrix2_f(n1,n2,ptr)
     int n1, n2;
     float **ptr;
{
  int i ;
  
  for (i=0;i<n1;i++) 
    free(ptr[i]);
  free(ptr);
}
  


float ****matrix4_f(n1,n2,n3,n4)
     int n1, n2, n3, n4;
{
  int i, j, k ;
  float ****m;
  
  m = (float ****) malloc((unsigned) n1*sizeof(float ***));
  if (!m) {
    fprintf(stderr,"allocation error in matrix2_d\n");
    exit(1);
  }
  for (i=0;i<n1;i++) {
    m[i]=(float ***) malloc((unsigned) n2*sizeof(float **));
    if (!m[i]) {
      fprintf(stderr,"allocation error in matrix2_d\n");
      exit(1);
    }
    for (j=0;j<n2;j++) {
      m[i][j]=(float **) 
	malloc((unsigned) n3*sizeof(float *));
      if (!m[i][j]) {
	fprintf(stderr,"allocation error in matrix2_d\n");
	exit(1);
      }
      for (k=0;k<n3;k++) {
	m[i][j][k]=(float *)
	  malloc((unsigned) n4*sizeof(float));
	if (!m[i][j][k]) {
	  fprintf(stderr, "allocation error in matrix2_d\n");
	  exit(1);
	}
      }
    }
  }
  return m;
}





unsigned char **matrix2_uch(n1,n2)
     int n1, n2;
{
  int i ;
  unsigned char **m;
  
  m = (unsigned char **) malloc((unsigned) n1*sizeof(unsigned char *));
  if (!m) {
    fprintf(stderr,"allocation error in matrix2_uch\n");
    exit(1);
  }
  for (i=0;i<n1;i++) {
    m[i]=(unsigned char*) malloc((unsigned) n2*sizeof(unsigned char));
    if (!m[i]) {
      fprintf(stderr,"allocation error in matrix2_uch\n");
      exit(1);
    }
  }
  return m;
}



void nrerror(error_text)
     char error_text[];
{
  void exit();
  
  fprintf(stderr,"Numerical Recipes run-time error...\n");
  fprintf(stderr,"%s\n",error_text);
  fprintf(stderr,"...now exiting to system...\n");
  exit(1);
}



float *vector(nl,nh)
     int nl,nh;
{
  float *v;
  
  v=(float *)malloc((unsigned) (nh-nl+1)*sizeof(float));
  if (!v) nrerror("allocation failure in vector()");
  return v-nl;
}

int *ivector(nl,nh)
     int nl,nh;
{
  int *v;
  
  v=(int *)malloc((unsigned) (nh-nl+1)*sizeof(int));
  if (!v) nrerror("allocation failure in ivector()");
  return v-nl;
}

double *dvector(nl,nh)
     int nl,nh;
{
  double *v;
  
  v=(double *)malloc((unsigned) (nh-nl+1)*sizeof(double));
  if (!v) nrerror("allocation failure in dvector()");
  return v-nl;
}



float **matrix(nrl,nrh,ncl,nch)
     int nrl,nrh,ncl,nch;
{
  int i;
  float **m;
  
  m=(float **) malloc((unsigned) (nrh-nrl+1)*sizeof(float*));
  if (!m) nrerror("allocation failure 1 in matrix()");
  m -= nrl;
  
  for(i=nrl;i<=nrh;i++) {
    m[i]=(float *) malloc((unsigned) (nch-ncl+1)*sizeof(float));
    if (!m[i]) nrerror("allocation failure 2 in matrix()");
    m[i] -= ncl;
  }
  return m;
}

double **dmatrix(nrl,nrh,ncl,nch)
     int nrl,nrh,ncl,nch;
{
  int i;
  double **m;
  
  m=(double **) malloc((unsigned) (nrh-nrl+1)*sizeof(double*));
  if (!m) nrerror("allocation failure 1 in dmatrix()");
  m -= nrl;
  
  for(i=nrl;i<=nrh;i++) {
    m[i]=(double *) malloc((unsigned) (nch-ncl+1)*sizeof(double));
    if (!m[i]) nrerror("allocation failure 2 in dmatrix()");
    m[i] -= ncl;
  }
  return m;
}

int **imatrix(nrl,nrh,ncl,nch)
     int nrl,nrh,ncl,nch;
{
  int i,**m;
  
  m=(int **)malloc((unsigned) (nrh-nrl+1)*sizeof(int*));
  if (!m) nrerror("allocation failure 1 in imatrix()");
  m -= nrl;
  
  for(i=nrl;i<=nrh;i++) {
    m[i]=(int *)malloc((unsigned) (nch-ncl+1)*sizeof(int));
    if (!m[i]) nrerror("allocation failure 2 in imatrix()");
    m[i] -= ncl;
  }
  return m;
}



float **submatrix(a,oldrl,oldrh,oldcl,oldch,newrl,newcl)
     float **a;
     int oldrl,oldrh,oldcl,oldch,newrl,newcl;
{
  int i,j;
  float **m;
  
  m=(float **) malloc((unsigned) (oldrh-oldrl+1)*sizeof(float*));
  if (!m) nrerror("allocation failure in submatrix()");
  m -= newrl;
  
  for(i=oldrl,j=newrl;i<=oldrh;i++,j++) m[j]=a[i]+oldcl-newcl;
  
  return m;
}



void free_vector(v,nl,nh)
     float *v;
     int nl,nh;
{
  free((char*) (v+nl));
}

void free_ivector(v,nl,nh)
     int *v,nl,nh;
{
  free((char*) (v+nl));
}

void free_dvector(v,nl,nh)
     double *v;
     int nl,nh;
{
  free((char*) (v+nl));
}



void free_matrix(m,nrl,nrh,ncl,nch)
     float **m;
     int nrl,nrh,ncl,nch;
{
  int i;
  
  for(i=nrh;i>=nrl;i--) free((char*) (m[i]+ncl));
  free((char*) (m+nrl));
}

void free_dmatrix(m,nrl,nrh,ncl,nch)
     double **m;
     int nrl,nrh,ncl,nch;
{
  int i;
  
  for(i=nrh;i>=nrl;i--) free((char*) (m[i]+ncl));
  free((char*) (m+nrl));
}

void free_imatrix(m,nrl,nrh,ncl,nch)
     int **m;
     int nrl,nrh,ncl,nch;
{
  int i;
  
  for(i=nrh;i>=nrl;i--) free((char*) (m[i]+ncl));
  free((char*) (m+nrl));
}



void free_submatrix(b,nrl,nrh,ncl,nch)
     float **b;
     int nrl,nrh,ncl,nch;
{
  free((char*) (b+nrl));
}



float **convert_matrix(a,nrl,nrh,ncl,nch)
     float *a;
     int nrl,nrh,ncl,nch;
{
  int i,j,nrow,ncol;
  float **m;
  
  nrow=nrh-nrl+1;
  ncol=nch-ncl+1;
  m = (float **) malloc((unsigned) (nrow)*sizeof(float*));
  if (!m) nrerror("allocation failure in convert_matrix()");
  m -= nrl;
  for(i=0,j=nrl;i<=nrow-1;i++,j++) m[j]=a+ncol*i-ncl;
  return m;
}



void free_convert_matrix(b,nrl,nrh,ncl,nch)
     float **b;
     int nrl,nrh,ncl,nch;
{
  free((char*) (b+nrl));
}

