/*----------------------- Rogaland University Centre -------------------------
/*---------------------------------------------------------------------------
## \section{seq2dec}
## 
## \subsection*{Purpose}
##  Convert a sequence to a format like the output from "qmf -d ..."
##	(which is a mosaic of critically decimated subbands).
##	This actually implies decimation of the sequence.
## 
## \subsection*{Usage }
## \begin{verbatim}
## seq2dec 
##    [-sn struct<integer> | 
##     -s struct<string> | 
##     -sb struct<string>]
##    [imagefile]
## \end{verbatim}
## \begin{description}
##   \item[-s]
##      Filter structure string, defined as a quad tree.
##      See the module {\em filtertree} (CompileFiltStructTree)
##      for a description of the structure.
##   \item[-sb]
##      Filterstructure string, defined as a binary tree.
##      See the module {\em filtertree} (CompileFiltStructTreeFromBin)
##      for a description of the structure.
##   \item[-sn]
##      Filter structure number, between 1 and 11 (default 4).
##      See the module {\em filtertree} (BuildFiltStructTree)
##      for a description of the structure.
## \end{description}
## 
## \subsection*{Date created}
## 7/3-94
+---------------------------------------------------------------------------+
|	Trygve Randen                                                       |
|	Rogaland University Center, Dept. of Elect. and Computer Eng.       |
|	P.O. Box 2557 Ullandhaug, 4004 Stavanger, Norway                    |
|	Tel.: +47 51 87 43 36, Fax.: +47 51 87 42 36                        |
|	E-mail: tranden@hsr.no                                              |
+--------------------------------------------------------------------------*/



/* Header files */
#include "qmf.h"
#include "mcf.h"

static Flag_Format flagfmt[]={
  {"sn",{"s","sb",LASTFLAG},1,{
    {PTBOOLEAN,"TRUE"},
    {PTINT,"4","struct"},
    LASTPARAMETER}},
  {"sb",{"sn","s",LASTFLAG},1,{
    {PTBOOLEAN,"FALSE"},
    {PTSTRING,"(,)","struct"},
    LASTPARAMETER}},
  {"s",{"sn","sb",LASTFLAG},1,{
    {PTBOOLEAN,"FALSE"},
    {PTSTRING,"(,)","struct"},
    LASTPARAMETER}},
  LASTFLAG};

int types[]={PFFLOAT,LASTTYPE};

main(argc,argv)
     int argc;
     char *argv[];
{
  struct header hd;		/* Input file header (any type) */
  struct header hdp;		/* Input file header (float) */
  struct header hdo;		/* Output file header (float) */
  int method;			/* Conversion methods */
  Boolean structStr1DFlag;	/* Structure string specified (bin) */
  Boolean structStrFlag;	/* Structure string specified (quad) */
  Boolean structNoFlag;		/* Structure number specified */
  int structNo;
  char *structStr1D;		/* Structure description string (bin) */
  char *structStr;		/* Structure description string (quad) */
  FilterTree filtertree;	/* Filter structure tree */
  Filename filename;
  FILE *fp;
  int leastCommonDecimation;
  int frame;

  hipserrprt = HEL_ERROR;	/* Suppress messaging from HIPS routines */

  /* Initialize variables */
  Progname=strsave(*argv);
  parseargs(argc,argv,flagfmt,&structNoFlag,&structNo,
	    &structStr1DFlag,&structStr1D,&structStrFlag,&structStr,
	    FFONE,&filename);

  /* Build the filter structure binary tree */
  if (structStrFlag) 
    filtertree = CompileFiltStructTree(structStr,"");
  else if (structStr1DFlag) 
    filtertree = CompileFiltStructTreeFromBin(structStr1D,"");
  else if (structNoFlag)
    filtertree = BuildFiltStructTree(structNo,"");

  /* Read the input header */
  fp = hfopenr(filename);
  fread_hdr_a(fp,&hd,filename);
  method = fset_conversion(&hd,&hdp,types,filename);

  /* Allocate the output image */
  dup_headern(&hdp,&hdo);
  alloc_imagez(&hdo);

  /* Extract and write the subbands */
  get_sub (fp, filename, &hd,&hdp,&hdo,method,filtertree,
	   0,0,hd.orows-1,hd.ocols-1,1,&frame);

  /* Write frame */
  h_copy(&hdo,&hdp);
  hd.num_frame=1;
  write_headeru(&hd,argc,argv);
  write_imagec(&hd,&hdp,method,TRUE,0);
}




get_sub (fp, filename, hd,hdp,hdo,method,ft,fr,fc,tr,tc,decimation,frame)
     FILE *fp;
     Filename filename;
     struct header *hd, *hdp, *hdo;
     int method;
     FilterTree ft;
     int fr,fc,tr,tc;
     int decimation;
     int *frame;
{
  int ro, kappa;
  int i,j,k,l;
  float **x, **y;

  if (ft != NULL) 
    for (ro=0; ro<2; ro++) 
      for (kappa=0; kappa<2; kappa++) 
        get_sub (fp, filename, hd, hdp, hdo, method,
		 ft->subband[ro][kappa],
		 (ro==0 ? fr : (fr+tr+1)/2),
		 (kappa==0 ? fc : (fc+tc+1)/2),
		 (ro==1 ? tr : (fr+tr-1)/2),
		 (kappa==1 ? tc : (fc+tc-1)/2),
		 decimation*2,frame);
  else {
    fread_imagec(fp,hd,hdp,method,(*frame)++,filename);	
    /* Memory allocation etc. */
    x=(float **) matrix2(hdp);
    y=(float **) matrix2(hdo);

    /* Transfer pixels */
    for (i=fr; i<=tr; i++) {
      for (j=fc; j<=tc; j++) {
	y[i][j]=x[(i-fr)*decimation][(j-fc)*decimation];
      }
    }
	    
    /* Release allocated memory */
    free_matrix2(y);
    free_matrix2(x);
  }
}
  
