/*----------------------- Rogaland University Centre -------------------------
## \section{struct2latex}
## 
## 
## \subsection*{Purpose}
## Convert a filter structure description to a LaTeX picture indicating
## the subband-split on a square.
## This is a handy documentation tool for documenting filterstructures.
## The output image will be within a square of 1 x 1 in a standard
## unit determined by the enclosing LaTeX document.
## 
## \subsection*{Description}
## Read a filtertree specification from the commandline in one of
## three formats, and print a LaTeX compatible output to the 
## standard output.
## For any of the inputs, the quad-tree specification is printed to
## the standard error.
## 
## 
## \subsection*{Usage}
## \begin{verbatim}
## struct2latex
##    [-sn struct<integer> | 
##     -s struct<string> | 
##     -sb struct<string>]
## \end{verbatim}
## \begin{description}
##   \item[-s]
## 	Filter structure string, defined as a quad tree.
## 	See the module {\em filtertree} (CompileFiltStructTree)
## 	for a description of the structure.
##   \item[-sb]
## 	Filterstructure string, defined as a binary tree.
## 	See the module {\em filtertree} (CompileFiltStructTreeFromBin)
## 	for a description of the structure.
##   \item[-sn]
## 	Filter structure number, between 1 and 11 (default 4).
## 	See the module {\em filtertree} (BuildFiltStructTree)
## 	for a description of the structure.
## \end{description}
## 
## 
## 
## \subsection*{Date created}
## April 1993
 *------------------------ Rogaland University Centre -----------------------*/



/* Header files */
#include "qmf.h"
#include "math.h"

static Flag_Format flagfmt[]={
  {"sn",{"s","sb",LASTFLAG},1,{
    {PTINT,"4","struct"},
    LASTPARAMETER}},
  {"sb",{"sn","s",LASTFLAG},1,{
    {PTBOOLEAN,"FALSE"},
    {PTSTRING,"(,)","struct"},
    LASTPARAMETER}},
  {"s",{"sn","sb",LASTFLAG},1,{
    {PTBOOLEAN,"FALSE"},
    {PTSTRING,"(,)","struct"},
    LASTPARAMETER}},
  LASTFLAG};



main(argc,argv)
     int argc;
     char *argv[];
{
  Boolean structStr1DFlag;	/* Structure string specified (bin) */
  Boolean structStrFlag;	/* Structure string specified (quad) */
  int struct_no;
  char *structStr1D;		/* Structure description string (bin) */
  char *structStr;		/* Structure description string (quad) */
  FilterTree filtertree;	/* Filter structure tree */

  /* Initialize variables */
  Progname=strsave(*argv);
  parseargs(argc,argv,flagfmt,&struct_no,
	    &structStr1DFlag,&structStr1D,&structStrFlag,&structStr,
	    FFNONE);

  /* Build the filter structure binary tree */
  if (structStrFlag) 
    filtertree = CompileFiltStructTree(structStr,"");
  else if (structStr1DFlag) 
    filtertree = CompileFiltStructTreeFromBin(structStr1D,"");
  else
    filtertree = BuildFiltStructTree(struct_no,"");

  /* Print the correspondig quad-tree specification to stderr */
  PrintFiltStructTree (filtertree);

  /* Make the LaTeX picture */
  printf("\\begin{picture}(1,1)(0,-1)\n");
  print_line(0.,0.,1.,0.);
  print_line(0.,1.,1.,0.);
  print_line(0.,1.,0.,1.);
  print_line(1.,1.,0.,1.);

  print_subband (filtertree,0.0,0.0,1.0,1.0);

  printf("\\end{picture}\n");
}



print_subband (ft,x_orig,y_orig,x_dim,y_dim)
     FilterTree ft;
     float x_orig, y_orig;
     float x_dim, y_dim;
{
  int ro, kappa;

  if (ft != NULL) {
    print_line (x_orig, y_orig+y_dim/2, x_dim, 0.0);
    print_line (x_orig+x_dim/2, y_orig+y_dim, 0.0, y_dim);
    
    for (ro=0; ro<2; ro++)
      for (kappa=0; kappa<2; kappa++)
	print_subband(ft->subband[ro][kappa],
		      x_orig + (kappa == 0 ? 0 : x_dim/2),
		      y_orig + (ro    == 0 ? 0 : y_dim/2),
		      x_dim/2, y_dim/2);
  }
}



print_line (x_orig,y_orig,x_dim,y_dim)
     float x_orig, y_orig, x_dim, y_dim;
     /* Remember that for LaTeX, x is horizontal coordinate */
{
  printf("\\put(%10.6f,%10.6f){\\line(%d,%d){%10.6f}}\n",
	 x_orig,-y_orig,
	 (x_dim == 0.0 ? 0 : 1),
	 (y_dim == 0.0 ? 0 : 1),
	 x_dim + y_dim);
}



