/*----------------------- Rogaland University Centre -------------------------
## \section{struct\_gabor}
## 
## \subsection*{Purpose}
## Filter the input image by a even-symmetric Gabor filter with the filter
## structure determined by a quad-filertree descriptor.
## 
## \subsection*{Usage}
## \subsubsection*{struct\_gabor}
## \begin{verbatim}
## struct_gabor(ft,hdi,hd,method,sf,alpha,
##      nonlin,blurr)
##      FilterTree ft;
##      struct header *hdi;
##      struct header *hd;
##      int method;
##      double sf;
##      double alpha;
##      int nonlin;
##      int blurr;
## \end{verbatim}
## \begin{itemize}
##   \item	        Filter structure tree
##   \item	        Input (nonfiltered) header
##   \item	        Input format (nonconverted) header
##   \item	        Conversion method
##   \item	        Factor to sigma in lowpass filter
##   \item	        Alpha in nonlinear transform
##   \item	        Nonlinear transform method
##   \item	        Smoothing method
##   \end{itemize}
## 
## \subsection*{Date created}
## April 1993
 *------------------------ Rogaland University Centre -----------------------*/

#include <hipl_format.h>
#include <math.h>
#include <float.h>
#include "struct_defs.h"
#include "mcf.h"



struct_gabor(ft,hdi,hd,method,sf,alpha,nonlin,blurr)
     FilterTree ft;		/* Filter structure tree */
     struct header *hdi;	/* Input (nonfiltered) header */
     struct header *hd;		/* Input format (nonconverted) header */
     int method;		/* Conversion method */
     double sf;			/* Factor to sigma in lowpass filter */
     double alpha;		/* Alpha in nonlinear transform */
     int nonlin;		/* Nonlinear transform method */
     int blurr;			/* Smoothing method */
{
  int frame=0;

  s_g_recurse(ft,0.,0.,1.,1.,hdi,hd,method,sf,alpha,nonlin,blurr,&frame);
}



s_g_recurse(ft,xmin,ymin,xmax,ymax,hdi,hd,method,sf,alpha,nonlin,blurr,frame)
     FilterTree ft;		/* Filter structure tree */
     float xmin,ymin,xmax,ymax;	/* Subband delimitiers */
     struct header *hdi;	/* Input (nonfiltered) header */
     struct header *hd;		/* Input format (nonconverted) header */
     int method;		/* Conversion method */
     double sf;			/* Factor to sigma in lowpass filter */
     double alpha;		/* Alpha in nonlinear transform */
     int nonlin;		/* Nonlinear transform method */
     int blurr;			/* Smoothing method */
     int *frame;		/* Output frame number */
{
  int ro, kappa;	/* Subband numbers */
  struct header hdo;	/* Output (filtered) image */
  double U0;		/* Centre frequency */
  double theta;		/* Orientation */

  for (ro=0; ro<2; ro++) {
    for (kappa=0; kappa<2; kappa++) {
      /* Filtering subbands recursively */
      if (ft->subband[ro][kappa] != NULL) {
	s_g_recurse(ft->subband[ro][kappa],
		    (ro==0 ? xmin : (xmin+xmax)/2.0),
		    (kappa==0 ? ymin : (ymin+ymax)/2.0),
		    (ro==1 ? xmax : (xmin+xmax)/2.0),
		    (kappa==1 ? ymax : (ymin+ymax)/2.0),
		    hdi,hd,method,sf,alpha,nonlin,blurr,frame);
      }
      else {
        /* No more subbands, filter and write this subband */
	dup_headern(hdi,&hdo);
	alloc_image(&hdo);
	U0 = 0.5 * sqrt(sqr(ro*(xmin+xmax)/2. + (xmin+xmax)/4.) +
			sqr(kappa*(ymin+ymax)/2. + (ymin+ymax)/4.));
	theta = H_180_PI * atan((kappa*(ymin+ymax)/2. + (ymin+ymax)/4.) /
				(ro*(xmin+xmax)/2. + (xmin+xmax)/4.));
	one_channel_filter(hdi, &hdo, U0, 0., theta,
			   (xmax-xmin)/4. * sqrt(log(H_SQRT2) / 4.),
			   (ymax-ymin)/4. * sqrt(log(H_SQRT2) / 4.),
			   sf/U0, alpha, EVEN_GABOR, nonlin, blurr,TRUE);
	/* This calculation of SigmaU and SigmaV implies a cutoff at the
	   pass-band -- stop-band limit of -3dB */
	fprintf(stderr,"write\n");
	mcf_write_image(hd,&hdo,(*frame)++,TRUE);
	free_image(&hdo);
      }
    } /* for-kappa */
  } /* for-ro */
}
