/*----------------------- Rogaland University Centre -------------------------
## \section{tanh}
## 
## 
## \subsection*{Purpose}
## Apply the tanh(..) nonlinearity.
## 
## \subsection*{Description}
## See: ``Multi-channel filtering techniques for texture segmentation and
##      surface quality inspection'',
##      Farshid Farrokhnia, Ph.D. dissertation,
##      Michigan State University, Dept. of Elect. Eng., 1990
## If any structure specification is given (parameter {\tt -s},
## {\tt -sn} or {\tt -sb}), it is assumed that the input is decimated,
## as generated by {\tt qmf -d $\cdots$}, and each subband is treated
## separately (important when scaling is concerned).
## 
## \subsection*{Usage}
## \begin{verbatim}
## tanh [-sn struct<integer> | 
##       -s struct<string> | 
##       -sb struct<string>]
##    [-alpha alpha in nonlin. 
##            transf.<real>] 
##    [imagefile]
## \end{verbatim}
## \begin{description}
##   \item[-s]
##      Filter structure string, defined as a quad tree.
##      See the module {\em filtertree} (CompileFiltStructTree)
##      for a description of the structure.
##   \item[-sb]
##      Filterstructure string, defined as a binary tree.
##      See the module {\em filtertree} (CompileFiltStructTreeFromBin)
##      for a description of the structure.
##   \item[-sn]
##      Filter structure number, between 1 and 11 (default 4).
##      See the module {\em filtertree} (BuildFiltStructTree)
##      for a description of the structure.
##   \item[-alpha]
## 	$\alpha$ parameter in tanh(..) nonlinearity
## \end{description}
## 
## 
## 
## \subsection*{Date created}
## March 1993
 *------------------------ Rogaland University Centre -----------------------*/


#include <hipl_format.h>
#include "qmf.h"
#include "mcf.h"

/* Define the commandline-format */
static Flag_Format flagfmt[] = {
  {"sn",{"s","sb",LASTFLAG},1,{
    {PTBOOLEAN,"FALSE"},
    {PTINT,"4","struct"},
    LASTPARAMETER}},
  {"sb",{"sn","s",LASTFLAG},1,{
    {PTBOOLEAN,"FALSE"},
    {PTSTRING,"(,)","struct"},
    LASTPARAMETER}},
  {"s",{"sn","sb",LASTFLAG},1,{
    {PTBOOLEAN,"FALSE"},
    {PTSTRING,"(,)","struct"},
    LASTPARAMETER}},
  {"alpha",{LASTFLAG},1,
     {{PTDOUBLE,"0.25","alpha in nonlin. transf."}, LASTPARAMETER}},
  LASTFLAG};

int types[] = {PFFLOAT,LASTTYPE};
double alpha;			/* Parameter to the nonlinearity */

main(argc,argv)
     int argc;
     char **argv;
{
  int fr;			/* Current frame */
  int method;
  FILE *fp;
  Filename filename;
  struct header hd;
  struct header hdp;
  Boolean structStr1DFlag;	/* Structure string specified (bin) */
  Boolean structStrFlag;	/* Structure string specified (quad) */
  Boolean structNoFlag;		/* Structure number specified */
  int structNo;
  char *structStr1D;		/* Structure description string (bin) */
  char *structStr;		/* Structure description string (quad) */
  FilterTree filtertree;	/* Filter structure tree */
  Boolean decimated=TRUE;

  hipserrprt = HEL_ERROR;	/* Suppress messaging from HIPS routines */

  /* Read the commandline */
  Progname = strsave(*argv);
  parseargs(argc,argv,flagfmt,&structNoFlag,&structNo,
	    &structStr1DFlag,&structStr1D,&structStrFlag,&structStr,
	    &alpha,FFONE,&filename);

  /* Build the filter structure binary tree */
  if (structStrFlag) 
    filtertree = CompileFiltStructTree(structStr,"");
  else if (structStr1DFlag) 
    filtertree = CompileFiltStructTreeFromBin(structStr1D,"");
  else if (structNoFlag)
    filtertree = BuildFiltStructTree(structNo,"");
  else
    decimated = FALSE;

  /* Read in the image header */
  fp = hfopenr(filename);
  fread_hdr_a(fp,&hd,filename);
  method = fset_conversion(&hd,&hdp,types,filename);
  write_headeru(&hdp,argc,argv);
  
  /* Nonlinear transform */
  if (decimated) {
    fread_imagec(fp,&hd,&hdp,method,0,filename);
    decimated_tanh(&hdp,filtertree,0,0,hdp.orows-1,hdp.ocols-1);
    write_image(&hd,0);
  }
  else 
    for (fr=0; fr<hdp.num_frame; fr++) {
      fread_imagec(fp,&hd,&hdp,method,fr,filename);
      tanh_subband(&hdp,&hdp);
      write_image(&hdp,fr);
    }
}




decimated_tanh (hd,ft,fr,fc,tr,tc)
     struct header *hd;
     FilterTree ft;
     int fr,fc,tr,tc;
{
  int ro, kappa;

  if (ft != NULL) 
    for (ro=0; ro<2; ro++) 
      for (kappa=0; kappa<2; kappa++) 
        decimated_tanh(hd, ft->subband[ro][kappa],
		       (ro==0 ? fr : (fr+tr+1)/2),
		       (kappa==0 ? fc : (fc+tc+1)/2),
		       (ro==1 ? tr : (fr+tr-1)/2),
		       (kappa==1 ? tc : (fc+tc-1)/2));
  
  else {
    setroi(hd,fr,fc,tr-fr+1,tc-fc+1);
    tanh_subband(hd,hd);
    setroi(hd,0,0,hd->orows,hd->ocols);
  }
}




tanh_subband (hdi, hdo)
     struct header *hdi, *hdo; 
{
  float mean,min,max,scale;

  /* Set the mean to 0 */
  h_mean(hdi,&mean,FALSE);
  h_linscale(hdi,hdo,1.,-mean);
  
  /* Scale the image to [-1,1] */
  h_minmax(hdo,&min,&max,FALSE);
  scale = MAX(abs(max),abs(min));
  h_divscale(hdo,hdo,&scale);

  /* Transform the image */
  mcf_tanh_nonlin(hdo,hdo,alpha);
}
