/*---------------------------------------------------------------------------
## \section{winstat}
## 
## \subsection*{Purpose}
## Compute the local statistics (variance or standard deviation) 
## of an image.
## 
## \subsection*{Description}
## 
## \subsection*{Usage }
## \begin{verbatim}
## winstat [-w rows<integer> [cols<integer>]]
##    [-lm | -gm | -nm]
##    [-var | -sd] [imagefile]
## \end{verbatim}
## \begin{description}
##   \item[-w]		Window size
##   \item[-lm]		Use local mean when computing the variance (default)
##   \item[-gm]		use global mean...
##   \item[-nm]		Don't use mean (assume that the mean is zero)
##   \item[-var]	Calculate windowed variance (default)
##   \item[-sd]		Calculate windowed standard deviation
## \end{description}
## 
## \subsection*{Date created}
## June 1993
+---------------------------------------------------------------------------+
|	Trygve Randen                                                       |
|	Rogaland University Center, Dept. of Elect. and Computer Eng.       |
|	P.O. Box 2557 Ullandhaug, 4004 Stavanger, Norway                    |
|	Tel.: +47 51 87 43 36, Fax.: +47 51 87 42 36                        |
|	E-mail: tranden@hsr.no                                              |
+--------------------------------------------------------------------------*/




#include <stdlib.h>
#include <limits.h>
#include <float.h>
#include <hipl_format.h>
#include <math.h>
#include "math.h"

/* Conversion */
int types[] = {PFFLOAT,LASTTYPE};

/* Define the commandline-format */
static Flag_Format flagfmt[] = {
  {"w",{LASTFLAG},1,{
    {PTINT,"1","rows"}, 
    {PTINT,"-1","cols"}, 
    LASTPARAMETER}},
  {"lm",{"gm","nm",LASTFLAG},0,{
    {PTBOOLEAN,"TRUE"},
    LASTPARAMETER}},
  {"gm",{"lm","nm",LASTFLAG},0,{
    {PTBOOLEAN,"FALSE"},
    LASTPARAMETER}},
  {"nm",{"lm","gm",LASTFLAG},0,{
    {PTBOOLEAN,"FALSE"},
    LASTPARAMETER}},
  {"var",{"sd",LASTFLAG},0,{
    {PTBOOLEAN,"TRUE"},
    LASTPARAMETER}},
  {"sd",{"var",LASTFLAG},0,{
    {PTBOOLEAN,"FALSE"},
    LASTPARAMETER}},
  LASTFLAG};

main(argc,argv)
     int argc;
     char **argv;
{
  struct header hd, hdi, hdo;	/* Nonconverted and conv. in, conv. out. */
  FILE *fp;			/* Input HIPS image file pointer */
  Filename filename;		/* Input HIPS image file name */
  int wr, wc;			/* Window sizes (rows/columns) */
  int x,y;			/* Image coordiantes */
  int i,j;			/* Window coordinates */
  int fr;			/* Current frame */
  int method;			/* Conversion method */
  float **img, **var;		/* Input/output image referenced as an array */
  Boolean lm, gm, nm;		/* Mean flags */
  Boolean varFlag, sdFlag;	/* Compute variance or standard deviation */
  float mean;			/* Mean value */
  float winsize;		/* Local window size (number of pixels) */

  /* Initialize variables and parse commandline */
  hipserrprt = HEL_ERROR;
  Progname = strsave(*argv);
  parseargs(argc,argv,flagfmt,&wr,&wc,&lm,&gm,&nm,&varFlag,&sdFlag,
	    FFONE,&filename);
  if (wc < 0)
    wc = wr;

  fp = hfopenr(filename);
  fread_hdr_a(fp,&hd,filename);
  method = fset_conversion(&hd,&hdi,types,filename);
  dup_headern(&hdi,&hdo);
  alloc_image(&hdo);
  var = (float **) matrix2 (&hdo);
  write_headeru(&hdo,argc,argv);

  for (fr=0; fr<hd.num_frame; fr++) {

    /* Read input */
    fread_imagec(fp,&hd,&hdi,method,0,filename);
    img = (float **) matrix2 (&hdi);

    if (gm) {
      /* Compute global mean */
      mean = 0.;
      for(x=0; x<hd.rows; x++) {
	for(y=0; y<hd.cols; y++) {
	  mean += img[x][y];
	}
      }
      mean = mean / (hd.rows * hd.cols);
    }
    else if (nm) {
      /* No mean, use mean equal to zero */
      mean = 0.;
    }
  
    for(x=0; x<hd.rows; x++) {
      for(y=0; y<hd.cols; y++) {

	/* Compute local mean and variance */
        var[x][y] = 0.;
	if (lm)
	  mean = 0;
	for (i=MAX(0,x-wr/2); i<MIN(hd.rows,x+wr/2); i++) {
	  for (j=MAX(0,y-wc/2); j<MIN(hd.cols,y+wc/2); j++) {
	    if (lm)	/* Compute local mean */
	      mean += img[i][j];
	    var[x][y] += sqr(img[i][j]);
	  }
	}
	winsize = (float) (MIN(hd.rows,x+wr/2) - MAX(0,x-wr/2))
	  *(MIN(hd.cols,y+wc/2) - MAX(0,y-wc/2));
	if (lm) 
	  mean = mean / winsize;
	var[x][y] = var[x][y] / winsize - sqr(mean);

	/* Convert to standard deviation if specified on command line */
	if (sdFlag) {
	  var[x][y] = sqrt(var[x][y]);
	}
      }
    }
    write_image(&hdo,fr);
    free_matrix2(img);
  }

  return(0);
}
