/*---------------------------------------------------------------------------
## \section{inspec2bibtex}
## 
## \subsection*{Purpose}
## Convert inspec export data to bibtex format
## 
## \subsection*{Description}
## Read inspec data from standard in and write bibtex data to standard
## out.
## 
## \subsection*{Usage }
## \begin{verbatim}
## inspec2bibtex < inspec-file > bibtex-file
## \end{verbatim}
##
## \subsection*{Date created}
## 30 January 1994
+---------------------------------------------------------------------------+
|	Trygve Randen                                                       |
|	Rogaland University Center, Dept. of Elect. and Computer Eng.       |
|	P.O. Box 2557 Ullandhaug, 4004 Stavanger, Norway                    |
|	Tel.: +47 51 87 43 36, Fax.: +47 51 87 42 36                        |
|	E-mail: tranden@hsr.no                                              |
+--------------------------------------------------------------------------*/

#include <stdio.h>
#include <string.h>

#define MAX_LINE_LEN 100
#define MAX_LINES 500
#define FALSE 0
#define TRUE 1

#define SUPPRESS 0
#define ADDRESS 1
#define ANNOTE 2
#define AUTHOR 3
#define BOOKTITLE 4
#define CHAPTER 5
#define EDITION 6
#define EDITOR 7
#define HOWPUBLISHED 8
#define INSTITUTION 9
#define JOURNAL 10
#define KEY 11
#define MONTH 12
#define NOTE 13
#define NUMBER 14
#define ORGANIZATION 15
#define PAGES 16
#define PUBLISHER 17
#define SCHOOL 18
#define SERIES 19
#define TITLE 20
#define TYPE 21
#define VOLUME 22
#define YEAR 23
#define _DOC_TYPE_ 50
#define _SUBMITTED_ 51

#define UNKNOWN_DOC_TYPE 0
#define BOOK 1
#define BOOK_CHAPTER 2
#define CONFERENCE_PAPER 3
#define CONFERENCE_PAPER_IN_JOURNAL 4
#define CONFERENCE_PROCEEDINGS 5
#define CONFERENCE_PROCEEDINGS_IN_JOURNAL 6
#define DISSERTATION 7
#define JOURNAL_PAPER 8
#define REPORT 9
#define REPORT_SECTION 10


typedef int Boolean;



char *
read_line(char *line)
{
  return gets(line);
}


int
read_record(char *lines)
{
  Boolean contin=TRUE;
  Boolean blanks=TRUE;
  char line[MAX_LINE_LEN];
  int line_no=0;
  
  /* Skip introducing blank lines and read first line of record */
  while(contin && blanks) {
    contin=(read_line(line)!=NULL);
    blanks=(line[0]==0);
  }
  strcpy((lines+MAX_LINE_LEN*line_no++),line);

  while(contin && !blanks) {
    contin=(read_line(line)!=NULL);
    blanks=(line[0]==0);
    strcpy((lines+MAX_LINE_LEN*line_no++),line);
  }

  return line_no-1;
}



modify_author(char *str)
{
  /* Replace `;' by `and' */
  int i,j;
  char newstr[MAX_LINE_LEN];
  
  j=0;
  for(i=0;i<strlen(str);i++) {
    if (str[i]==';') {
      newstr[j++]=' ';
      newstr[j++]='a';
      newstr[j++]='n';
      newstr[j++]='d';
      newstr[j++]=' ';
    }
    else
      newstr[j++]=str[i];
  }
  newstr[j++]='\0';
  strcpy(str,newstr);
}


int
extract_year(char *line)
{
  /* Will extract the first four digit number in this line as the year */
  int i,j,year;
  char year_str[5];

  j=0;
  for (i=0; i<strlen(line); i++) {
    if (line[i]<='9' && line[i]>='0')
      j++;
    else
      j=0;
    if (j==4)
      break;
  }
  strncpy(year_str,&line[i-3],4);
  year_str[4]='\0';
  year=atoi(year_str);
  if (year>2100 || year<1900)
    fprintf(stderr,"Year in illegal range: %d\n",year);
  return year;
}



extract_author_for_key(char *author, 
		       char *line)
{
  int i;

  strncpy(author,line,5);
  author[5]='\0';
  for(i=0; i<5; i++) {
    if(author[i]==',')
      author[i]='\0';
    else if(author[i]==' ')
      author[i]='-';
    else
      author[i]=tolower(author[i]);
  }
}


char *
extract_word(char *word,
	     char *line)
{
  int i;

  strncpy(word,line,MAX_LINE_LEN);
  for (i=0; i<strlen(word); i++)
    if (word[i]==' ')
      word[i]='\0';
  return word;
}



int
extract_doc_type(char *doc_type,
		 char *line)
{
  int doc_type_no;

  if      (strcmp(line,"Book")==0) {
    doc_type_no=BOOK;
    strcpy(doc_type,"book");
  }
  else if (strcmp(line,"Book Chapter")==0) {
    doc_type_no=BOOK_CHAPTER;
    strcpy(doc_type,"incollection");
  }
  else if (strcmp(line,"Conference Paper")==0) {
    doc_type_no=CONFERENCE_PAPER;
    strcpy(doc_type,"inproceedings");
  }
  else if (strcmp(line,"Conference Paper in Journal")==0) {
    doc_type_no=CONFERENCE_PAPER_IN_JOURNAL;
    strcpy(doc_type,"article");
  }
  else if (strcmp(line,"Conference Proceedings in Journal")==0)  {
    doc_type_no=CONFERENCE_PROCEEDINGS;
    strcpy(doc_type,"article"); 
  }
  else if (strcmp(line,"Conference Proceedings")==0) {
    doc_type_no=CONFERENCE_PROCEEDINGS_IN_JOURNAL;
    strcpy(doc_type,"proceedings");
  }
  else if (strcmp(line,"Dissertation")==0) {
    doc_type_no=DISSERTATION;
    strcpy(doc_type,"phdthesis");
  }
  else if (strcmp(line,"Journal Paper")==0) {
    doc_type_no=JOURNAL_PAPER;
    strcpy(doc_type,"article");
  }
  else if (strcmp(line,"Report")==0) {
    doc_type_no=REPORT;
    strcpy(doc_type,"techreport");
  }
  else if (strcmp(line,"Report Section")==0){
    doc_type_no=REPORT_SECTION;
    strcpy(doc_type,"inproceedings");
  }
  else {
    fprintf(stderr,"ERROR:   Unknown document type: %s\n",line);
    fprintf(stderr,"         Modify this conversion program!\n");
    doc_type_no=UNKNOWN_DOC_TYPE;
  } 
  return doc_type_no;
}



make_bibtex_line(int field, 
		 char *data,
		 char *line,
		 int *line_cnt,
		 Boolean first_field_line)
{
  char field_txt[15];
  char volume[MAX_LINE_LEN];
  char number[MAX_LINE_LEN];
  char pages[MAX_LINE_LEN];

  /* Trim leading blanks */
  while (data[0]==' ')
    data++;

  /* Replace field number with text */
  if (first_field_line) {
    switch (field) {
      case ADDRESS	: strcpy(field_txt,"ADDRESS"); 		break;
      case ANNOTE	: strcpy(field_txt,"ANNOTE"); 		break;
      case AUTHOR	: strcpy(field_txt,"AUTHOR"); 		break;
      case BOOKTITLE    : strcpy(field_txt,"BOOKTITLE");	break;
      case CHAPTER      : strcpy(field_txt,"CHAPTER"); 		break;
      case EDITION      : strcpy(field_txt,"EDITION"); 		break;
      case EDITOR	: strcpy(field_txt,"EDITOR"); 		break;
      case HOWPUBLISHED	: strcpy(field_txt,"HOWPUBLISHED");	break;
      case INSTITUTION  : strcpy(field_txt,"INSTITUTION"); 	break;
      case JOURNAL      : strcpy(field_txt,"JOURNAL"); 		break;
      case KEY  	: strcpy(field_txt,"KEY"); 		break;
      case MONTH	: strcpy(field_txt,"MONTH"); 		break;
      case NOTE 	: strcpy(field_txt,"NOTE"); 		break;
      case NUMBER	: strcpy(field_txt,"NUMBER"); 		break;
      case ORGANIZATION : strcpy(field_txt,"ORGANIZATION"); 	break;
      case PAGES	: strcpy(field_txt,"PAGES"); 		break;
      case PUBLISHER    : strcpy(field_txt,"PUBLISHER"); 	break;
      case SCHOOL	: strcpy(field_txt,"SCHOOL"); 		break;
      case SERIES	: strcpy(field_txt,"SERIES"); 		break;
      case TITLE	: strcpy(field_txt,"TITLE"); 		break;
      case TYPE 	: strcpy(field_txt,"TYPE"); 		break;
      case VOLUME	: strcpy(field_txt,"VOLUME"); 		break;
      case YEAR 	: strcpy(field_txt,"YEAR"); 		break;
      default		: field_txt[0]='\0';
    }
  }
  else
    field_txt[0]='\0';

  if (field==YEAR)
    sprintf(data,"%d",extract_year(data));
  if (field==VOLUME) {
    extract_word (volume,data);
    if (strstr(data,"Iss: ") != NULL)
      make_bibtex_line(NUMBER,
		       extract_word (number,strstr(data,"Iss: ")+5),
		       line, line_cnt, TRUE);
    if (strstr(data,"p. ") != NULL)
      make_bibtex_line(PAGES,
		       extract_word (pages ,strstr(data,"p. ")+3),
		       line, line_cnt, TRUE);
    strcpy(data,volume);
  }

  if (field_txt[0]=='\0')
    sprintf(line+((*line_cnt)++)*MAX_LINE_LEN,"  %-15s    %s",field_txt,data);
  else
    sprintf(line+((*line_cnt)++)*MAX_LINE_LEN,"  %-15s = {%s",field_txt,data);
}
  


int
convert_record(char *in_lines, int in_line_cnt)
{
  char out_lines[MAX_LINES][MAX_LINE_LEN];
  char main_author[MAX_LINE_LEN];
  char bibtex_key[MAX_LINE_LEN];
  char line[MAX_LINE_LEN];
  char key[MAX_LINE_LEN];
  char doc_type[MAX_LINE_LEN];
  int in_line_no;
  int out_line_cnt=0;
  int field;
  int year;
  int i;
  int doc_type_no;
  Boolean first_field_line;
  static Boolean key_NN_warned;

  main_author[0]='\0';
  for (in_line_no=0;in_line_no<in_line_cnt;in_line_no++) {
    strncpy(line,in_lines+MAX_LINE_LEN*in_line_no,MAX_LINE_LEN);
    first_field_line=TRUE;

    if      (strncmp(line," INSPEC "                               , 8)==0) {field=SUPPRESS;	strcpy(line,line+ 8);}
    else if (strncmp(line," Abstract:     "                        ,15)==0) {field=SUPPRESS;	strcpy(line,line+15);}
    else if (strncmp(line," Affiliation:  "                        ,15)==0) {field=SUPPRESS;	strcpy(line,line+15);}
    else if (strncmp(line," Authors:      "                        ,15)==0) {field=AUTHOR;	strcpy(line,line+15);}
    else if (strncmp(line," Availability: "                        ,15)==0) {field=SUPPRESS;	strcpy(line,line+15);}
    else if (strncmp(line," Book Title:   "                        ,15)==0) {field=BOOKTITLE;	strcpy(line,line+15);}
    else if (strncmp(line," Classification: "                      ,17)==0) {field=SUPPRESS;	strcpy(line,line+17);}
    else if (strncmp(line," Conf. Title:  "                        ,15)==0) {field=BOOKTITLE;	strcpy(line,line+15);}
    else if (strncmp(line," Doc Type:     "                        ,15)==0) {field=_DOC_TYPE_;	strcpy(line,line+15);}
    else if (strncmp(line," Editors:      "                        ,15)==0) {field=EDITOR;	strcpy(line,line+15);}
    else if (strncmp(line," Free Terms: "                          ,13)==0) {field=SUPPRESS;	strcpy(line,line+13);}
    else if (strncmp(line," Issued by:    "                        ,15)==0) {field=SCHOOL;	strcpy(line,line+15);}
    else if (strncmp(line," Journal:      "                        ,15)==0) {field=JOURNAL;	strcpy(line,line+15);}
    else if (strncmp(line,"               Vol: "                   ,20)==0) {field=VOLUME;	strcpy(line,line+20);}
    else if (strncmp(line,"               Date: "                  ,21)==0) {field=YEAR;	strcpy(line,line+21);}
    else if (strncmp(line,"               Country of Publication: ",39)==0) {field=SUPPRESS;	strcpy(line,line+39);}
    else if (strncmp(line,"               ISSN: "                  ,21)==0) {field=SUPPRESS;	strcpy(line,line+21);}
    else if (strncmp(line,"               ISBN: "                  ,21)==0) {field=SUPPRESS;	strcpy(line,line+21);}
    else if (strncmp(line,"               CCC: "                   ,20)==0) {field=SUPPRESS;	strcpy(line,line+20);}
    else if (strncmp(line,"               Conf. Date: "            ,27)==0) {field=SUPPRESS;	strcpy(line,line+27);}
    else if (strncmp(line,"               Conf. Loc: "             ,26)==0) {field=SUPPRESS;	strcpy(line,line+26);}
    else if (strncmp(line,"               Conf. Sponsor: "         ,30)==0) {field=ORGANIZATION;strcpy(line,line+30);}
    else if (strncmp(line,"               p. "                     ,18)==0) {field=PAGES;	strcpy(line,line+18);}
    else if (strncmp(line," Language:     "                        ,15)==0) {field=SUPPRESS;	strcpy(line,line+15);}
    else if (strncmp(line," Publisher:    "                        ,15)==0) {field=PUBLISHER;	strcpy(line,line+15);}
    else if (strncmp(line," Submitted:    "                        ,15)==0) {field=_SUBMITTED_;	strcpy(line,line+15);}
    else if (strncmp(line," Thesaurus: "                           ,12)==0) {field=SUPPRESS;	strcpy(line,line+12);}
    else if (strncmp(line," Title:        "                        ,15)==0) {field=TITLE;	strcpy(line,line+15);}
    else if (strncmp(line," Treatment:    "                        ,15)==0) {field=SUPPRESS;	strcpy(line,line+15);}
    else if (strncmp(line," Numerical Index: "                     ,18)==0) {field=SUPPRESS;	strcpy(line,line+18);}
    else if (strncmp(line," Chemical Index: "                      ,17)==0) {field=SUPPRESS;	strcpy(line,line+17);}
    else if (strncmp(line," Report No:    "                        ,15)==0) {field=NUMBER;	strcpy(line,line+15);}
    else if (strncmp(line," Report Title: "                        ,15)==0) {field=BOOKTITLE;	strcpy(line,line+15);}
    else if (strncmp(line,"   "                                    , 3)==0) first_field_line=FALSE;
    else {
      fprintf(stderr,"ERROR:   Unknown field: %s\n",line);
      fprintf(stderr,"         Modify this conversion program!\n");
      field=SUPPRESS;
    }
    switch (field) 
      {
      case _DOC_TYPE_:
	doc_type_no=extract_doc_type(doc_type,line);
	break;
      case AUTHOR:
	modify_author(line); 
	if (first_field_line)
	  extract_author_for_key(main_author,line); 
	break;
      case EDITOR:
	modify_author(line); 
	break;
      case YEAR:
	if (doc_type_no != DISSERTATION) 
	  year=extract_year(line); 
	else
	  field=SUPPRESS;
	break;
      case _SUBMITTED_:
	year=extract_year(line); 
	field=YEAR;
	break;
      case BOOKTITLE:
	if (doc_type_no == CONFERENCE_PROCEEDINGS 
	    || doc_type_no == CONFERENCE_PROCEEDINGS_IN_JOURNAL)
	  field=TITLE;
	break;
      case SCHOOL:
	if (doc_type_no == REPORT)
	  field = INSTITUTION;
	else if (doc_type_no == REPORT_SECTION)
	  field = ORGANIZATION;
      break;
      }
    if(field != SUPPRESS) 
      make_bibtex_line(field,line,*out_lines,&out_line_cnt,first_field_line);
  }

  if (main_author[0]=='\0') {
    if (!key_NN_warned) {
      fprintf(stderr,"WARNING: No author available for field(s) with key NN.\n");
      fprintf(stderr,"         Modify the output manually.\n");
      key_NN_warned=TRUE;
    }
    strcpy(key,"NN");
  }
  else
    sprintf(key,"%s%d",main_author,year%100);

  /* Transfer out_lines to in_lines -- before return */
  sprintf(in_lines,"@%s{%s,",doc_type,key);
  for (i=1; i<out_line_cnt; i++) {
    strcpy(line,out_lines[i]);
    if (strncmp(out_lines[i+1],"   ",3)!=0 || i==out_line_cnt-1) {
      sprintf(line,"%s}",line);
      if (i<out_line_cnt-1)
	sprintf(line,"%s,",line);
    }
    strcpy(in_lines+i*MAX_LINE_LEN,line);
  }
  sprintf(in_lines+(i++)*MAX_LINE_LEN,"}");
  return i;
}



print_bibtex_record(char *data_record,
		    int line_cnt)
{
  int i;

  for(i=0;i<line_cnt;i++)
    printf("%s\n",data_record+i*MAX_LINE_LEN);
}





main() 
{
  char data_record[MAX_LINES][MAX_LINE_LEN];
  int i,line_cnt;

  while((line_cnt=read_record(*data_record))!=0) {
    line_cnt=convert_record(*data_record,line_cnt);
    print_bibtex_record(*data_record,line_cnt);
  }
}
