/*---------------------------------------------------------------------------
## \section{hips2eps}
## 
## \subsection*{Purpose}
## Convert a HIPS image to Encapsulated PostScript
## 
## \subsection*{Usage }
## \begin{verbatim}
## hips2eps [-bpp bits per pixel<integer>] 
##   [-scale] [imagefile]
## \end{verbatim}
## \begin{description}
##   \item[-bpp]
##	Bits per pixel. If this parameter is not specified, it is determined
##	automatically by the input header.
##   \item[-scale]
##	Scale the image to exploit the maximum dynamic range of the output.
## \end{description}
## 
## \subsection*{Date created}
## 21 February 1994
+---------------------------------------------------------------------------+
|	Trygve Randen                                                       |
|	Rogaland University Center, Dept. of Elect. and Computer Eng.       |
|	P.O. Box 2557 Ullandhaug, 4004 Stavanger, Norway                    |
|	Tel.: +47 51 87 43 36, Fax.: +47 51 87 42 36                        |
|	E-mail: tranden@hsr.no                                              |
+--------------------------------------------------------------------------*/


#include <hipl_format.h>
#include <time.h>
#include <math.h>

#define PS_OFFSET 36.000
#define SCREEN_FREQ 60
#define MAX_LL 80

static Flag_Format flagfmt[]={
  {"bpp",{LASTFLAG},1,{
    {PTINT,"-1","bits per pixel"},
    LASTPARAMETER}},
  {"scale",{LASTFLAG},0,{
    {PTBOOLEAN,"FALSE"},
    LASTPARAMETER}},
  LASTFLAG};

int types[]={PFINT,LASTTYPE};





main(int argc,char **argv)
{
  struct header hd, hdp;
  int method,i,j,f,k;
  Filename filename;
  FILE *fp;
  int **img;
  time_t tloc;
  int line_len;
  int opix;
  int bpp;			/* Bits per pixel */
  Boolean maxScale;		/* Scale to exploit max range? */

  /* Read commandline */
  Progname = strsave(*argv);
  parseargs(argc,argv,flagfmt,&bpp,&maxScale,FFONE,&filename);
  
  fp = hfopenr(filename);
  fread_hdr_a(fp,&hd,filename);
  method = fset_conversion(&hd,&hdp,types,filename);
  fread_imagec(fp,&hd,&hdp,method,0,filename);
  img = (int **) matrix2(&hdp);
  time(&tloc);

  /* Determine bits per pixel in output */
  if (bpp < 0) 
    if (hd.sizepix==0)
      bpp=1;
    else
      bpp=8*hd.sizepix;

  /* If necessary: scale the image */
  if (bpp < 8*hd.sizepix || maxScale) {
    linscale(&hdp,&hdp,0,(int) pow(2.0,bpp)-1);
  }

  /* Print header of PostScript file */
  printf("%%!\n");
  printf("%%%%Creator: hips2eps (Trygve Randen)\n");
  printf("%%%%Title:\n");
  printf("%%%%BoundingBox: %.3f %.3f %.3f %.3f\n",
	 PS_OFFSET,PS_OFFSET,
	 PS_OFFSET+141.73,
	 PS_OFFSET+141.73*hdp.rows/hdp.cols);
  printf("%%%%CreationDate: %s\n", ctime(&tloc));
  printf("\n");
  printf("%%%%Pages: 1\n");
  printf("%%%%EndComments\n");
  printf("\n");
  printf("/picstr %d string def \n",hdp.cols);
  printf("gsave \n");
  /*
     printf("%d 45 {180 mul cos exch 180 mul cos add 2 div} setscreen\n",
     SCREEN_FREQ);
     */
  printf("%.3f %.3f translate\n",PS_OFFSET,PS_OFFSET);
  printf("%.3f %.3f scale\n", 141.73, 141.73*hdp.rows/hdp.cols);
  printf("%d %d %d\n",hdp.cols,hdp.rows,bpp);
  printf("[%d 0 0 -%d 0 %d]\n",hdp.cols,hdp.rows,hdp.rows);
  printf("{currentfile picstr readhexstring pop}\n");
  printf("image\n");
  printf("\n");

  line_len=0;
  for(i=0; i<hdp.rows;i++) {
    for(j=0;j<hdp.cols;j++) {
      switch (bpp) 
	{
	case 1: 
	  opix=0;
	  for (k=0; k<4; k++) {
	    if (j+k < hdp.cols && img[i][j+k])
	      opix+=(int) pow(2.0,3-k);
	  }
	  j+=3;
	  printf("%01x", opix); 
	  line_len+=1;
	  break;
	case 2: 
	  printf("%01x", img[i][j]*4+img[i][j+1]);
	  j+=1;
	  line_len+=1;
	  break;
	case 4:
	  printf("%01x", img[i][j]); 
	  line_len+=1;
	  break;
	case 8:	 
	  printf("%02x", img[i][j]); 
	  line_len+=2;
	  break;
	case 16:
	  printf("%04x", img[i][j]); 
	  line_len+=4;
	  break;
	case 32:
	  printf("%08x", img[i][j]); 
	  line_len+=8;
	  break;
	default:
	  perr(HE_IMSG,"Unsupported number of bits per pixel in EPS\n");
	}
      if(line_len >= MAX_LL) {
	printf("\n");
	line_len=0;
      }
    }
  }

  /* Print footer of PostScript file */
  printf("\n");
  printf("grestore\n");


  free_matrix2(img);
  return(0);
}



linscale (hdi,hdo,n_min,n_max)
     struct header *hdi,*hdo;
     int n_min,n_max;
{
  int o_min, o_max;	/* Current minimum and maximum */
  float b, c;		/* Compute opix=b*ipix+c */
  int **in, **out;
  int i,j;

  in = (int **) matrix2(hdi);
  out = (int **) matrix2(hdo);
  h_minmax(hdi,&o_min,&o_max,FALSE);
  if (o_max > o_min) {
    b = ((float) (n_max - n_min)) / ((float) (o_max - o_min)) ;
    c = (float) n_min - b * o_min;
  }
  else {
    b = 1.0;
    c = 0.0;
  }

  for (i=0; i<hdi->orows; i++) {
    for (j=0; j<hdi->ocols; j++) {
      out[i][j]=(int) ((float) in[i][j]*b + c);
    }
  }

  free_matrix2(out);
  free_matrix2(in);
}			      
